/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   Portions Copyright (C) 2001 Adam Loghry
** 
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
//#include "cmds.h"
//#include "../lib/confparse/configparse.h"

RCSID("$Id: commands.c,v 1.20 2001/10/02 01:40:35 cbehrens Exp $");
//typedef struct _ircd_mode               IRCD_MODE;
//typedef struct _ircd_capab              IRCD_CAPAB;
#if 0
struct _ircd_mode
{
        char                            mode;
        char                            plusminus;
        IRCD_CLIENT                     *cptr;
        
        char                            *banstring;
        char                            *exceptstring;  /* all three of these variables could probably share the same space, it might */
        char                            *quietstring;   /* be worth coming back and setting it up that way ...... later */
        
        IRCD_MODE                       *next;
};
#endif

/* channel modes are here because this is channel.c .. doh
 * user modes are here because this file also houses m_umode
 * user flags may be found in user.c
 *
 * FIXME: break the mode commands into mode.c ? */
int _channel_modes[] =
{
        CHANNEL_MODES_PRIVATE,          'p',
        CHANNEL_MODES_SECRET,           's',
        CHANNEL_MODES_MODERATED,        'm',
        CHANNEL_MODES_NOPRIVMSGS,       'n',
        CHANNEL_MODES_TOPICLIMIT,       't',
        CHANNEL_MODES_INVITEONLY,       'i', 
        CHANNEL_MODES_OPERONLY,         'O', 
        CHANNEL_MODES_REGGEDONLY,       'R',
        0, 0
};

int _user_modes[] =
{
        CLIENT_UMODES_OPER,             'o',
        CLIENT_UMODES_LOCOP,            'O',
        CLIENT_UMODES_FMODE,            'f',
        CLIENT_UMODES_IMODE,            'i',
        CLIENT_UMODES_KMODE,            'k',
        CLIENT_UMODES_SMODE,            's',
        CLIENT_UMODES_WMODE,            'w',
        CLIENT_UMODES_ZMODE,            'z',
        CLIENT_UMODES_YMODE,            'y',
        CLIENT_UMODES_CMODE,            'c',
        CLIENT_UMODES_NMODE,            'n',
        CLIENT_UMODES_SERVICE,          'x',
        CLIENT_UMODES_REGISTERED,       'r',
        CLIENT_UMODES_ONLYREGMSGS,      'R',
        CLIENT_UMODES_DEAF,             'd',
        0, 0
};

int _channel_canjoin(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr, char *key);
IRCD_CHANNELLIST *_channel_ismember(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr);
// int _channel_cansend(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr);
IRCD_MODE *_ircdmode_create(IRCD_MODE **ircdmodes, IRCD_MODE **lastircdmode);
void _ircdmode_free(IRCD_MODE **bucket);
// void _user_del_channels(IRCD_CLIENT *pptr);
void _channel_invite_del(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr);
// void _send_channel_modes(IRCD_CLIENT *cptr, IRCD_CHANNEL *chptr);
// void _channel_add_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, int flag);
// void _channel_del_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, IRCD_CHANNELLIST *chlptr);

static void _build_channel_modes(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr, char *mbuf, char *pbuf);
static void _channel_invite_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr);
static int _channel_change_opflag(IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr, IRCD_MODE *opmodeptr, int mine);
static char *_make_nick_user_host(char *nick, char *user, char *host, char *buf, char **hostptr);
static int _channel_ban_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask);
static int _channel_ban_del(IRCD_CHANNEL *chptr, char *str);
static int _channel_except_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask);
static int _channel_except_del(IRCD_CHANNEL *chptr, char *str);
static int _channel_quiet_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask);
static int _channel_quiet_del(IRCD_CHANNEL *chptr, char *str);
static void _channelname_clean(unsigned char *chname);

#ifdef HMC
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *nick, char *prefix, char *args, char *hmc);
#else
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *nick, char *prefix, char *args);
#endif

static char *_make_nick_user_host(char *nick, char *user, char *host, char *buf, char **hostptr)
{
	static char	star = '*';
	char	*ptr = buf;
	char	*sptr;
	int		n;

	if (!nick || !*nick)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=nick,n=NICKLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr++ = '!';
	if (!user || !*user)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=user,n=USERLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr++ = '@';
	if (hostptr)
		*hostptr = ptr;
	if (!host || !*host)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=host,n=HOSTLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr = '\0';
	return buf;
}

void _channel_del_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, IRCD_CHANNELLIST *chlptr)
{
	IRCD_CLIENTLIST		*clptr;

	if (chlptr == NULL)
		if (!(chlptr = IsMember(chptr, pptr)))
			return;
	pptr->c_numchannels--;
	clptr=chlptr->clptr;
	chlptr->clptr = NULL;
	del_clientlist_from_list(&(chptr->members), clptr);
	if (!--chptr->numusers)
	{
		/* remove channel */
		_channel_ban_del(chptr, NULL);
		assert(!chptr->nbans);
		hash_del_channel(chptr);
		if (chptr->topic)
			MyFree(chptr->topic);
		if (chptr->topic_who)
			MyFree(chptr->topic_who);
		if (chptr->key)
			MyFree(chptr->key);
		MyFree(chptr);
		Globalinfo->channels--;
	}
	MyFree(clptr);
}

void _channel_add_user(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, int flag)
{
	IRCD_CHANNELLIST	*chlptr;
	IRCD_CLIENTLIST		*clptr;

	chlptr = (IRCD_CHANNELLIST *)MyMalloc(sizeof(IRCD_CHANNELLIST));
	clptr = (IRCD_CLIENTLIST *)MyMalloc(sizeof(IRCD_CLIENTLIST));
	chlptr->chptr = chptr;
	clptr->cptr = pptr;
	clptr->modes = flag;
	add_channellist_to_list(&(pptr->c_channels), chlptr);
	add_clientlist_to_list(&(chptr->members), clptr);
	chlptr->clptr = clptr;
	clptr->chlptr = chlptr;
	pptr->c_numchannels++;
	chptr->numusers++;
}

static void _build_channel_modes(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr, char *mbuf, char *pbuf)
{
	int		ismem = 0;

	if (!IsServer(cptr) && (chptr->limit || chptr->key))
		ismem = (IsMember(chptr, cptr) != NULL);
	*mbuf++ = '+';
	if (chptr->modes & CHANNEL_MODES_SECRET)
		*mbuf++ = 's';
	/* else */
	if (chptr->modes & CHANNEL_MODES_PRIVATE)
		*mbuf++ = 'p';
	if (chptr->modes & CHANNEL_MODES_MODERATED)
		*mbuf++ = 'm';
	if (chptr->modes & CHANNEL_MODES_TOPICLIMIT)
		*mbuf++ = 't';
	if (chptr->modes & CHANNEL_MODES_INVITEONLY)
		*mbuf++ = 'i';
	if (chptr->modes & CHANNEL_MODES_NOPRIVMSGS)
		*mbuf++ = 'n';
	if (chptr->limit)
	{
		*mbuf++ = 'l';
		if (IsServer(cptr) || ismem)
			pbuf += sprintf(pbuf, "%d ", chptr->limit);
	}
	if (chptr->key)
	{
		*mbuf++ = 'k';
		if (IsServer(cptr) || ismem)
			pbuf += strlen(strcpy(pbuf, chptr->key));
	}
	*mbuf = '\0';
	*pbuf = '\0';
}

void _send_channel_modes(IRCD_CLIENT *cptr, IRCD_CHANNEL *chptr)
{
	char				modebuf[64];
	char				parabuf[512];
	char				buf[512];
	char				*t;
	IRCD_CLIENTLIST		*clptr;
	IRCD_CLIENTLIST		*anop = NULL;
	IRCD_CLIENTLIST		*skip = NULL;
	IRCD_BAN			*bans;
	IRCD_EXCEPT			*excepts;
	IRCD_QUIET			*quiets;
	int					n = 0;
	int					paralen;
	int					namlen;
	char				*cp;
	int					count;
	int					send;

	if (*chptr->name != '#') /* don't send '&' channels */
		return;
	_build_channel_modes(chptr, cptr, modebuf, parabuf);

	t = buf + sprintf(buf, ":%s SJOIN %u %s %s %s%s:",
			meptr->name, chptr->creation, chptr->name, modebuf,
			parabuf, *parabuf ? " " : "");

	for(clptr=chptr->members;clptr;clptr=clptr->next)
	{
		if (clptr->modes & CHANNEL_MODES_CHANOP)
		{
			anop = clptr;
			break;
		}
	}
	clptr = NULL;
	for(;;)
	{
		if (anop)
		{
			clptr = skip = anop;
			anop = NULL;
		}
		else
		{
			if (clptr == NULL || clptr == skip)
				clptr = chptr->members;
			else
				clptr = clptr->next;
			if (clptr && clptr == skip)
				clptr = clptr->next;
			if (clptr == NULL)
				break;
		}
		if (clptr->modes & CHANNEL_MODES_CHANOP)
			*t++ = '@';
		if (clptr->modes & CHANNEL_MODES_VOICE)
			*t++ = '+';
		strcpy(t, clptr->cptr->name);
		t += strlen(t);
		*t++ = ' ';
		n++;
		if (t - buf > 512 - HOSTLEN - 3)
		{
			*t++ = '\0';
			if (t[-1] == ' ')
				t[-1] = '\0';
			send_to_client(cptr, "%s", buf);
			t = buf + sprintf(buf, ":%s SJOIN %u %s 0 :",
					meptr->name, chptr->creation, chptr->name);
			n = 0;
		}
	}
	if (n)
	{
		*t++ = '\0';
		if (t[-1] == ' ')
			t[-1] = '\0';
		send_to_client(cptr, "%s", buf);
	}
	*parabuf = '\0';
	*modebuf = '+';
	cp = modebuf + 1;
	count = 0;
	paralen = 0;
	send = 0;
		
	/* send bans */

	for(bans=chptr->bans;bans;bans=bans->next)
	{
		if (paralen + (namlen = strlen(bans->banstring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, bans->banstring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'b';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, bans->banstring);
				paralen = namlen;
				*cp++ = 'b';
			}
			count = 1;
		}
	}

	/* then send exceptions and quiets */
	for(excepts=chptr->excepts;excepts;excepts=excepts->next)
	{
		if (paralen + (namlen = strlen(excepts->estring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, excepts->estring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'e';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, excepts->estring);
				paralen = namlen;
				*cp++ = 'e';
			}
			count = 1;
		}
	}

	for(quiets=chptr->quiets;quiets;quiets=quiets->next)
	{
		if (paralen + (namlen = strlen(quiets->qstring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, quiets->qstring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'q';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, quiets->qstring);
				paralen = namlen;
				*cp++ = 'q';
			}
			count = 1;
		}
	}


	if (*parabuf)
	{
		*cp = '\0';
		send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
				chptr->name, modebuf, parabuf);
	}
}

static int _channel_change_opflag(IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr, IRCD_MODE *opmodeptr, int mine)
{
	int	m =
		(opmodeptr->mode == 'o' ? CHANNEL_MODES_CHANOP : CHANNEL_MODES_VOICE);
	int	irm = (int)vars_value_get_bynum(IRCD_VAR_IGNORE_REDUNDANT_MODES);

	if (!clptr)
		return 1;
	if (opmodeptr->plusminus == '+')
	{
		if (irm && mine && (clptr->modes & m))
			return 0;
		clptr->modes |= m;
		if (m == CHANNEL_MODES_CHANOP)
			clptr->modes &= ~CHANNEL_MODES_DEOPPED;
	}
	else
	{
		if (irm && mine && !(clptr->modes & m))
			return 0;
		clptr->modes &= ~m;
	}
	return 1;
}

static int _channel_ban_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask)
{
	IRCD_BAN		*ban;

	if (MyClient(pptr))
	{
		collapse(str);
		if (chptr->nbans >= MAXBANS)
			return 0;
		hostptr = strchr(str, '@');
		assert(hostptr != NULL);
		hostptr++;
	}
	for(ban=chptr->bans;ban;ban=ban->next)
	{
		if (MyClient(pptr) &&
				(!match(ban->banstring, str) /* || !match(str, ban->banstring) */ ))
			return 0;
		else if (!mycmp(ban->banstring, str))
			return 0;
	}

	ban = (IRCD_BAN *)MyMalloc(sizeof(IRCD_BAN));
	ban->banstring = MyStrdup(str);
	if (ban->net != -1)
	{
		ban->net = net;
		ban->mask = mask;
		ban->firstpart = (char *)MyMalloc(hostptr - str + 1);
		memcpy(ban->firstpart, str, hostptr - str);
		ban->firstpart[hostptr - str] = '\0';
	}
	ban->bantime = Now;
	if (IsServer(pptr))
		ban->who = MyStrdup(vars_value_get_bynum(IRCD_VAR_NETWORK_HOST)); /* used to be the server name */
	else
	{
		ban->who = MyMalloc(strlen(pptr->name)+strlen(pptr->username)+
					strlen(pptr->hostname)+3);
		sprintf(ban->who, "%s!%s@%s", pptr->name, pptr->username,
				pptr->hostname);
	}
	add_ban_to_list(&(chptr->bans), ban);
	chptr->nbans++;

	return 1;
}

static int _channel_except_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask)
{
	IRCD_EXCEPT	*except;
	
	if (MyClient(pptr))
	{
		collapse(str);
		if (chptr->nexcepts >= MAXEXCEPTS)
			return 0;
		hostptr = strchr(str, '@');
		assert(hostptr != NULL);
		hostptr++;
	}
	for(except=chptr->excepts;except;except=except->next)
	{
		if (MyClient(pptr) &&
				(!match(except->estring, str) /* || !match(str, except->estring) */ ))
			return 0;
		else if (!mycmp(except->estring, str))
			return 0;
	}

	except = (IRCD_EXCEPT *)MyMalloc(sizeof(IRCD_EXCEPT));
	except->estring = MyStrdup(str);
	if (except->net != -1)
	{
		except->net = net;
		except->mask = mask;
		except->firstpart = (char *)MyMalloc(hostptr - str + 1);
		memcpy(except->firstpart, str, hostptr - str);
		except->firstpart[hostptr - str] = '\0';
	}
	except->etime = Now;
	if (IsServer(pptr))	/* was the server name.. */
		except->who = MyStrdup(vars_value_get_bynum(IRCD_VAR_NETWORK_HOST));
	else
	{
		except->who = MyMalloc(strlen(pptr->name)+strlen(pptr->username)+
					strlen(pptr->hostname)+3);
		sprintf(except->who, "%s!%s@%s", pptr->name, pptr->username,
				pptr->hostname);
	}
	add_except_to_list(&(chptr->excepts), except);
	chptr->nexcepts++;

	return 1;
}

static int _channel_quiet_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr, char *str, char *hostptr, ircd_uint32_t net, ircd_uint32_t mask)
{
	IRCD_QUIET		*quiet;
	
	if (MyClient(pptr))
	{
		collapse(str);
		if (chptr->nquiets >= MAXQUIETS)
			return 0;
		hostptr = strchr(str, '@');
		assert(hostptr != NULL);
		hostptr++;
	}
	for(quiet=chptr->quiets;quiet;quiet=quiet->next)
	{
		if (MyClient(pptr) &&
				(!match(quiet->qstring, str) /* || !match(str, quiet->qstring) */ ))
			return 0;
		else if (!mycmp(quiet->qstring, str))
			return 0;
	}

	quiet = (IRCD_QUIET *)MyMalloc(sizeof(IRCD_QUIET));
	quiet->qstring = MyStrdup(str);
	if (quiet->net != -1)
	{
		quiet->net = net;
		quiet->mask = mask;
		quiet->firstpart = (char *)MyMalloc(hostptr - str + 1);
		memcpy(quiet->firstpart, str, hostptr - str);
		quiet->firstpart[hostptr - str] = '\0';
	}
	quiet->qtime = Now;
	if (IsServer(pptr))
		quiet->who = MyStrdup(vars_value_get_bynum(IRCD_VAR_NETWORK_HOST)); /* used to be the server name (pptr->name, duh) */
	else
	{
		quiet->who = MyMalloc(strlen(pptr->name)+strlen(pptr->username)+
					strlen(pptr->hostname)+3);
		sprintf(quiet->who, "%s!%s@%s", pptr->name, pptr->username,
				pptr->hostname);
	}
	add_quiet_to_list(&(chptr->quiets), quiet);
	chptr->nquiets++;

	return 1;
}

static int _channel_ban_del(IRCD_CHANNEL *chptr, char *str)
{
	IRCD_BAN	*ban;
	IRCD_BAN	*nextban;

	for(ban=chptr->bans;ban;ban=nextban)
	{
		nextban = ban->next;
		if (!str || !mycmp(str, ban->banstring))
		{
			MyFree(ban->banstring);
			if (ban->firstpart)
				MyFree(ban->firstpart);
			MyFree(ban->who);
			del_ban_from_list(&(chptr->bans), ban);
			MyFree(ban);
			chptr->nbans--;
		}
	}
	return 1;
}

static int _channel_except_del(IRCD_CHANNEL *chptr, char *str)
{
	IRCD_EXCEPT	*except;
	IRCD_EXCEPT	*nextexcept;

	for(except=chptr->excepts;except;except=nextexcept)
	{
		nextexcept = except->next;
		if (!str || !mycmp(str, except->estring))
		{
			MyFree(except->estring);
			if (except->firstpart)
				MyFree(except->firstpart);
			MyFree(except->who);
			del_except_from_list(&(chptr->excepts), except);
			MyFree(except);
			chptr->nexcepts--;
		}
	}
	return 1;
}

static int _channel_quiet_del(IRCD_CHANNEL *chptr, char *str)
{
	IRCD_QUIET	*quiet;
	IRCD_QUIET	*nextquiet;

	for(quiet=chptr->quiets;quiet;quiet=nextquiet)
	{
		nextquiet = quiet->next;
		if (!str || !mycmp(str, quiet->qstring))
		{
			MyFree(quiet->qstring);
			if (quiet->firstpart)
				MyFree(quiet->firstpart);
			MyFree(quiet->who);
			del_quiet_from_list(&(chptr->quiets), quiet);
			MyFree(quiet);
			chptr->nquiets--;
		}
	}
	return 1;
}

static void _channel_invite_add(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr)
{
	IRCD_INVITE	*invite;

	for(invite=cptr->c_invites_list;invite;invite=invite->next)
		if (!strcasecmp(chptr->name, invite->channel))
			return; /* already exists */
	invite = (IRCD_INVITE *)MyMalloc(sizeof(IRCD_INVITE));
	invite->channel = (char *)MyStrdup(chptr->name);
	add_invite_to_list(&(cptr->c_invites_list), invite);
	if (cptr->c_invites_num == 10) /* remove oldest invite */
	{
		IRCD_INVITE	*i;
		for(i=cptr->c_invites_list;i->next;i=i->next)
			continue;
		del_invite_from_list(&(cptr->c_invites_list), i);
	}
	else
		cptr->c_invites_num++;
	return;
}

int _channel_canjoin(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr, char *key)
{
	IRCD_INVITE	*i = NULL;

	assert(IsLocal(pptr));

	if (!IsRegistered(pptr) && (chptr->modes & CHANNEL_MODES_REGGEDONLY))
		return ERR_REGYOURNICK;
	if (!IsAnOper(pptr) && (chptr->modes & CHANNEL_MODES_OPERONLY))
		return ERR_NOPRIVILEGES; 
	if (chptr->limit && (chptr->numusers >= chptr->limit))
		return ERR_CHANNELISFULL;
	if (chptr->key && (!key || mycmp(chptr->key, key)))
		return ERR_BADCHANNELKEY;
	if (chptr->modes & CHANNEL_MODES_INVITEONLY)
	{
		for(i=pptr->c_invites_list;i;i=i->next)
			if (!strcasecmp(i->channel, chptr->name))
				break;
		if (!i)
			return ERR_INVITEONLYCHAN;
	}
	if (chptr->bans)
	{
		IRCD_BAN	*ban = chptr->bans;
		char		banstr1[NICKLEN+USERLEN+HOSTLEN+6];
		char		banstr2[NICKLEN+USERLEN+HOSTLEN+6];
		char		banstr3[NICKLEN+USERLEN+HOSTLEN+6];
		char		*hostptr;
		int		tb	= 0;

		_make_nick_user_host(pptr->name, pptr->username, pptr->hostname,
						banstr1, &hostptr);
		strcpy(banstr2, banstr1);
		strcpy(banstr3, banstr1);

		*hostptr = '\0';
		strcpy(banstr2+(hostptr-banstr1), pptr->c_ipstring);

		/* banstr1 just has nick + user@ */
		/* banstr2 has n!u@IP */
		/* banstr3 has n!u@HOST */

		for(;ban;ban=ban->next)
		{
			if (ban->firstpart)
			{
				if (!cidr_match(ban->net, ban->mask, pptr->c_rem_ipaddr)
						&& !match(ban->firstpart, banstr1))
					tb = 1;
			}
			if (!match(ban->banstring, banstr2) ||
						!match(ban->banstring, banstr3))
					tb = 1;

			if(tb && chptr->excepts)
			{
				IRCD_EXCEPT	*except = chptr->excepts;
				for(;except;except=except->next)
				{
					if(except->firstpart)
					{
						if(!cidr_match(except->net, except->mask, pptr->c_rem_ipaddr) 
								&& !match(except->firstpart, banstr1))
							return 0; /* not banned */
					}
					if (!match(except->estring, banstr2) ||
							!match(except->estring, banstr3))
							return 0; /* not banned */	
				}

				/* if we've gotten through the whole loop, and none of the +e
				 * masks match.. your banned, kiddo */
				return ERR_BANNEDFROMCHAN; 
			}
			else
				return ERR_BANNEDFROMCHAN;
		}
	}
	if (i)
	{
		del_invite_from_list(&(pptr->c_invites_list), i);
		pptr->c_invites_num--;
		MyFree(i->channel);
		MyFree(i);
	}
	return 0;
}

IRCD_CHANNELLIST *_channel_ismember(IRCD_CHANNEL *chptr, IRCD_CLIENT *pptr)
{
	IRCD_CHANNELLIST	*chlptr;

	for(chlptr=pptr->c_channels;chlptr;chlptr=chlptr->next)
		if (chlptr->chptr == chptr)
			break;
	return chlptr;
}

int _channel_cansend(IRCD_CLIENT *pptr, IRCD_CHANNEL *chptr)
{
	IRCD_CHANNELLIST	*chlptr = NULL;

	if (chptr->modes & CHANNEL_MODES_MODERATED)
	{
		if ((chlptr = IsMember(chptr, pptr)) == NULL)
			return -1;
		if (!(chlptr->clptr->modes &
						(CHANNEL_MODES_CHANOP|CHANNEL_MODES_VOICE)))
			return -1;
	}
	if (!IsRegistered(pptr) && (chptr->modes & CHANNEL_MODES_REGGEDONLY))
	{
		if ((chlptr = IsMember(chptr, pptr)) == NULL)
			return 0;
		if (!(chlptr->clptr->modes & (CHANNEL_MODES_CHANOP|CHANNEL_MODES_VOICE)))
			return 0;

	}
	if (chptr->modes & CHANNEL_MODES_NOPRIVMSGS)
	{
		if (!chlptr && (IsMember(chptr, pptr) == NULL))
			return -2;
	}
	if (chptr->quiets)
	{
		/* _channel_canjoin() */
		IRCD_QUIET 	*quiet = chptr->quiets;
		char		qstr1[NICKLEN+USERLEN+HOSTLEN+6];
		char		qstr2[NICKLEN+USERLEN+HOSTLEN+6];
		char		qstr3[NICKLEN+USERLEN+HOSTLEN+6];
		char		*hostptr;

                if ((chlptr = IsMember(chptr, pptr)) != NULL) /* if they're in the room.... */
		{
			if (chlptr->clptr->modes & (CHANNEL_MODES_CHANOP|CHANNEL_MODES_VOICE)) 	/* .... and they're opped or voiced.... */
				return 0; /* then they may speak regardless of +q */
	        }

		_make_nick_user_host(pptr->name, pptr->username, pptr->hostname,
						qstr1, &hostptr);
		strcpy(qstr2, qstr1);
		strcpy(qstr3, qstr1);

		*hostptr = '\0';
		strcpy(qstr2+(hostptr-qstr1), pptr->c_ipstring);

		for(;quiet;quiet=quiet->next)
		{
			if (quiet->firstpart)
			{
				if (!cidr_match(quiet->net, quiet->mask, pptr->c_rem_ipaddr)
						&& !match(quiet->firstpart, qstr1))
					return -1; /* cant speak */
			}
			if (!match(quiet->qstring, qstr2) ||
						!match(quiet->qstring, qstr3))
					return -1; /* cant speak here either */
		}
	}

	return 0;
}

static void _channelname_clean(unsigned char *chname)
{
	for(;*chname;chname++)
	{
		if (*chname == '\007' || (*chname == ' ') ||
				(*chname == ',') ||
#if 0
				(*chname == ':') ||
#endif
				((*chname > 127) && (*chname <= 160)))
		{
			*chname = '\0';
			return;
		}
	}
}

/* m_list() */
int m_list(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*channels = ircd_nextarg(&args, 0);
	char			*channel;
	IRCD_CHANNEL	*chptr;
	int				sc;

	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LISTSTART, pptr->name));

	if (!channels || !*channels)
	{
		IRCD_CHANNELLIST	*chlptr;

		for(chptr=Channels;chptr;chptr=chptr->next)
		{
			if (!CanSSC(pptr) && !(chlptr = IsMember(chptr, pptr)) &&
						(chptr->modes & CHANNEL_MODES_SECRET))
				continue;
			sc = (chlptr || !((chptr->modes & (CHANNEL_MODES_SECRET|CHANNEL_MODES_PRIVATE)) && !CanSSC(pptr)));
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_LIST, pptr->name),
					sc ? chptr->name : "*", chptr->numusers,
					sc ? (chptr->topic ? chptr->topic : "[No topic set]") : "[Topic hidden for non-member]");
		}
	}
	else for(;(channel = ircd_next_subarg_comma(&channels)) != NULL;)
	{
		chptr = hash_find_channel(channel);
		if (chptr && 
			(((chptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET))) ||
					IsMember(chptr, pptr) || CanSSC(pptr)))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_LIST, pptr->name), channel,
				chptr->numusers, chptr->topic ? chptr->topic : "[No topic set]");
		}
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LISTEND, pptr->name));
	return 3;
}

int m_topic(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
#define MAX_TOPIC_CHANS	10
	char				*channels = ircd_nextarg(&args, 0);
	char				*topic = ircd_nextarg(&args, 0);
	char				*name;
	IRCD_CHANNEL		*chptr;
	IRCD_CHANNELLIST	*chlptr;
	char				*sendtos[MAX_TOPIC_CHANS+1];
	char				**st = NULL;
	char				**std;
	int					numtried;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!channels || !*channels)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "TOPIC");
		return 3;
	}
	if (topic && (strlen(topic) > TOPICLEN))
		topic[TOPICLEN] = '\0';

	if (IsLocal(pptr))
	{
		numtried = 0;
		st = sendtos;
		for(;(name = ircd_next_subarg_comma(&channels)) != NULL;)
		{
			if (!*name)
				continue;
			if (++numtried > MAX_TOPIC_CHANS)
			{
				send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Too many recipients", pptr->name);
				return 3;
			}
			*st = name;
			st++;
		}
		*st = NULL;
		st = sendtos;
	}

	numtried = 0;

	for(;;)
	{
		if (st)
		{
			name = *st;
			st++;
			if (name)/* check for already sent */
			{
				for(std=sendtos;*std != name;std++)
					if (!strcmp(*std, name))
						break;
				if (*std != name)
				{
					DEBUG(DEBUG_ALL, ("got dup %s for topic\n", name));
					continue;
				}
			}
		}
		else
			name = ircd_next_subarg_comma(&channels);
		if (!name)
			break;
		numtried++;
		chptr = hash_find_channel(name);
		if (!chptr || ((chlptr = IsMember(chptr, pptr)) == NULL))
		{
			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
					name);
			continue;
		}
		if (!topic)
		{
			if (!chptr->topic)
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_NOTOPIC, pptr->name), chptr->name);
			else
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TOPIC, pptr->name),
					chptr->name, chptr->topic);
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TOPICWHOTIME, pptr->name),
					chptr->name, chptr->topic_who, chptr->topic_time);
			}
		}
		else
		{
			IRCD_CLIENTLIST	*clptr = chlptr->clptr;

			if (!IsServer(lptr) && (chptr->modes & CHANNEL_MODES_TOPICLIMIT))
			{
				if (!(clptr->modes & CHANNEL_MODES_CHANOP))
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					continue;
				}
			}
			if (chptr->topic)
			{
				MyFree(chptr->topic);
				MyFree(chptr->topic_who);
				chptr->topic = NULL;
				chptr->topic_who = NULL;
			}
			if (*topic)
			{
				chptr->topic = MyStrdup(topic);
				chptr->topic_who = MyStrdup(pptr->name);
				chptr->topic_time = Now;
			}
			send_to_channel_butone(IsServer(lptr) ? lptr : NULL,
					CHANNEL_SEND_SERVERS|CHANNEL_SEND_CLIENTS,
					chptr, pptr, "TOPIC %s :%s", chptr->name, topic);
		}
	}
	numtried /= 5;
	return numtried > 3 ? numtried : 3;
}

void _channel_invite_del(IRCD_CHANNEL *chptr, IRCD_CLIENT *cptr)
{
	IRCD_INVITE	*inext;
	IRCD_INVITE	*invite = cptr->c_invites_list;

	for(;invite;invite=inext)
	{
		inext = invite->next;
		if (!chptr || !strcasecmp(chptr->name, invite->channel))
		{
			del_invite_from_list(&(cptr->c_invites_list), invite);
			cptr->c_invites_num--;
			MyFree(invite->channel);
			MyFree(invite);
		}
	}
}

int m_invite(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*user = ircd_nextarg(&args, 0);
	char			*channel = ircd_nextarg(&args, 0);
	int				doit = 0;
	IRCD_CLIENT		*acptr;
	IRCD_CHANNEL	*chptr;
	IRCD_CHANNELLIST	*chlptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!channel || !*channel)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "INVITE");
		return 3;
	}

	if ((!(acptr = hash_find_client(user))) || !IsClient(acptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHNICK,
				pptr->name), user);
		return 3;
	}

	if (!(chptr = hash_find_channel(channel)))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL,
			pptr->name), channel);
#if 0
		send_prefix_to_client(pptr, acptr, "INVITE %s :%s",
				user, channel);
#endif
		return 3;
	}
	if ((chlptr = IsMember(chptr, pptr)) == NULL)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL,
			pptr->name), channel);
		return 3;
	}
	if (IsMember(chptr, acptr) != NULL)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_USERONCHANNEL,
			pptr->name), user, channel);
		return 3;
	}

	if (chptr->modes & CHANNEL_MODES_INVITEONLY)
	{
		IRCD_CLIENTLIST	*clptr = chlptr->clptr;

		if (!(clptr->modes & CHANNEL_MODES_CHANOP))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CHANOPRIVSNEEDED,
				pptr->name), chptr->name);
			return 3;
		}
		doit = 1;
	}

	if (IsLocal(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_INVITING,
				pptr->name), acptr->name, chptr->name);
		if (acptr->c_away)
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_AWAY,
					pptr->name), acptr->name, acptr->c_away);
	}

	if (IsLocal(acptr) && doit)
		_channel_invite_add(chptr, acptr);
	send_prefix_to_client(pptr, acptr, "INVITE %s :%s",
			acptr->name, chptr->name);
	return 3;
}

int m_names(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*channel;
	IRCD_CHANNEL	*acptr;
	char			buf[512];
	char			*dptr;
	char			*saveptr;
	int				ismem;
	int				flag;
	int				mylen;
	IRCD_CLIENTLIST	*clptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	else if (IsServer(lptr))
		return 3; /* drop it */

	channel = ircd_nextarg(&args, 0);

	if (!channel || !*channel)
	{
		if (IsLocal(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "NAMES");
		return 3;
	}

	acptr = hash_find_channel(channel);
	if (!acptr)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHCHANNEL,
				pptr->name), channel);
		return 3;
	}

	ismem = (IsMember(acptr, pptr) != NULL);
	if (!ismem && !CanSSC(pptr) && 
			(acptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET)))
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_ENDOFNAMES, pptr->name), channel);
		return 3;
	}
	mylen = strlen(meptr->name);
	dptr = buf;
	*dptr++ = '*';
	*dptr++ = ' ';
	dptr += strlen(strcpy(dptr, acptr->name));
	*dptr++ = ' ';
	*dptr++ = ':';
	saveptr = dptr;
	if (acptr->modes & CHANNEL_MODES_SECRET)
		*buf = '@';
	else if (!(acptr->modes & CHANNEL_MODES_PRIVATE))
		*buf = '=';
	flag = 1;
	for(clptr=acptr->members;clptr;clptr=clptr->next)
	{
		if (!ismem && !CanSSC(pptr) && (clptr->cptr->c_umodes & CLIENT_UMODES_IMODE))
			continue;
		if (clptr->modes & CHANNEL_MODES_CHANOP)
			*dptr++ = '@';
		else if (clptr->modes & CHANNEL_MODES_VOICE)
			*dptr++ = '+';
		dptr += strlen(strcpy(dptr, clptr->cptr->name));
		*dptr++ = ' ';
		flag = 1;
		if (mylen + (dptr - buf) > 475)
		{
			*(dptr-1) = '\0';
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_NAMREPLY,
					pptr->name), buf);
			dptr = saveptr;
			flag = 0;
		}
	}
	if (flag)
	{
		if (*(dptr-1) == ' ')
			*(dptr-1) = '\0';
		else
			*dptr = '\0';
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_NAMREPLY,
				pptr->name), buf);
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFNAMES, pptr->name),
		channel);

	return 3;
}

void _user_del_channels(IRCD_CLIENT *pptr)
{
	IRCD_CHANNELLIST		*nextch;

	for(;pptr->c_channels;pptr->c_channels=nextch)
	{
		nextch = pptr->c_channels->next;
		_channel_del_user(pptr->c_channels->chptr, pptr, pptr->c_channels);
		MyFree(pptr->c_channels);
	}
}

int m_join(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*channel = ircd_nextarg(&args, 0);
	char	*key = ircd_nextarg(&args, 0);
	char	*ptr;
	char	*k;
	char	*name;
	int		chlen;
	int		flag;
	int		i;
	int		isnewc = 0;
	IRCD_CHANNEL	*chptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	else if (IsServer(pptr))
		return 0;

	if (!channel || !*channel)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
			"JOIN");
		return 3;
	}

	for(ptr=channel;;)
	{
		if ((*ptr == ',') && (*(ptr+1) == '0') && !atoi(ptr+1))
		{
			channel = ptr+1;
			ptr++;
		}
		if (!*ptr++)
			break;
	}

	k = ircd_next_subarg_comma(&key);
	for(;(name=ircd_next_subarg_comma(&channel)) != NULL;
							k=ircd_next_subarg_comma(&key))
	{
		_channelname_clean(name);

		if (!*name)
			continue;

		if (*name == '0' && !atoi(name))
		{
			IRCD_CHANNELLIST		*nextch;

			for(;pptr->c_channels;pptr->c_channels=nextch)
			{
				nextch = pptr->c_channels->next;
				send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						pptr->c_channels->chptr,
						pptr, "PART %s", pptr->c_channels->chptr->name);
				_channel_del_user(pptr->c_channels->chptr, pptr,
						pptr->c_channels);
				MyFree(pptr->c_channels);
			}
			send_to_servers(lptr, ":%s JOIN 0", pptr->name);
			continue;
		}
		if (IsServer(lptr) && (*name == '&'))
			continue;
		if ((chlen=strlen(name)) > CHANNELLEN)
		{
			name[CHANNELLEN] = '\0';
			chlen = CHANNELLEN;
		}
		if (!IsChannel(name))
		{
			if (MyClient(pptr))
				send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHCHANNEL,
						pptr->name), name);
			continue;
		}
		if (IsLocal(pptr))
		{
			int		maxchannels = pptr->c_class->maxchannels;

			if (maxchannels < 0)
				maxchannels = (int)vars_value_get_bynum(IRCD_VAR_DEFAULT_MAXCHANNELS);

			if (pptr->c_numchannels >= maxchannels)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_TOOMANYCHANNELS, pptr->name), name);
				continue;
			}

			if (!(pptr->c_class->flags & CLASS_FLAGS_BYPASSCHANNELDENY))
			{
				STRINGDENY		*sd;

				sd = stringdeny_match(&(_ircd_conf.channel_denies),
							name, Now);
				if (sd)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_UNAVAILRESOURCE, pptr->name),
						name);
					continue;
				}
			}
		}
		flag = 0;
		chptr = hash_find_channel(name);
		if (!chptr)
		{
			/* new chan */

			if (!IsLocal(pptr))
			{
				send_to_opers("User %s on %s remotely JOINing new channel (non-TS behavior)",
						pptr->name, pptr->servername);
			}
			else
			{
				int chanopsonsplit = (int)vars_value_get_bynum(IRCD_VAR_CHANOPSONSPLIT);

				if (IsAnOper(pptr) &&
					vars_value_get_bynum(IRCD_VAR_OPERSCANJOINONSPLIT))
				{
					if (chanopsonsplit && (*name != '&') &&
						((Globalinfo->g_servers -
							Globalinfo->g_serversbursting) <
												chanopsonsplit))
					{
						send_to_opers("%s is creating channel %s during split",
							pptr->name, name);
					}
					flag |= CHANNEL_MODES_CHANOP;
				}
				else if (!chanopsonsplit || (*name == '&') ||
						((Globalinfo->g_servers -
							Globalinfo->g_serversbursting) >=
												chanopsonsplit))
				{
					flag |= CHANNEL_MODES_CHANOP;
				}
			}
			chptr = (IRCD_CHANNEL *)MyMalloc(sizeof(IRCD_CHANNEL)+chlen);
			strcpy(chptr->name, name);
			hash_add_channel(chptr);
			if (++Globalinfo->channels > Globalinfo->maxchannels)
			{
				Globalinfo->maxchannels = Globalinfo->channels;
				Globalinfo->maxchannelstime = Now;
			}
			chptr->creation = Now;

			if(vars_value_get_bynum(IRCD_VAR_SETDEFAULTCMODES))
				isnewc	= 1;
		}

		if (IsMember(chptr, pptr))
		{
			if (IsLocal(pptr))
			{
				send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_USERONCHANNEL,
					pptr->name), pptr->name, name);
			}
			continue;
		}

		if (IsClient(lptr) && (i = _channel_canjoin(pptr, chptr, k)))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(i, pptr->name),
					name);
			continue;
		}

		_channel_add_user(chptr, pptr, flag);

		if (*name != '&')
		{
			if (IsClient(lptr) && (flag & CHANNEL_MODES_CHANOP))
			{
				send_to_servers(NULL, ":%s SJOIN %u %s +%s :@%s",
					meptr->name, chptr->creation, name, (isnewc) ? "nt" : "", pptr->name);
			}
			else if (IsClient(lptr))
				send_to_servers(NULL, ":%s SJOIN %u %s +%s :%s",
					meptr->name, chptr->creation, name, (isnewc) ? "nt" : "", pptr->name);
			else
			{
				/* this won't happen with all TS servers */
				send_to_servers(lptr, ":%s JOIN :%s", pptr->name, name);
				if(isnewc)
					send_to_servers(lptr, ":%s MODE %s +nt", meptr->name, name);
			}
		}

		send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "JOIN :%s", name);

		if (MyClient(pptr))
		{
			if ((chptr->numusers == 1) && !(flag & CHANNEL_MODES_CHANOP))
				send_prefix_to_client(meptr, pptr,
						"NOTICE %s :*** Channel ops on %s are unavailable during split ***",
						pptr->name, name);
#if 0
			_channel_invite_del(chptr, pptr);
/* already done in _canjoin() */
#endif
			if (chptr->topic)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TOPIC, pptr->name),
					name, chptr->topic);
				if (chptr->topic_who)
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TOPICWHOTIME, pptr->name),
						name, chptr->topic_who, chptr->topic_time);
			}

			m_names(lptr, pptr, prefix, name);

			if(isnewc)
			{
				chptr->modes |=         CHANNEL_MODES_NOPRIVMSGS;
				chptr->modes |=		CHANNEL_MODES_TOPICLIMIT;
				send_to_client(pptr, ":%s MODE %s +nt", vars_value_get_bynum(IRCD_VAR_NETWORK_HOST), chptr->name);
			}
		}
	}

	return 3;
}

int m_part(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_CHANNEL	*chptr;
	char			*channels = ircd_nextarg(&args, 0);
	char			*name;
	IRCD_CHANNELLIST	*chlptr;
	char			buf[512];
	char			*dptr = buf;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	else if (IsServer(pptr))
		return 0;

	if (!channels || !*channels)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "PART");
		return 3;
	}

	for(;(name = ircd_next_subarg_comma(&channels)) != NULL;)
	{
		chptr = hash_find_channel(name);
		if (!chptr)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHCHANNEL,
					pptr->name), name);
			continue;
		}
		if ((chlptr = IsMember(chptr, pptr)) == NULL)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL,
				pptr->name), name);
			continue;
		}

		if (*name != '&')
		{
			if (dptr != buf)
				*dptr++ = ',';
			dptr += strlen(strcpy(dptr, name));
		}

		send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "PART %s", name);
		_channel_del_user(chptr, pptr, chlptr);
		del_channellist_from_list(&(pptr->c_channels), chlptr);
		MyFree(chlptr);
	}
	if (dptr != buf)
	{
		*dptr = '\0';
		send_to_servers(lptr, ":%s PART %s", pptr->name, buf);
	}
	return 3;
}

int m_omode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if(MyClient(pptr))
	{
		if(!IsAnOper(pptr))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
			return 3;
		}
		if(!(pptr->c_oline_flags & OLINE_FLAGS_CANOMODE))
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :You are not allowed to use OMODE.", pptr->name);
			return 3;
		}
	}

	return do_mode(lptr, pptr, prefix, args, 1);
}
int m_mode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return do_mode(lptr, pptr, prefix, args, 0);
}

int do_mode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int omode)
{
	IRCD_CLIENT		*tptr;
	IRCD_CHANNEL	*chptr;
	IRCD_CHANNELLIST	*chlptr;
	IRCD_CHANNELLIST	*chlptr2;
	char			*channel = ircd_nextarg(&args, 0);
	char			*modes = ircd_nextarg(&args, 0);
	char			*modeargs = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	char			*ptr;
	char			plusminus = '+';
	char			newban[NICKLEN+USERLEN+HOSTLEN+6];
	char			newexcept[NICKLEN+USERLEN+HOSTLEN+6];
	char			newquiet[NICKLEN+USERLEN+HOSTLEN+6];
	IRCD_MODE		*opmodes = NULL;
	IRCD_MODE		*lastopmode = NULL;
	IRCD_MODE		*banmodes = NULL;
	IRCD_MODE		*lastbanmode = NULL;
	IRCD_MODE		*exceptmodes	= NULL;
	IRCD_MODE		*lastexceptmode = NULL;
	IRCD_MODE		*quietmodes = NULL;
	IRCD_MODE		*lastquietmode = NULL;
	IRCD_MODE		*opmodeptr;
	char			parabuf[512];
	char			modebuf[512];
	char			modeargbuf[512];
	char			*modeptr = modebuf;
	char			*modeargptr = modeargbuf;
	char			*newkey = NULL;
	int				limitchange = 0;
	int				limit = 0;
	int				tlimit;
	int				keychange = 0;
	int				opcnt = 0;
	int				*i;
	int				newmodes;
	int				oldmodes;
	char			*sptr;
	char			*target;
	int				isok;
	int				ischop;
	int				isdeop;
	int				got_ovk = 0;
	IRCD_BAN		*ban;
	IRCD_EXCEPT		*except;
	IRCD_QUIET		*quiet;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	if (!channel || !*channel)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "MODE");
		return 3;
	}
	chptr = hash_find_channel(channel);
	if (!chptr)
#ifdef HMC
		return m_umode(lptr, pptr, channel, prefix, modes, modeargs);
#else
		return m_umode(lptr, pptr, channel, prefix, modes);
#endif

	if (!modes || !*modes)
	{
		_build_channel_modes(chptr, pptr, modebuf, parabuf);
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_CHANNELMODEIS, pptr->name), chptr->name,
			modebuf, parabuf);
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_CREATIONTIME, pptr->name), chptr->name,
			chptr->creation);
		return 3;
	}

	chlptr = IsMember(chptr, pptr);

	ischop = IsServer(pptr) || (omode) || 
				(chlptr && (chlptr->clptr->modes & CHANNEL_MODES_CHANOP));
	isdeop = !ischop && !IsServer(pptr) &&
				(chlptr && (chlptr->clptr->modes & CHANNEL_MODES_DEOPPED));
	isok = ischop || (!isdeop && IsServer(lptr) && chptr->creation) || !MyClient(pptr);

	oldmodes = newmodes = chptr->modes;

	for(ptr=modes;ptr && *ptr;)
	{
		switch(*ptr)
		{
			case '+':
			case '-':
				plusminus = *ptr;
				break;
			case 'o':
			case 'v':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
#if 0
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
						"MODE-ov");
#endif
					break;
				}
				if (IsClient(lptr) && !omode)
				{
					if (!chlptr)
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
							chptr->name);
						break;
					}
					if (!ischop)
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
							chptr->name);
						break;
					}

					if (opcnt++ >= MAXMODEPARAMS)
						break;
				}
				tptr = hash_find_chasing(target);
				if (!tptr)
					break;
				if (!IsMember(chptr, tptr))
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_USERNOTINCHANNEL, pptr->name),
						tptr->name, chptr->name);
					break;
				}
				if (IsService(tptr) && !omode && (plusminus == '-') && (*ptr == 'o'))
				{
						send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_ITSASERVICE, pptr->name),
								tptr->name, chptr->name);
						break;
				}
				if ((tptr == lptr) && (plusminus == '+') && (*ptr == 'o') && !omode && MyClient(tptr))
				{
					break;
				}
				if (!got_ovk)
					got_ovk = *ptr;
				else if (IsClient(lptr) && (got_ovk != *ptr))
					break;
				/* keep track of 'tptr' and what mode change */
				opmodeptr = _ircdmode_create(&opmodes, &lastopmode);
				opmodeptr->cptr = tptr;
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				break;
			case 'k':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
					if (MyClient(pptr))
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
							"MODE-k");
					break;
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (keychange)
					break;
				else
				{
					register u_char		*s;

					if (IsClient(lptr))
					{
						register u_char		*d;
						register u_char		ch;

						for(d=s=(u_char *)target;(ch=*s);s++)
						{
							ch &= 0x7f;
							if ((ch != ':') && (ch > 0x20))
							{
								*d++ = ch;
							}
						}
						*d = '\0';
					}
					else
					{
						for(s=(u_char *)target;*s;s++)
						{
							*s &= 0x7f;
							if ((*s == ' ') || (*s == '\t') || (*s == '\n'))
							{
								*s = '\0';
								break;
							}
						}
					}
				}
				if (!*target || ((IsClient(lptr) &&
									(opcnt++ >= MAXMODEPARAMS))))
					break;
				if (!got_ovk)
					got_ovk = *ptr;
				else if (IsClient(lptr) && (*ptr != got_ovk))
					break;
				keychange = 1;
				newkey = chptr->key;
				if (plusminus == '+')
				{
					if (chptr->key && IsClient(lptr))
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_KEYSET, pptr->name),
							chptr->name);
					else if (!chptr->key || IsServer(lptr))
						newkey = target;
				}
				else
				{
					if (IsServer(lptr) ||
							(chptr->key && !mycmp(chptr->key, target)))
					{
						newkey = NULL;
					}
				}
				break;
			case 'b':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
					/* return banlist */
					if(HideCBans && chptr)
					{
						if(!IsAnOper(pptr) && !IsMember(chptr, pptr))
						{
							send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name), chptr->name);
							break;
						}
						if(!IsAnOper(pptr) && !(chlptr->clptr->modes & CHANNEL_MODES_CHANOP))
						{
							send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name), chptr->name);
							break;
						}

                                        }

					for(ban=chptr->bans;ban;ban=ban->next)
					{
						send_prefix_to_client(meptr, pptr, 
							IRCD_REPLY(RPL_BANLIST, pptr->name),
							chptr->name,
							ban->banstring,
							ban->who,
							ban->bantime);
					}
					send_prefix_to_client(meptr, pptr, 
						IRCD_REPLY(RPL_ENDOFBANLIST, pptr->name),
						chptr->name);
					break;
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (IsClient(lptr) && opcnt++ >= MAXMODEPARAMS)
					break;
				opmodeptr = _ircdmode_create(&banmodes, &lastbanmode);
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				opmodeptr->banstring = target;
				break;
			case 'e':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
                                        if(HideCExcepts && chptr)
                                        {
                                                if(!IsAnOper(pptr) && !IsMember(chptr, pptr))
                                                {
                                                        send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name), chptr->name);
                                                        break;
                                                }
                                                if(!IsAnOper(pptr) && !(chlptr->clptr->modes & CHANNEL_MODES_CHANOP))
                                                {
                                                        send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name), chptr->name);
                                                        break;
                                                }
                                                
                                        }

					for(except=chptr->excepts;except;except=except->next)
					{
						send_prefix_to_client(meptr, pptr, 
							IRCD_REPLY(RPL_EXCEPTLIST, pptr->name),
							chptr->name,
							except->estring,
							except->who,
							except->etime);
					}
					send_prefix_to_client(meptr, pptr, 
						IRCD_REPLY(RPL_ENDOFEXCEPTLIST, pptr->name),
						chptr->name);
					break;
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (IsClient(lptr) && opcnt++ >= MAXMODEPARAMS)
					break;
				opmodeptr = _ircdmode_create(&exceptmodes, &lastexceptmode);
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				opmodeptr->exceptstring = target; /* could probably share this with 'banstring', might do so later */
				break;
			case 'q':
				target = ircd_next_subarg_space(&modeargs);
				if (!target || !*target)
				{
                                        if(HideCQuiets && chptr)
                                        {
                                                if(!IsAnOper(pptr) && !IsMember(chptr, pptr))
                                                {
                                                        send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name), chptr->name);
                                                        break;
                                                }
                                                if(!IsAnOper(pptr) && !(chlptr->clptr->modes & CHANNEL_MODES_CHANOP))
                                                {
                                                        send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name), chptr->name);
                                                        break;
                                                }
                                                
                                        }
					for(quiet=chptr->quiets;quiet;quiet=quiet->next)
					{
						send_prefix_to_client(meptr, pptr, 
							IRCD_REPLY(RPL_QUIETLIST, pptr->name),
							chptr->name,
							quiet->qstring,
							quiet->who,
							quiet->qtime);
					}
					send_prefix_to_client(meptr, pptr, 
						IRCD_REPLY(RPL_ENDOFQUIETLIST, pptr->name),
						chptr->name);
					break;
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (IsClient(lptr) && opcnt++ >= MAXMODEPARAMS)
					break;
				opmodeptr = _ircdmode_create(&quietmodes, &lastquietmode);
				opmodeptr->mode = *ptr;
				opmodeptr->plusminus = plusminus;
				opmodeptr->quietstring = target;
				break;
			case 'l':
				tlimit = 0;
				if (plusminus == '+')
				{
					target = ircd_next_subarg_space(&modeargs);
					if (!target || !*target || !(tlimit = atoi(target)))
					{
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name),
							"MODE-l");
						break;
					}
				}
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				if (limitchange++)
					break;
				limit = tlimit;
				break;
			case 'O':
				if(!IsAnOper(pptr))
				{
					send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
					break;
				}
			case 'i':
			case 's':
			case 'p':
			case 't':
			case 'n':
			case 'm':
			case 'R':
				if (!IsServer(pptr) && !chlptr && !omode)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
						chptr->name);
					break;
				}
				if (!ischop)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
						chptr->name);
					break;
				}
				for(i=_channel_modes;*i;i+=2)
				{
					if (*(i+1) != *ptr)
						continue;
					if (plusminus == '+')
						newmodes |= *i;
					else
						newmodes &= ~(*i);
				}
				break;
			default:
				if (IsClient(pptr))
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(ERR_UNKNOWNMODE, pptr->name), *ptr);
				break;
		}
		if (!*++ptr)
			ptr = ircd_next_subarg_space(&modeargs);
	}

    /* if it's a non-deopped non-op on a TS channel from a TS
    ** server, allow only ops and deops and ignore everything
    ** else (to avoid having desynched ops) -orabidoo
    */
	if (!ischop && isok)
	{
		newmodes = oldmodes;
		limitchange = 0;
		keychange = 0;
		_ircdmode_free(&banmodes);
		_ircdmode_free(&exceptmodes);
		_ircdmode_free(&quietmodes);
	}

	plusminus = '\0';

	if (newmodes != oldmodes)
	{
		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & newmodes) && !(*i & oldmodes))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
				chptr->modes |= *i;
			}
		}

		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & oldmodes) && !(*i & newmodes))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
				chptr->modes &= ~(*i);
			}
		}
	}

	if (limitchange)
	{
		if (limit && (chptr->limit != limit))
		{
			char	buf[20];

			if (plusminus != '+')
			{
				plusminus = '+';
				*modeptr++ = plusminus;
			}
			*modeptr++ = 'l';
			sprintf(buf, "%-15d", limit);
			if (modeargptr != modeargbuf)
				*modeargptr++ = ' ';
			for(sptr=buf;*sptr && *sptr != ' ';)
					*modeargptr++ = *sptr++;
			chptr->limit = limit;
		}
		if (!limit && chptr->limit)
		{
			if (plusminus != '-')
			{
				plusminus = '-';
				*modeptr++ = plusminus;
			}
			*modeptr++ = 'l';
			chptr->limit = 0;
		}
	}

	/* build +ov */

	for(opmodeptr=opmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		if ((modeargptr-modeargbuf) +
				strlen(opmodeptr->cptr->name) + 2 >= sizeof(modeargbuf))
			break;
		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		chlptr2 = IsMember(chptr, opmodeptr->cptr);
		if (opmodeptr->mode == 'o')
		{
			if (IsServer(pptr))
			{
				send_to_opers("Server %s doing mode %co and removing TS on %s",
					pptr->name, plusminus, chptr->name);
				chptr->creation = 0;
			}

			if (isdeop && (plusminus == '+'))
			{
				if (chlptr2 && !(chlptr2->clptr->modes & CHANNEL_MODES_CHANOP))
					chlptr2->clptr->modes |= CHANNEL_MODES_DEOPPED;
			}
		}
		if (isok)
		{
			isok = _channel_change_opflag(chptr, chlptr2->clptr,
							opmodeptr, IsClient(lptr) /* , pptr->c_oline_forcing */ );
			if (isok || IsServer(lptr))
			{
				*modeptr++ = opmodeptr->mode;
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=opmodeptr->cptr->name;*sptr;)
					*modeargptr++ = *sptr++;
			}
		}
	}
	_ircdmode_free(&opmodes);

	/* +k */

	if (keychange)
	{
		if (chptr->key && !newkey)
		{
			if (!((modeargptr-modeargbuf) +
				strlen(chptr->key) + 2 >= sizeof(modeargbuf)))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = 'k';
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=chptr->key;*sptr;)
					*modeargptr++ = *sptr++;
				MyFree(chptr->key);
				chptr->key = NULL;
			}
		}
		else if (newkey && newkey != chptr->key)
		{
			if (!((modeargptr-modeargbuf) +
				strlen(newkey) + 2 >= sizeof(modeargbuf)))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = 'k';
				if (modeargptr != modeargbuf)
					*modeargptr++ = ' ';
				for(sptr=newkey;*sptr;)
					*modeargptr++ = *sptr++;
				if (chptr->key)
					MyFree(chptr->key);
				chptr->key = MyStrdup(newkey);
			}
		}
	}

	/* bans */
	for(opmodeptr=banmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		char			*user, *host;
		int				ret;
		ircd_uint32_t	net;
		ircd_uint32_t	mask;
		char			*hostptr;

		if ((user = strchr(opmodeptr->banstring, '!')) != NULL)
			*user++ = '\0';
		if ((host = strrchr(user?user:opmodeptr->banstring, '@')) != NULL)
			*host++ = '\0';
		sptr = _make_nick_user_host(opmodeptr->banstring,
								user, host, newban, &hostptr);
		ret = cidr_ip2netmask(hostptr, &net, &mask);
		if (!ret)
			net = -1;

		if ((modeargptr-modeargbuf) +
				strlen(sptr) + 2 >= sizeof(modeargbuf))
			break;

		if (((opmodeptr->plusminus == '+') &&
					!_channel_ban_add(chptr, pptr, sptr, hostptr, net, mask)) ||
				((opmodeptr->plusminus == '-') &&
					!_channel_ban_del(chptr, sptr)))
			continue;

		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		*modeptr++ = opmodeptr->mode;
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(;*sptr;)
			*modeargptr++ = *sptr++;
	}
	_ircdmode_free(&banmodes);

	/* excepts */
	for(opmodeptr=exceptmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		char			*user, *host;
		int				ret;
		ircd_uint32_t	net;
		ircd_uint32_t	mask;
		char			*hostptr;

		if ((user = strchr(opmodeptr->exceptstring, '!')) != NULL)
			*user++ = '\0';
		if ((host = strrchr(user?user:opmodeptr->exceptstring, '@')) != NULL)
			*host++ = '\0';
		sptr = _make_nick_user_host(opmodeptr->exceptstring,
								user, host, newexcept, &hostptr);
		ret = cidr_ip2netmask(hostptr, &net, &mask);
		if (!ret)
			net = -1;

		if ((modeargptr-modeargbuf) +
				strlen(sptr) + 2 >= sizeof(modeargbuf))
			break;

		if (((opmodeptr->plusminus == '+') &&
					!_channel_except_add(chptr, pptr, sptr, hostptr, net, mask)) ||
				((opmodeptr->plusminus == '-') &&
					!_channel_except_del(chptr, sptr)))
			continue;

		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		*modeptr++ = opmodeptr->mode;
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(;*sptr;)
			*modeargptr++ = *sptr++;
	}
	_ircdmode_free(&exceptmodes);

	/* quiets */
	for(opmodeptr=quietmodes;opmodeptr;opmodeptr=opmodeptr->next)
	{
		char			*user, *host;
		int				ret;
		ircd_uint32_t	net;
		ircd_uint32_t	mask;
		char			*hostptr;

		if ((user = strchr(opmodeptr->quietstring, '!')) != NULL)
			*user++ = '\0';
		if ((host = strrchr(user?user:opmodeptr->quietstring, '@')) != NULL)
			*host++ = '\0';
		sptr = _make_nick_user_host(opmodeptr->quietstring,
								user, host, newquiet, &hostptr);
		ret = cidr_ip2netmask(hostptr, &net, &mask);
		if (!ret)
			net = -1;

		if ((modeargptr-modeargbuf) +
				strlen(sptr) + 2 >= sizeof(modeargbuf))
			break;

		if (((opmodeptr->plusminus == '+') &&
					!_channel_quiet_add(chptr, pptr, sptr, hostptr, net, mask)) ||
				((opmodeptr->plusminus == '-') &&
					!_channel_quiet_del(chptr, sptr)))
			continue;

		if (plusminus != opmodeptr->plusminus)
		{
			plusminus = opmodeptr->plusminus;
			*modeptr++ = plusminus;
		}
		*modeptr++ = opmodeptr->mode;
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(;*sptr;)
			*modeargptr++ = *sptr++;
	}
	_ircdmode_free(&quietmodes);

	*modeptr = '\0';
	*modeargptr = '\0';

	if (*modebuf && isok)
	{
		if (IsClient(pptr))
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);
		else
			send_to_channel_butone2(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, prefix, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);

		if (*chptr->name != '&')
		{
			if(!omode)
				send_to_servers(IsServer(lptr) ? lptr : NULL, ":%s MODE %s %s %s",
					prefix, chptr->name, modebuf, modeargbuf);
			else
			{
				if(MyClient(pptr))
				{
					send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr, 
						"%s[@%s] is using OMODE in %s", 
							pptr->name, meptr->name, chptr->name);
					send_to_servers(NULL, ":%s WALLOPS :%s[@%s] is using OMODE in %s",
								meptr->name, pptr->name, meptr->name, chptr->name);
				}
				send_to_servers(lptr, ":%s MODE %s %s %s",
							prefix, chptr->name, modebuf, modeargbuf);
			}
		} /* if() '&' */
	}

	return 3;
}

int m_okick(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if(MyClient(pptr))
	{
		if(!IsAnOper(pptr))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
			return 3;
		}
		if(!(pptr->c_oline_flags & OLINE_FLAGS_CANOKICK))
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :You are not allowed to use OKICK.", pptr->name);
			return 3;
		}
	}

	/* okay, they're good, lets go.. */

	return do_kick(lptr, pptr, prefix, args, 1);
}

int m_kick(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return do_kick(lptr, pptr, prefix, args, 0);
}

int do_kick(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int okick)
{
	char				*channels = ircd_nextarg(&args, 0);
	char				*nickskeep = ircd_nextarg(&args, 0);
	char				*comment = ircd_nextarg(&args, 0);
	char				*channel;
	char				*nick;
	IRCD_CHANNEL		*chptr;
	char				buf[512];
	char				buf2[512];
	char				*dptr = buf;
	char				*dptr2 = buf2;
	IRCD_CLIENT			*acptr;
	char				nicksbuf[512];
	char				*nicks = nickskeep;
	char				*tptr;
	IRCD_CHANNELLIST	*chlptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!nicks || !*nicks)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "KICK");
		return 3;
	}

	if (!comment)
		comment = pptr->name;

	if (IsServer(pptr))
		send_to_opers("KICK from %s for %s %s", pptr->name,
					channels, nicks);

	for(;;)
	{
		if (!IsServer(lptr))
		{
			if ((channel = channels) == NULL)
				break;
			channels = NULL;
			tptr = strchr(channel, ',');
			if (tptr)
				*tptr = '\0';
		}
		else if ((channel = ircd_next_subarg_comma(&channels)) == NULL)
			break;
		chptr = hash_find_channel(channel);
		if (!chptr)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHCHANNEL,
					pptr->name), channel);
			continue;
		}
		if (!IsServer(pptr) && !okick && MyClient(pptr))
		{
			chlptr = IsMember(chptr, pptr);
			if (!chlptr)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOTONCHANNEL, pptr->name),
					chptr->name);
				continue;
			}
			else if (!(chlptr->clptr->modes & CHANNEL_MODES_CHANOP))
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_CHANOPRIVSNEEDED, pptr->name),
					chptr->name);
				continue;
			}
		}

		if (*channel != '&')
		{
			if (dptr != buf)
				*dptr++ = ',';
			else
				dptr += strlen(strcpy(dptr, channel));
		}

		support_strncpy(nicksbuf, nickskeep, sizeof(nicksbuf));
		nicks = nicksbuf;

		for(;;)
		{
			if (!IsServer(lptr))
			{
				if ((nick = nicks) == NULL)
					break;
				nicks = NULL;
				tptr = strchr(nick, ',');
				if (tptr)
					*tptr = '\0';
			}
			else if ((nick = ircd_next_subarg_comma(&nicks)) == NULL)
				break;

			if (!(acptr = hash_find_chasing(nick)))
				continue;
			if (IsService(acptr) && !okick && MyClient(pptr))
			{
				send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_ITSASERVICE, pptr->name), acptr->name, chptr->name);
				continue;
			}
			if ((chlptr = IsMember(chptr, acptr)) != NULL || okick || !MyClient(pptr))
			{
				if(okick && (acptr == pptr) && ((chlptr = IsMember(chptr, acptr)) == NULL))
					continue;
				if (*channel != '&')
				{
					if (dptr2 != buf2)
						*dptr2++ = ',';
					else
						dptr2 += strlen(strcpy(dptr2, acptr->name));
				}
				send_to_channel_butone(NULL,
					CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
					chptr, pptr, "KICK %s %s :%s", channel,
						acptr->name, comment);
				_channel_del_user(chptr, acptr, chlptr);
				del_channellist_from_list(&(acptr->c_channels), chlptr);
				MyFree(chlptr);
			}
			else
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_USERNOTINCHANNEL, pptr->name),
					nick, channel);
		}
	}

	if ((dptr != buf) && (dptr2 != buf2))
	{
		*dptr = '\0';
		*dptr2 = '\0';
		if(!okick)
			send_to_servers(lptr, ":%s KICK %s %s :%s",
				pptr->name, buf, buf2, comment);
		else
		{
			if(MyClient(pptr))
			{
				send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, meptr, 
						"%s[@%s] is using OKICK in %s", 
							pptr->name, meptr->name, chptr->name, buf2, comment);
				send_to_servers(lptr, ":%s WALLOPS :%s[@%s] is using OKICK in %s",
							meptr->name, pptr->name, meptr->name, chptr->name, buf2, comment);
			}
			send_to_servers(lptr, ":%s KICK %s %s :%s",
					pptr->name, buf, buf2, comment);
		}
	}

	return 3;
}

/*
** m_sjoin
** parv[0] - sender
** parv[1] - TS
** parv[2] - channel
** parv[3] - modes + n arguments (key and/or limit)
** parv[4+n] - flags+nick list (all in one parameter)
*/
int m_sjoin(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*ts = ircd_nextarg(&args, 0);
	char	*channel = ircd_nextarg(&args, 0);
	char	*modes = ircd_nextarg(&args, 0);
	char	*modeargs = args;
	char	*newkey = NULL;
	int		newlimit = 0;
	char	*tempptr;
	char	*ptr;
	int		newts;
	int		oldts;
	int		isnew = 0;
	int		doesop = 0;
	int		haveops = 0;
	int		keepourmodes = 1;
	int		keepnewmodes = 1;
	int		newmodes;
	int		oldmodes;
	int		tstosend;
	char	plusminus;
	char	modebuf[512];
	char	modeargbuf[512];
	char	*modeptr = modebuf;
	char	*modeargptr = modeargbuf;
	char	largebuf[1024];
	int		people;
	IRCD_CLIENT	*acptr;
	IRCD_CHANNEL	*chptr;
	IRCD_CLIENTLIST	*clptr;
	char	*name;
	int		flags;
	char	*sptr;
	int		pargs;
	int		*i;

	if (!IsServer(pptr))
	{
		if (IsLocal(pptr))
			;
		return 3;
	}

	if (!modeargs || !*modeargs)
		return 3;

	if (*channel == '&')
		return 3;

	newts = atoi(ts);
	chptr = hash_find_channel(channel);

	newmodes = 0;

	for(ptr=modes;*ptr;ptr++)
	{
		switch(*ptr)
		{
			case 'i':
				newmodes |= CHANNEL_MODES_INVITEONLY;
				break;
			case 'm':
				newmodes |= CHANNEL_MODES_MODERATED;
				break;
			case 'n':
				newmodes |= CHANNEL_MODES_NOPRIVMSGS;
				break;
			case 'p':
				newmodes |= CHANNEL_MODES_PRIVATE;
				break;
			case 's':
				newmodes |= CHANNEL_MODES_SECRET;
				break;
			case 't':
				newmodes |= CHANNEL_MODES_TOPICLIMIT;
				break;
			case 'k':
				newkey = ircd_next_subarg_space(&modeargs);
				if (newkey && !*newkey)
					newkey = NULL;
				break;
			case 'l':
				tempptr = ircd_next_subarg_space(&modeargs);
				if (tempptr)
					newlimit = atoi(tempptr);
				break;
		}
	}
			
	if (!chptr)
	{
		isnew = 1;
		chptr = (IRCD_CHANNEL *)MyMalloc(sizeof(IRCD_CHANNEL)+strlen(channel));
		strcpy(chptr->name, channel);
		hash_add_channel(chptr);
		if (++Globalinfo->channels > Globalinfo->maxchannels)
		{
			Globalinfo->maxchannels = Globalinfo->channels;
			Globalinfo->maxchannelstime = Now;
		}
	}
	oldts = chptr->creation;

	if (modeargs && (*modeargs == ':'))
		modeargs++;

	if (modeargs && !isnew && (newts > 0) && (oldts > 0) && (newts != oldts))
	{
		char	*start = modeargs;
		char	save;

		for(ptr=start;;ptr++)
		{
			if (*ptr && *ptr != ' ')
				continue;
			save = *ptr;
			*ptr = '\0';
			if ((*start == '@') || (*(start+1) == '@'))
			{
				if ((*++start == '@') || (*start == '+'))
					start++;
				if (((acptr = hash_find_chasing(start)) != NULL) &&
						(acptr->c_from == lptr))
				{
					*ptr = save;
					doesop = 1;
					break;
				}
			
			}
			start = ptr+1;
			if (save == '\0')
				break;
			else
				*ptr = save;
		}
	}

	for(clptr=chptr->members;clptr;clptr=clptr->next)
	{
		if (clptr->modes & CHANNEL_MODES_CHANOP)
		{
			haveops = 1;
			break;
		}
	}

	oldmodes = chptr->modes;

	if (isnew)
		chptr->creation = tstosend = newts;
	else if (!newts || !oldts)
		chptr->creation = tstosend = 0;
	else if (newts == oldts)
		tstosend = oldts;
	else if (newts < oldts)
	{
#ifdef TS5
		keepourmodes = 0;
		chptr->creation = tstosend = newts;
#else
		if (doesop)
		{
			keepourmodes = 0;
			chptr->creation = tstosend = newts;
		}
		else if (haveops)	/* !doesop and we have ops */
		{
			tstosend = oldts;
		}
		else /* !doesop and we don't have ops */
		{
			chptr->creation = tstosend = newts;
		}
#endif
	}
	else
	{
#ifdef TS5
		keepnewmodes = 0;
		tstosend = oldts;
#else
		if (haveops)
		{
			keepnewmodes = 0;
			tstosend = oldts;
		}
		else if (doesop) /* !haveops && doesop */
		{
			chptr->creation = tstosend = newts;
			send_to_opers("Hacked ops on opless channel: %s from %s",
					chptr->name, pptr->name);
		}
		else /* !haveops && !doesop */
			tstosend = oldts;
#endif
	}

	if (!keepnewmodes)
	{
		newmodes = oldmodes;
		newkey = chptr->key;
		newlimit = chptr->limit;
	}
	else if (keepourmodes)
	{
		newmodes |= oldmodes;
		if (chptr->limit > newlimit)
			newlimit = chptr->limit;
		if (strcmp(newkey ? newkey : "", chptr->key ? chptr->key : "") < 0)
			newkey = chptr->key;
	}

	plusminus = '\0';

	if (newmodes != oldmodes)
	{
		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & newmodes) && !(*i & oldmodes))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
			}
		}

		for(i=_channel_modes;*i;i+=2)
		{
			if ((*i & oldmodes) && !(*i & newmodes))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(i+1);
			}
		}
	}

	if (chptr->limit && !newlimit)
	{
		if (plusminus != '-')
		{
			plusminus = '-';
			*modeptr++ = plusminus;
		}
		*modeptr++ = 'l';
	}

	if (chptr->key && !newkey)
	{
		if (plusminus != '-')
		{
			plusminus = '-';
			*modeptr++ = plusminus;
		}
		*modeptr++ = 'k';
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(sptr=chptr->key;*sptr;)
			*modeargptr++ = *sptr++;
	}

	if (newlimit && (newlimit != chptr->limit))
	{
		char buf[20];

		if (plusminus != '+')
		{
			plusminus = '+';
			*modeptr++ = plusminus;
		}
		*modeptr++ = 'l';
		sprintf(buf, "%-15d", newlimit);
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(sptr=buf;*sptr && (*sptr != ' ');)
			*modeargptr++ = *sptr++;
	}

	if (newkey && (!chptr->key || mycmp(chptr->key, newkey)))
	{
		if (plusminus != '+')
		{
			plusminus = '+';
			*modeptr++ = plusminus;
		}
		*modeptr++ = 'k';
		if (modeargptr != modeargbuf)
			*modeargptr++ = ' ';
		for(sptr=newkey;*sptr;)
			*modeargptr++ = *sptr++;
	}

	chptr->modes = newmodes;

	if (newkey != chptr->key)
	{
		if (chptr->key)
			MyFree(chptr->key);
		chptr->key = !newkey ? newkey : MyStrdup(newkey);
	}
	if (newlimit != chptr->limit)
		chptr->limit = newlimit;

	if (!keepourmodes)
	{
		char	meplusminus = '\0';
		int		mpargs = 0;
		char	memodebuf[512];
		char	memodeargbuf[512];
		char	*memodeptr = memodebuf;
		char	*memodeargptr = memodeargbuf;

		for(clptr=chptr->members;clptr;clptr=clptr->next)
		{
			if (clptr->modes & CHANNEL_MODES_CHANOP)
			{
				if (meplusminus != '-')
				{
					meplusminus = '-';
					*memodeptr++ = meplusminus;
				}
				*memodeptr++ = 'o';
				if (memodeargptr != memodeargbuf)
					*memodeargptr++ = ' ';
				for(sptr=clptr->cptr->name;*sptr;)
					*memodeargptr++ = *sptr++;
				if (++mpargs == MAXMODEPARAMS)
				{
					*memodeptr = '\0';
					*memodeargptr = '\0';
					send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, meptr, "MODE %s %s %s", chptr->name,
						memodebuf, memodeargbuf);
					memodeptr = memodebuf;
					memodeargptr = memodeargbuf;
					meplusminus = '\0';
					mpargs = 0;
				}
				clptr->modes &= ~CHANNEL_MODES_CHANOP;
			}
			if (clptr->modes & CHANNEL_MODES_VOICE)
			{
				if (meplusminus != '-')
				{
					meplusminus = '-';
					*memodeptr++ = meplusminus;
				}
				*memodeptr++ = 'v';
				if (memodeargptr != memodeargbuf)
					*memodeargptr++ = ' ';
				for(sptr=clptr->cptr->name;*sptr;)
					*memodeargptr++ = *sptr++;
				if (++mpargs == MAXMODEPARAMS)
				{
					*memodeptr = '\0';
					*memodeargptr = '\0';
					send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, meptr, "MODE %s %s %s", chptr->name,
						memodebuf, memodeargbuf);
					memodeptr = memodebuf;
					memodeargptr = memodeargbuf;
					meplusminus = '\0';
					mpargs = 0;
				}
				clptr->modes &= ~CHANNEL_MODES_VOICE;
			}
		} /* for() */
		if (memodeptr != memodebuf)
		{
			*memodeptr = '\0';
			*memodeargptr = '\0';
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, meptr, "MODE %s %s %s", chptr->name,
				memodebuf, memodeargbuf);
		}
	} /* (!keepourmodes) */

	if (modeptr != modebuf)
	{
		*modeptr = '\0';
		*modeargptr = '\0';
		if (IsClient(pptr))
		{
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "MODE %s %s %s", chptr->name,
				modebuf, modeargbuf);
		}
		else
		{
			send_to_channel_butone2(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, prefix, "MODE %s %s %s", chptr->name,
				modebuf, modeargbuf);
		}
		modeptr = modebuf;
		modeargptr = modeargbuf;
	}

	if ((*modes != '0') && keepnewmodes)
		_build_channel_modes(chptr, pptr, modebuf, modeargbuf);
	else
	{
		modebuf[0] = '0';
		modebuf[1] = '\0';
		modeargbuf[0] = '\0';
	}

	tempptr = largebuf;
	tempptr += sprintf(tempptr, ":%s SJOIN %d %s %s %s :",
				prefix, tstosend, chptr->name,
				modebuf, modeargbuf);

	*modeptr++ = '+';

	people = 0;
	pargs = 0;


	for(name = ircd_next_subarg_space(&modeargs);name;
					name = ircd_next_subarg_space(&modeargs))
	{
		char *savename = name;

		flags = 0;
		if (*name == '@' || *(name+1) == '@')
			flags |= CHANNEL_MODES_CHANOP;
		if (*name == '+' || *(name+1) == '+')
			flags |= CHANNEL_MODES_VOICE;

		if (flags && ((*++name == '@') || (*name == '+')))
			name++;

		if (!keepnewmodes)
		{
			if (flags & CHANNEL_MODES_CHANOP)
				flags = CHANNEL_MODES_DEOPPED;
			else
				flags = 0;
		}

		if (!(acptr = hash_find_chasing(name)) || (acptr->c_from != lptr))
			continue;
		people++;

		if (!IsMember(chptr, acptr))
		{
			_channel_add_user(chptr, acptr, flags);
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, acptr, "JOIN :%s", chptr->name);
		}
		if (keepnewmodes)
			strcpy(tempptr, savename);
		else
			strcpy(tempptr, name);
		tempptr += strlen(tempptr);
		*tempptr++ = ' ';
		if (flags & CHANNEL_MODES_CHANOP)
		{
			*modeptr++ = 'o';
			if (modeargptr != modeargbuf)
				*modeargptr++ = ' ';
			for(sptr=name;*sptr;)
				*modeargptr++ = *sptr++;
			if (++pargs == MAXMODEPARAMS)
			{
				*modeptr = '\0';
				*modeargptr = '\0';
				if (IsClient(pptr))
					send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, pptr, "MODE %s %s %s", chptr->name,
						modebuf, modeargbuf);
				else
					send_to_channel_butone2(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, prefix, "MODE %s %s %s", chptr->name,
						modebuf, modeargbuf);
				pargs = 0;
				modeptr = modebuf;
				modeargptr = modeargbuf;
				*modeptr++ = '+';
			}
		}
		if (flags & CHANNEL_MODES_VOICE)
		{
			*modeptr++ = 'v';
			if (modeargptr != modeargbuf)
				*modeargptr++ = ' ';
			for(sptr=name;*sptr;)
				*modeargptr++ = *sptr++;
			if (++pargs == MAXMODEPARAMS)
			{
				*modeptr = '\0';
				*modeargptr = '\0';
				if (IsClient(pptr))
					send_to_channel_butone(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, pptr, "MODE %s %s %s", chptr->name,
						modebuf, modeargbuf);
				else
					send_to_channel_butone2(NULL,
						CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
						chptr, prefix, "MODE %s %s %s", chptr->name,
						modebuf, modeargbuf);
				pargs = 0;
				modeptr = modebuf;
				modeargptr = modeargbuf;
				*modeptr++ = '+';
			}
		}
	}

	if (modeptr != (modebuf+1))
	{
		*modeptr = '\0';
		*modeargptr = '\0';
		if (IsClient(pptr))
			send_to_channel_butone(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, pptr, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);
		else
			send_to_channel_butone2(NULL,
				CHANNEL_SEND_CLIENTS|CHANNEL_SEND_LOCALONLY,
				chptr, prefix, "MODE %s %s %s", chptr->name, modebuf, modeargbuf);
	}

	if (people)
	{
		if (*(tempptr-1) == ' ')
			tempptr--;
		*tempptr = '\0';
		send_to_servers(lptr, "%s", largebuf);
	}

	return 3;
}

#if 0
static void _send_channel_modes(IRCD_CLIENT *cptr, IRCD_CHANNEL *chptr)
{
	char				modebuf[64];
	char				parabuf[512];
	char				buf[512];
	char				*t;
	IRCD_CLIENTLIST		*clptr;
	IRCD_CLIENTLIST		*anop = NULL;
	IRCD_CLIENTLIST		*skip = NULL;
	IRCD_BAN			*bans;
	IRCD_EXCEPT			*excepts;
	IRCD_QUIET			*quiets;
	int					n = 0;
	int					paralen;
	int					namlen;
	char				*cp;
	int					count;
	int					send;

	if (*chptr->name != '#') /* don't send '&' channels */
		return;
	_build_channel_modes(chptr, cptr, modebuf, parabuf);

	t = buf + sprintf(buf, ":%s SJOIN %u %s %s %s%s:",
			meptr->name, chptr->creation, chptr->name, modebuf,
			parabuf, *parabuf ? " " : "");

	for(clptr=chptr->members;clptr;clptr=clptr->next)
	{
		if (clptr->modes & CHANNEL_MODES_CHANOP)
		{
			anop = clptr;
			break;
		}
	}
	clptr = NULL;
	for(;;)
	{
		if (anop)
		{
			clptr = skip = anop;
			anop = NULL;
		}
		else
		{
			if (clptr == NULL || clptr == skip)
				clptr = chptr->members;
			else
				clptr = clptr->next;
			if (clptr && clptr == skip)
				clptr = clptr->next;
			if (clptr == NULL)
				break;
		}
		if (clptr->modes & CHANNEL_MODES_CHANOP)
			*t++ = '@';
		if (clptr->modes & CHANNEL_MODES_VOICE)
			*t++ = '+';
		strcpy(t, clptr->cptr->name);
		t += strlen(t);
		*t++ = ' ';
		n++;
		if (t - buf > 512 - HOSTLEN - 3)
		{
			*t++ = '\0';
			if (t[-1] == ' ')
				t[-1] = '\0';
			send_to_client(cptr, "%s", buf);
			t = buf + sprintf(buf, ":%s SJOIN %u %s 0 :",
					meptr->name, chptr->creation, chptr->name);
			n = 0;
		}
	}
	if (n)
	{
		*t++ = '\0';
		if (t[-1] == ' ')
			t[-1] = '\0';
		send_to_client(cptr, "%s", buf);
	}
	*parabuf = '\0';
	*modebuf = '+';
	cp = modebuf + 1;
	count = 0;
	paralen = 0;
	send = 0;
		
	/* send bans */

	for(bans=chptr->bans;bans;bans=bans->next)
	{
		if (paralen + (namlen = strlen(bans->banstring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, bans->banstring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'b';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, bans->banstring);
				paralen = namlen;
				*cp++ = 'b';
			}
			count = 1;
		}
	}

	/* then send exceptions and quiets */
	for(excepts=chptr->excepts;excepts;excepts=excepts->next)
	{
		if (paralen + (namlen = strlen(excepts->estring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, excepts->estring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'e';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, excepts->estring);
				paralen = namlen;
				*cp++ = 'e';
			}
			count = 1;
		}
	}

	for(quiets=chptr->quiets;quiets;quiets=quiets->next)
	{
		if (paralen + (namlen = strlen(quiets->qstring)) + 10 < 200)
		{
			*(parabuf+paralen) = ' ';
			strcpy(parabuf+paralen+1, quiets->qstring);
			paralen += namlen + 1;
			count++;
			*cp++ = 'q';
		}
		else if (*parabuf)
			send = 1;
		if (count == 4)
			send = 1;
		if (send)
		{
			*cp = '\0';
			send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
					chptr->name, modebuf, parabuf);
			send = 0;
			*parabuf = '\0';
			paralen = 0;
			cp = modebuf;
			*cp++ = '+';
			if (count != 4)
			{
				strcpy(parabuf, quiets->qstring);
				paralen = namlen;
				*cp++ = 'q';
			}
			count = 1;
		}
	}


	if (*parabuf)
	{
		*cp = '\0';
		send_prefix_to_client(meptr, cptr, "MODE %s %s %s",
				chptr->name, modebuf, parabuf);
	}
}
#endif

void _umode_build_change(IRCD_CLIENT *pptr, char *buf, int oldmodes, int sendmask)
{
	int		*s;
	int		flag;
	char	plusminus = '\0';
	char	*modeptr = buf;

	if ((oldmodes & sendmask) != (pptr->c_umodes & sendmask))
	{
		for(s=_user_modes;(flag=*s);s+=2)
		{
			if (MyClient(pptr) && !(flag & sendmask))
				continue;
			if ((flag & oldmodes) && !(pptr->c_umodes & flag))
			{
				if (plusminus != '-')
				{
					plusminus = '-';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(s+1);
			}
			else if (!(flag & oldmodes) && (pptr->c_umodes & flag))
			{
				if (plusminus != '+')
				{
					plusminus = '+';
					*modeptr++ = plusminus;
				}
				*modeptr++ = *(s+1);
			}
		}
	}
	*modeptr = '\0';
}

void _umode_build(IRCD_CLIENT *pptr, char *buf, int mask)
{
	int		*s;
	int		flag;

	*buf++ = '+';
	for(s=_user_modes;(flag=*s);s+=2)
	{
		if (!(flag & mask))
			continue;
		if (pptr->c_umodes & flag)
			*buf++ = *(s+1);
	}
	*buf = '\0';
}

#ifdef HMC
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *nick, char *prefix, char *args, char *hmc)
#else
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *nick, char *prefix, char *args)
#endif
{
	char			*modes 	= ircd_nextarg(&args, 0);
	char			plusminus = '+';
	IRCD_CLIENT		*acptr;
	char			ubuf[64];
	int				oldmodes;
	char			*ptr;
	int				*i;
	int			showflags	= 0;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!nick || !*nick) /* should always be false, checked in m_mode() */
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				pptr->name), "UMODE");
		return 3;
	}

	acptr = hash_find_client(nick);
	if (!acptr || IsServer(acptr))
	{
		if (IsLocal(pptr))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOSUCHCHANNEL, pptr->name),
				nick);
		}
		return 3;
	}

	if (acptr != pptr)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_USERSDONTMATCH, pptr->name));
		return 3;
	}

	if (!modes || !*modes)
	{
		_umode_build(acptr, ubuf, -1);
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_UMODEIS, pptr->name), ubuf);
		return 3;
	}

	oldmodes = pptr->c_umodes;

	for(ptr=modes;*ptr;ptr++)
	{
		switch(*ptr)
		{
			case '+':
			case '-':
				plusminus = *ptr;
				break;
			default:
				for(i=_user_modes;*i;i+=2)
					if (*ptr == *(i+1))
						break;
				if (!*i)
				{
					send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(ERR_UMODEUNKNOWNFLAG, pptr->name), *ptr);
					continue;
				}
				if (MyClient(pptr))
				{
#ifdef OPER_GOD_MODE /* made originally as a debugging feature on glines, left in because someone wanted it on their net;
		      * this is an "undocumented feature" which will probably be pulled completely later */
						/* +r */
					if(*i == CLIENT_UMODES_REGISTERED)
						continue;				/* +x */
					if(!IsAnOper(pptr) && MyClient(pptr) && (*i == CLIENT_UMODES_SERVICE))
						continue;
#else
						/* +x */			/* +r */
					if(*i == CLIENT_UMODES_SERVICE || *i == CLIENT_UMODES_REGISTERED)
						continue;
#endif
				}
				if (plusminus == '+')
					pptr->c_umodes |= *i;
				else
					pptr->c_umodes &= ~(*i);
				break;
		}
	}

	if (pptr->c_umodes == oldmodes)
		return 3;

	if (!(oldmodes & CLIENT_UMODES_OPER) &&
					(pptr->c_umodes & CLIENT_UMODES_OPER))
	{
		if (IsServer(lptr))
			Globalinfo->g_opers++;
		else
			pptr->c_umodes &= ~CLIENT_UMODES_OPER;
	}
	else if (!(pptr->c_umodes & CLIENT_UMODES_OPER) &&
				(oldmodes & CLIENT_UMODES_OPER))
	{
		/* deopering; remove all oper-only modes to return them to normal */

		if(pptr->c_umodes & CLIENT_UMODES_FMODE)
			pptr->c_umodes &= ~CLIENT_UMODES_FMODE;
                if(pptr->c_umodes & CLIENT_UMODES_KMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_KMODE;
	        if(pptr->c_umodes & CLIENT_UMODES_WMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_WMODE;
		if(pptr->c_umodes & CLIENT_UMODES_ZMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_ZMODE;
		if(pptr->c_umodes & CLIENT_UMODES_YMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_YMODE;
		if(pptr->c_umodes & CLIENT_UMODES_CMODE)
			pptr->c_umodes &= ~CLIENT_UMODES_CMODE;
		if(pptr->c_umodes & CLIENT_UMODES_NMODE)
		        pptr->c_umodes &= ~CLIENT_UMODES_NMODE;
#ifdef OPER_GOD_MODE /* see comments in m_mode() about this */
		if(pptr->c_umodes & CLIENT_UMODES_SERVICE)
		        pptr->c_umodes &= ~CLIENT_UMODES_SERVICE;
#endif
		/* strip any of the clients oper only flags */
		if(MyClient(pptr))
		{
			IRCD_USERFLAG 	*flptr;
			for(flptr = _user_flags;flptr->flagstr;flptr++)
		            if (flptr->opersonly && (pptr->c_nflags & flptr->flag))
				    pptr->c_nflags &= ~flptr->flag;
			showflags = 1;

			pptr->c_wl = pptr->c_old_wl;
			pptr->c_oline_flags = 0;
		}
			
		Globalinfo->g_opers--;
		assert(pptr->c_oslist);
		del_client_from_oslist(pptr->c_oslist, pptr);
		add_client_to_oslist(Clients, pptr);

	}

/* can't get this one from remote server...*/
	if (!(oldmodes & CLIENT_UMODES_LOCOP) &&
					(pptr->c_umodes & CLIENT_UMODES_LOCOP))
	{
		if (!IsServer(lptr))
			pptr->c_umodes &= ~CLIENT_UMODES_LOCOP;
	}

	if (!(oldmodes & CLIENT_UMODES_IMODE) &&
			(pptr->c_umodes & CLIENT_UMODES_IMODE))
	{
		if (IsLocal(pptr))
		{
			Globalinfo->m_invis++;
			pptr->c_nflags |= CLIENT_NFLAGS_INVISIBLE; /* yes, the flags are usually for notices and things, but
								    * adding flags for invisible and deaf umodes saves people asking
								    * the opers how things work --myriad */
			showflags = 1;
		}

		Globalinfo->g_invis++;
	}
	else if (!(pptr->c_umodes & CLIENT_UMODES_IMODE) &&
			(oldmodes & CLIENT_UMODES_IMODE))
	{
		if (IsLocal(pptr))
		{
			Globalinfo->m_invis--;
			pptr->c_nflags &= ~CLIENT_NFLAGS_INVISIBLE;
			showflags = 1;
		}

		Globalinfo->g_invis--;
		assert(pptr->c_oslist);
		del_client_from_oslist(pptr->c_oslist, pptr);
		add_client_to_oslist(Clients, pptr);
	}

       if (!(oldmodes & CLIENT_UMODES_DEAF) &&
                         (pptr->c_umodes & CLIENT_UMODES_DEAF))
        {
		if(IsLocal(pptr))
		{
                	pptr->c_nflags |= CLIENT_NFLAGS_DEAF;
        		showflags = 1;
		}
        }
        else if ((oldmodes & CLIENT_UMODES_DEAF) &&
                             !(pptr->c_umodes & CLIENT_UMODES_DEAF))
        {
		if(IsLocal(pptr))
		{
                	pptr->c_nflags &= ~CLIENT_NFLAGS_DEAF;
        		showflags = 1;
		}
        }

	if (!(oldmodes & CLIENT_UMODES_ONLYREGMSGS) &&
			(pptr->c_umodes & CLIENT_UMODES_ONLYREGMSGS))
	{
		if(IsLocal(pptr))
		{
			pptr->c_nflags |= CLIENT_NFLAGS_ONLYREGISTEREDMSGS;
			showflags = 1;
			if(!IsRegistered(pptr))
				send_prefix_to_client(meptr, pptr, "NOTICE %s :*** Note that umode +R only has an effect if you are using a registered nick", 
								pptr->name);
		}
	}
	else if ((oldmodes & CLIENT_UMODES_ONLYREGMSGS) &&
				!(pptr->c_umodes & CLIENT_UMODES_ONLYREGMSGS))
	{
		if(IsLocal(pptr))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_ONLYREGISTEREDMSGS;
			showflags = 1;
		}
	}

	if (IsLocal(pptr))
	{
		/* int showflags = 0; */

		if (!(oldmodes & CLIENT_UMODES_ZMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_ZMODE))
		{
			if (IsAnOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_OWALLOPS;
				showflags = 1;
			}
			else
			{
				pptr->c_umodes &= ~CLIENT_UMODES_ZMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_ZMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_ZMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_OWALLOPS;
			showflags = 1;
		}
		/* umode +s is no longer restricted because it shows NOTHING beyond a "highest connection count" notice */
		if (!(oldmodes & CLIENT_UMODES_SMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_SMODE))
		{
				pptr->c_nflags |= CLIENT_NFLAGS_SNOTICES;
				showflags = 1;
		}
		else if ((oldmodes & CLIENT_UMODES_SMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_SMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SNOTICES;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_FMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_FMODE))
		{
			if(IsAnOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_SKILLS;
				showflags = 1;
			}
			else
				pptr->c_umodes &= ~CLIENT_UMODES_FMODE;
		}
		else if ((oldmodes & CLIENT_UMODES_FMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_FMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SKILLS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_KMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_KMODE))
		{
			if(IsAnOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_OKILLS;
				showflags = 1;
			}
			else
				pptr->c_umodes &= ~CLIENT_UMODES_KMODE;
		}
		else if ((oldmodes & CLIENT_UMODES_KMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_KMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_OKILLS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_WMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_WMODE))
		{
			if(IsAnOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_SWALLOPS;
				showflags = 1;
			}
			else
				pptr->c_umodes &= ~CLIENT_UMODES_WMODE;
		}
		else if ((oldmodes & CLIENT_UMODES_WMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_WMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_SWALLOPS;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_CMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_CMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS;
				showflags = 1;
			}
			else
			{
				pptr->c_nflags &= ~(CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS);
				pptr->c_umodes &= ~CLIENT_UMODES_CMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_CMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_CMODE))
		{
			pptr->c_nflags &= ~(CLIENT_NFLAGS_CLICONNECTS|CLIENT_NFLAGS_CLIDISCONNECTS);
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_NMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_NMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_NICKCHANGES;
				showflags = 1;
			}
			else
			{
				pptr->c_nflags &= ~CLIENT_NFLAGS_NICKCHANGES;
				pptr->c_umodes &= ~CLIENT_UMODES_NMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_NMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_NMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_NICKCHANGES;
			showflags = 1;
		}
		if (!(oldmodes & CLIENT_UMODES_YMODE) &&
				(pptr->c_umodes & CLIENT_UMODES_YMODE))
		{
			if (IsOper(pptr))
			{
				pptr->c_nflags |= CLIENT_NFLAGS_STATSNOTICES;
				showflags = 1;
			}
			else
			{
				pptr->c_nflags &= ~CLIENT_NFLAGS_STATSNOTICES;
				pptr->c_umodes &= ~CLIENT_UMODES_YMODE;
			}
		}
		else if ((oldmodes & CLIENT_UMODES_YMODE) &&
				!(pptr->c_umodes & CLIENT_UMODES_YMODE))
		{
			pptr->c_nflags &= ~CLIENT_NFLAGS_STATSNOTICES;
			showflags = 1;
		}

		if (showflags)
			m_flags(lptr, pptr, prefix, NULL);

		if (oldmodes == pptr->c_umodes)
			return 3;
	}

#ifdef HMC
	if(hmc && !MyClient(pptr))
	{
		hmc++; 
		set_hmc_remote(pptr, hmc);
	}

	if(MyClient(pptr))	
		set_hmc(pptr);
#endif
	
	_umode_build_change(pptr, ubuf, oldmodes, SEND_UMODES);
	if (*ubuf)
#ifdef HMC
		send_to_servers(lptr, ":%s MODE %s %s !%s", pptr->name, pptr->name, ubuf, hmc_decide(pptr));
#else
		send_to_servers(lptr, ":%s MODE %s %s", pptr->name, pptr->name, ubuf);
#endif
	_umode_build_change(pptr, ubuf, oldmodes, -1);
#if 0
	if (MyClient(pptr))
		send_prefix_to_client(pptr, pptr, "MODE %s %s", pptr->name, buf);
/* umodes only send client's nick, not full n!u@h */
#else
	if (MyClient(pptr))
		send_to_client(pptr, ":%s MODE %s %s", pptr->name, pptr->name, ubuf);
#endif

	return 3;
}
