/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   Portions Copyright (C) 2001 Adam Loghry
** 
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
//#include "cmds.h"

/* just /flags
 *
 * FIXME: improve the flags system to have a var for umodes.. so i can get rid of this hardcoding bs -myr */

extern int _user_modes[];

IRCD_USERFLAG _user_flags[] =
{
        { CLIENT_NFLAGS_OWALLOPS,                       "OWALLOPS",                     1       },
        { CLIENT_NFLAGS_SWALLOPS,                       "SWALLOPS",                     1       },
        { CLIENT_NFLAGS_STATSNOTICES,                   "STATSNOTICES",                 1       },
        { CLIENT_NFLAGS_OKILLS,                         "OKILLS",                       1       },
        { CLIENT_NFLAGS_SKILLS,                         "SKILLS",                       1       },
        { CLIENT_NFLAGS_SNOTICES,                       "SNOTICES",                     0       },
        { CLIENT_NFLAGS_CLICONNECTS,                    "CLICONNECTS",                  1       },
        { CLIENT_NFLAGS_CLIDISCONNECTS,                 "CLIDISCONNECTS",               1       },
        { CLIENT_NFLAGS_THROTTLES,                      "THROTTLES",                    1       },
        { CLIENT_NFLAGS_NICKCHANGES,                    "NICKCHANGES",                  1       },
        { CLIENT_NFLAGS_IPMISMATCHES,                   "IPMISMATCHES",                 1       },
        { CLIENT_NFLAGS_LWALLOPS,                       "LWALLOPS",                     1       },
        { CLIENT_NFLAGS_CONNECTS,                       "CONNECTS",                     1       },
        { CLIENT_NFLAGS_SQUITS,                         "SQUITS",                       1       },
        { CLIENT_NFLAGS_DENYEXPIRES,                    "EXPIREDDENIES",                1,      },
        { CLIENT_NFLAGS_DENIES,                 	"DENIES",       		1, 	}, 
        { CLIENT_NFLAGS_INVISIBLE,              	"INVISIBLE",            	0       },
        { CLIENT_NFLAGS_DEAF,                   	"DEAF",                 	0       },
        { CLIENT_NFLAGS_ONLYREGISTEREDMSGS,     	"ONLYREGISTEREDMSGS",   	0       },
        { 0, NULL, 0 }
};

int m_flags(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_USERFLAG		*ufptr;
	IRCD_USERFLAG		*ufptrkeep;
	char				*flag;
	char				plusminus = '+';
	char				buffer[255];
	char				*ptr = buffer;
	int					len = 0;
	int					oldflags = 0;
	int					flaglen;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	if (!IsClient(lptr))
		return 3;

	oldflags = pptr->c_nflags;

	if (args && *args == ':')
		args++;

	if (args && *args)
	{
		while((flag = ircd_next_subarg_space(&args)) != NULL)
		{
			if ((*flag == '+') || (*flag == '-'))
				plusminus = *flag++;
			if (!strcasecmp(flag, "all"))
			{
				if (plusminus == '-')
				{
					pptr->c_nflags = 0;
					continue;
				}
				for(ufptr = _user_flags;ufptr->flagstr;ufptr++)
					if (IsAnOper(pptr) || !ufptr->opersonly)
						pptr->c_nflags |= ufptr->flag;
				continue;
			}

			ufptrkeep = NULL;
			flaglen = strlen(flag);

			for(ufptr = _user_flags;ufptr->flagstr;ufptr++)
			{
				if (!strncasecmp(ufptr->flagstr, flag, flaglen))
				{
					if (ufptrkeep)
						break;
					ufptrkeep = ufptr;
				}
			}
			if (!ufptrkeep || ufptr->flagstr)
			{
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :Unknown flag: %s", pptr->name, flag);
				pptr->c_nflags = oldflags; /* restore flags */
				return 3;
			}
			ufptr = ufptrkeep;
			if (plusminus == '+')
			{
				if (ufptr->opersonly && !IsAnOper(pptr))
				{
					send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Need to be Oper to set %s",
						pptr->name, ufptr->flagstr);
					continue;
				}

				/* had to hard code all this crap and didnt want to, but it IS more user friendly in the end --myr */
				if(!strcasecmp(ufptr->flagstr, "INVISIBLE") && !(pptr->c_umodes & CLIENT_UMODES_IMODE) && !(pptr->c_nflags & CLIENT_NFLAGS_INVISIBLE))
				{
					send_to_client(pptr, ":%s MODE %s +i", pptr->name, pptr->name);
					pptr->c_umodes |= CLIENT_UMODES_IMODE;
				}
				if(!strcasecmp(ufptr->flagstr, "DEAF") && !(pptr->c_umodes & CLIENT_UMODES_DEAF) && !(pptr->c_nflags & CLIENT_NFLAGS_DEAF))
				{
					send_to_client(pptr, ":%s MODE %s +d", pptr->name, pptr->name);
					pptr->c_umodes |= CLIENT_UMODES_DEAF;
				}
				if(!strcasecmp(ufptr->flagstr, "ONLYREGISTEREDMSGS") && !(pptr->c_umodes & CLIENT_UMODES_ONLYREGMSGS) && !(pptr->c_nflags & CLIENT_NFLAGS_ONLYREGISTEREDMSGS))
				{
					if(!IsRegistered(pptr))
						send_prefix_to_client(meptr, pptr, "NOTICE %s :*** Note that umode +R only has an effect if you are using a registered nick",
									pptr->name);
					send_to_client(pptr, ":%s MODE %s +R", pptr->name, pptr->name);
					pptr->c_umodes |= CLIENT_UMODES_ONLYREGMSGS;
				}

				pptr->c_nflags |= ufptr->flag;
			}
			else
			{
				if(!strcasecmp(ufptr->flagstr, "INVISIBLE") && (pptr->c_umodes & CLIENT_UMODES_IMODE))
				{
					send_to_client(pptr, ":%s MODE %s -i", pptr->name, pptr->name);
					pptr->c_umodes &= ~CLIENT_UMODES_IMODE;
				}
				if(!strcasecmp(ufptr->flagstr, "DEAF") && (pptr->c_umodes & CLIENT_UMODES_DEAF))
				{
					send_to_client(pptr, ":%s MODE %s -d", pptr->name, pptr->name);
					pptr->c_umodes &= ~CLIENT_UMODES_DEAF;
				}
				if(!strcasecmp(ufptr->flagstr, "ONLYREGISTEREDMSGS") && (pptr->c_umodes & CLIENT_UMODES_ONLYREGMSGS) && (pptr->c_nflags & CLIENT_NFLAGS_ONLYREGISTEREDMSGS)) 
				{
					send_to_client(pptr, ":%s MODE %s -R", pptr->name, pptr->name);
					pptr->c_umodes &= ~CLIENT_UMODES_ONLYREGMSGS;
				}

				pptr->c_nflags &= ~ufptr->flag;
			}
		}
	}

	for(ufptr = _user_flags;ufptr->flagstr;ufptr++)
	{
		if (!(lptr->c_nflags & ufptr->flag))
			continue;
		len = strlen(ufptr->flagstr);
		if (len + 1 + (ptr-buffer) > 250)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Current flags: %s", pptr->name, buffer);
			ptr = buffer;
		}
		if (ptr != buffer)
			*ptr++ = ' ';
		strcpy(ptr, ufptr->flagstr);
		ptr += len;
	}
	if (ptr != buffer)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Current flags: %s", pptr->name, buffer);
	}
	ptr = buffer;
	for(ufptr = _user_flags;ufptr->flagstr;ufptr++)
	{
		if ((lptr->c_nflags & ufptr->flag))
			continue;
		if (ufptr->opersonly && !IsAnOper(pptr))
			continue;
		len = strlen(ufptr->flagstr);
		if (len + 1 + (ptr-buffer) > 250)
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :Current missing flags: %s", pptr->name, buffer);
			ptr = buffer;
		}
		if (ptr != buffer)
			*ptr++ = ' ';
		strcpy(ptr, ufptr->flagstr);
		ptr += len;
	}
	if (ptr != buffer)
	{
		send_prefix_to_client(meptr, pptr,
			"NOTICE %s :Current missing flags: %s", pptr->name, buffer);
	}
	return 3;
}
