/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   Portions Copyright (C) 2001 Adam Loghry
** 
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
//#include "cmds.h"
//#include "../lib/confparse/configparse.h"

RCSID("$Id: commands.c,v 1.20 2001/10/02 01:40:35 cbehrens Exp $");

static int handle_message(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int not);

int m_privmsg(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	        return handle_message(lptr, pptr, prefix, args, 0);
}

int m_notice(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	        return handle_message(lptr, pptr, prefix, args, 1);
}

static int handle_message(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int not)
{
#define MAX_MESSAGE_NICKS	50			/* TODO: make this into a /set var -myriad */
	char	*tostr = ircd_nextarg(&args, 0);
	char	*message = ircd_nextarg(&args, 0);
	char	*cmd;
	char	*sendtos[100];
	char	**st = NULL;
	char	**std;
	char	*nick;
	char	*server;
	int		resetidle = 0;
	IRCD_CLIENT	*acptr;
	int		numtried = 0;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	if (IsServer(pptr) && !not)
		return 3;
	cmd = not ? "NOTICE" : "PRIVMSG";
	if (!tostr)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NORECIPIENT, pptr->name), cmd);
		return 3;
	}
	if (!message || !*message)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTEXTTOSEND, pptr->name));
		return 3;
	}
	if (IsLocal(pptr))
	{
		int		num = 0;

		st = sendtos;
		for(;(nick = ircd_next_subarg_comma(&tostr)) != NULL;)
		{
			if (!*nick)
				continue;
			if (++num > MAX_MESSAGE_NICKS)
			{
				send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Too many recipients", pptr->name);
				return 3;
			}
			*st = nick;
			st++;
		}
		*st = NULL;
		st = sendtos;
	}
	for(;;)
	{
		if (st)
		{
			nick = *st;
			st++;
			if (nick)/* check for already sent */
			{
				for(std=sendtos;*std != nick;std++)
					if (!strcmp(*std, nick))
						break;
				if (*std != nick)
				{
					DEBUG(DEBUG_ALL, ("got dup %s for message\n", nick));
					continue;
				}
			}
		}
		else
			nick = ircd_next_subarg_comma(&tostr);
		if (!nick)
			break;
		numtried++;
		if (IsChannel(nick))
		{
			IRCD_CHANNEL	*chptr = hash_find_channel(nick);

			if (chptr)
			{
				if (_channel_cansend(pptr, chptr) == 0)
				{
					if (IsLocal(pptr) && ((chptr->numusers > 1) ||
							(chptr->members && 
							(chptr->members->cptr != pptr))))
					{
						if (!not)
							resetidle = 1;
					}
					send_to_channel_butone(lptr,
							CHANNEL_SEND_SERVERS|CHANNEL_SEND_CLIENTS|CHANNEL_SEND_SPEAKING,
							chptr, pptr, "%s %s :%s",
							cmd, nick, message);
				}
				else if (!not)
				{
					send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_CANNOTSENDTOCHAN, pptr->name), nick);
				}
				continue;
			}
		}
		else
		{
			IRCD_CLIENT		*nptr = hash_find_client(nick);
			if (nptr && IsClient(nptr))
			{
                        	if(IsDeaf(nptr))
					continue;
			        if(!IsRegistered(pptr) && (IsUmodeR(nptr) && IsRegistered(nptr)))
					continue;
				if (IsClient(pptr))
					send_prefix_to_client(pptr, nptr, "%s %s :%s",
						cmd, nick, message);
				else
					send_to_client(nptr, ":%s %s %s :%s",
						prefix, cmd, nick, message);
				if (nptr != pptr)
					resetidle = 1;
				continue;
			}
		}
		if (IsAnOper(pptr) && ((*nick == '$') || (*nick == '#')))
		{
			char *s;

			if ((s = strrchr(nick, '.')) == NULL)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOTOPLEVEL, pptr->name), nick);
				continue;
			}
			while(*++s)
			{
				if (*s == '*' || *s == '?')
					break;
			}
			if (*s)
			{
				send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_WILDTOPLEVEL,
						pptr->name), nick);
				continue;
			}
			/* sendto_match_butone */
			for(acptr=Locals;acptr;acptr=acptr->c_lo_next)
			{
				if ((IsServer(acptr) && (acptr != lptr)) ||
					(IsClient(acptr) && 
					(((*nick == '$') && !match(nick+1, acptr->servername)) ||
					((*nick == '#') && !match(nick+1, acptr->hostname)))))
				{
					if (IsClient(pptr))
						send_prefix_to_client(pptr, acptr, "%s %s :%s",
							cmd, nick, message); 
					else
						send_to_client(acptr, ":%s %s %s :%s",
							prefix, cmd, nick, message); 
				}
			}
			continue;
		}
		if (((server = strchr(nick, '@')) != NULL) &&
				((acptr = hash_find_client_incl_hm(server+1)) != NULL))
		{
			if (IsServer(acptr))
			{
				if (!IsMe(acptr))
				{
					send_to_client(acptr, ":%s %s %s :%s",
						prefix, cmd, nick, message);
					continue;
				}
				*server = '\0';
				if (!strcasecmp(nick, "OPERS") && IsAnOper(pptr))
				{
#if 0
					send_to_opers("To opers: From %s: %s",
						pptr->name, message);
#else
					for(acptr=*Opers;acptr;acptr=acptr->c_os_next)
    				{
						if (!IsLocal(acptr))
							continue;
						send_prefix_to_client(pptr, acptr, "%s %s :** TO LOCAL OPERS ** %s",
							cmd, acptr->name, message); 
					}
#endif
					continue;
				}
			}
		}
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHNICK, pptr->name),
					nick);
	}
	if (resetidle && !not /* && IsLocal(pptr) */ )
		pptr->c_idletime = Now;

	numtried /= 5;
	return numtried > 3 ? numtried : 3;
}
