/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   Portions Copyright (C) 2001 Adam Loghry
** 
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
//#include "cmds.h"
//#include "../lib/confparse/configparse.h"

RCSID("$Id: commands.c,v 1.20 2001/10/02 01:40:35 cbehrens Exp $");

extern int _user_modes[];

static int hostname_check(char *hostname)
{
        char *tptr = hostname;
        char *dotptr = NULL;

        for(;*tptr;tptr++)
        {
                if ((*tptr <= ' ') || (*tptr > 'z') || (*tptr == ':') ||
                                (*tptr == '*') || (*tptr == '%'))
                        break;
                if (*tptr == '.')
                        dotptr = tptr;
        }
        if (*tptr || !*hostname || !dotptr || (strlen(dotptr) > 4))
                return 0;

        return 1;
}

static int nickname_check(char *nick)
{
        char    *ch;

        if ((*nick == '-') || isdigit(*nick))
                return 0;
        for (ch = nick; *ch && (ch - nick) < NICKLEN; ch++)
                if (!isvalid(*ch) || isspace(*ch))
                        break;
        *ch = '\0';
        return (ch - nick);
}
static int _register_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr);
static int _do_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *username, char *hostname, char *servername, char *ircname);
static int _lusers_send(IRCD_CLIENT *pptr, char *pattern, int sendcounter);                                                
static int _motd_send(IRCD_CLIENT *pptr);
// static void _sendnick(IRCD_CLIENT *cptr, IRCD_CLIENT *acptr); /* should end up in server.c */
#if 0
#ifdef HMC
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *target, char *prefix, char *args, char *hmc);
#else
static int m_umode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *target, char *prefix, char *args);
#endif
#endif
//static IRCD_MODE *_ircdmode_create(IRCD_MODE **ircdmodes, IRCD_MODE **lastircdmode);
//static void _ircdmode_free(IRCD_MODE **bucket);
//static char *_make_nick_user_host(char *nick, char *user, char *host, char *buf, char **hostptr);
// static void do_who(IRCD_CLIENT *pptr, IRCD_CLIENT *acptr, IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr);
// static char *datestr(int thetime, char *buf);
#ifdef KLINES
static char *_reason_fix(char *reason);
static char *_cluster(char *hostname, int *ip_or_host);
#endif
// static void _send_servers_recurs(IRCD_CLIENT *lptr, IRCD_CLIENT *cptr);
static void _count_servers_and_users(IRCD_CLIENT *acptr, int *numservers, int *numusers);
// static void _channelname_clean(unsigned char *chname);
static int m_useriphost(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int justip);

static int _register_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr)
{
	int					err;
	char				exitbuf[512];

	pptr->c_flags |= CLIENT_FLAGS_REG;
	add_client_to_oslist(Clients, pptr);
	if (pptr->c_umodes & CLIENT_UMODES_IMODE)
		Globalinfo->g_invis++;
	if (pptr->c_umodes & CLIENT_UMODES_OPER) /* can happen if remote client */
		Globalinfo->g_opers++;
	if (++Globalinfo->g_clients > Globalinfo->g_maxclients)
	{
		Globalinfo->g_maxclients = Globalinfo->g_clients;
		Globalinfo->g_maxclientstime = Now;
	}
	if (IsLocal(pptr))
	{
		int					ret;
#ifdef KLINES
		CONFHOSTIPENTRY		*entryptr;
#endif
		char				*ptr1 = NULL;
		char				*ptr2;
		char				*uchrs;
		int					num;

		*exitbuf = '\0';

		if (++Globalinfo->m_clients > Globalinfo->m_maxclients)
		{
			Globalinfo->m_maxclients = Globalinfo->m_clients;
			Globalinfo->m_maxclientstime = Now;
		}
		if (Globalinfo->m_clients + Globalinfo->m_servers > Globalinfo->m_maxconn)
		{
			Globalinfo->m_maxconn = Globalinfo->m_clients + Globalinfo->m_servers;
			Globalinfo->m_maxconntime = Now;
			if (!(Globalinfo->m_maxconn % 10))
				send_to_flag(CLIENT_NFLAGS_SNOTICES, NULL,
					"New highest connections: %d (%d clients)",
					Globalinfo->m_maxconn, Globalinfo->m_clients);
		}
		if (pptr->c_umodes & CLIENT_UMODES_IMODE)
			Globalinfo->m_invis++;
		if (pptr->c_umodes & CLIENT_UMODES_OPER) /* can't happen */
			Globalinfo->m_opers++;

		/* check username */
		uchrs = (char *)vars_value_get_bynum(IRCD_VAR_USERNAMECHARS);

		if (*pptr->username)
		{
#ifdef HMC
			int x = 0;
#endif
			for(ptr1=pptr->username;*ptr1;ptr1++)
			{
				if (isalnum(*ptr1) || (*ptr1 == '~'))
					continue;
				if (!uchrs)
					break;
#ifdef HMC
				if (!x && *ptr1 == '+')
				{
					x = 1;
					continue;
				}
#endif
				for(ptr2=uchrs;*ptr2;ptr2++)
					if (*ptr1 == *ptr2)
						break;
				if (!*ptr2)
					break;
			}
		}
		if (!*pptr->username || *ptr1)
		{
			client_del(meptr, pptr, "Invalid username");
			return -1;
		}

		if (!hostname_check(pptr->hostname))
		{
			client_del(meptr, pptr, "Invalid hostname");
			return -1;
		}

		ret = ircd_checkallow(pptr, &err);
		if (ret < 0)
		{

			*exitbuf = '\0';
			switch(ret)
			{
				case	-1: /* no auth */
					sprintf(exitbuf, "No authorization: %i", err);
					break;
				case	-2: /* no auth */
					sprintf(exitbuf, "No authorization due to class error: %i", err);
					break;
				case	-3: /* class full */
					strcpy(exitbuf, "Class is full");
					break;
				case	-4: /* host limit exceeded */
					strcpy(exitbuf, "Class host limit exceeded");
					break;
				case	-5: /* host limit exceeded */
					strcpy(exitbuf, "Class userhost limit exceeded");
					break;
				case	-6: /* host limit exceeded */
					strcpy(exitbuf, "Class vhost limit exceeded");
					break;
				case	-7: /* host limit exceeded */
					strcpy(exitbuf, "Class uservhost limit exceeded");
					break;
				default:
					sprintf(exitbuf, "Unknown error: %i", err);
					break;
			}
			client_del(meptr, pptr, exitbuf);
			return -1;
		}

		if (pptr->c_res_req)
		{
			if (pptr->c_res_req->rl_forward)
				MyFree(pptr->c_res_req->rl_forward);
			if (pptr->c_res_req->rl_reverse)
				MyFree(pptr->c_res_req->rl_reverse);
			MyFree(pptr->c_res_req);
			pptr->c_res_req = NULL;
		}

		if (!(pptr->c_class->flags & CLASS_FLAGS_BYPASSIRCNAMEDENY))
		{
			STRINGDENY	*sd;

			if ((sd = stringdeny_match(&(_ircd_conf.ircname_denies),
							pptr->ircname, Now)) != NULL)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_YOUREBANNEDCREEP, pptr->name),
					sd->reason);
				client_del(meptr, pptr, "Invalid ircname");
				return -1;
			}
		}

		if (!(pptr->c_class->flags & CLASS_FLAGS_BYPASSLIMITS))
		{
			num = (int)vars_value_get_bynum(IRCD_VAR_MAXCLIENTS);
			if ((num > 0) && (Globalinfo->m_clients >= num))
			{
				client_del(meptr, pptr,
					"Server is full -- reached client maximum");
				return -1;
			}
		}
		if (!(pptr->c_class->flags & CLASS_FLAGS_BYPASSDENY))
		{
#ifdef KLINES
			ret = access_match_ipaddr(&(_ircd_conf.deny),
                		pptr->username, pptr->c_ipstring,
                		pptr->c_rem_ipaddr, ACCESS_FLAGS_BESTMATCH,
                		&entryptr);

			if (ret < 0)
				ret = access_match_hostname(&(_ircd_conf.deny),
	                pptr->username, pptr->hostname, ACCESS_FLAGS_BESTMATCH,
					&entryptr);
			if (ret >= 0) /* banned */
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_YOUREBANNEDCREEP, pptr->name),
					entryptr->data ? entryptr->data : "No reason specified");
				sprintf(exitbuf, "Banned: %s",
					entryptr->data ? entryptr->data : "No reason specified");
				client_del(meptr, pptr, exitbuf);
				return -1;
			}
#endif
#if 0
			check_deny_on(pptr);
#else
			{
        			IRCD_DENY       *dptr;
        			char            hostmask[USERLEN+HOSTLEN+1];
        			int             islocal;
        			int             isperm;

        			sprintf(hostmask, "%s@%s", pptr->username, pptr->hostname);

        			for(dptr=Denies;dptr;dptr=dptr->next)
        			{
                			if(!(dptr->type & DENY_TYPE_AKILL) || !dptr->mask)
                        			continue;

                			islocal = (dptr->type & DENY_TYPE_LOCAL) ? 1 : 0;
                			isperm  = (dptr->type & DENY_TYPE_PERM) ? 1 : 0;

                			if(!match(dptr->mask, hostmask))
                			{
                        				send_prefix_to_client(meptr, pptr, "NOTICE %s :*** You are not welcome on this %s; you have been autokilled.",
                                                		                pptr->name, islocal ? "server" : "network");
                        			if(isperm)
                                			send_prefix_to_client(meptr, pptr, "NOTICE %s :*** Set on: %s by %s - does not expire (permanent) - Reason: %s",
                                        			        pptr->name, support_ctime(dptr->when), dptr->oper, dptr->reason);
                        			else
                                			send_prefix_to_client(meptr, pptr, "NOTICE %s :*** Set on: %s by %s - Expires on: %s - Reason: %s",
                                                			pptr->name, support_ctime(dptr->when), dptr->oper, support_ctime(dptr->expire), dptr->reason);
                        			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_YOUREBANNEDCREEP, pptr->name), dptr->reason);
						client_del(meptr, pptr, "BANNED");
						return -1;
					}
				}
			}
#endif
		}

		if (vars_value_get_bynum(IRCD_VAR_AUTH_CHECK) && !(pptr->c_flags & CLIENT_FLAGS_GOTAUTH) &&
					(pptr->c_class->flags & CLASS_FLAGS_IDENTONLY))
		{
			client_del(meptr, pptr,
				"Sorry, you must have ident to connect");
			return -1;
		}
		/* notilde and spoofing dont get along */
		if (vars_value_get_bynum(IRCD_VAR_AUTH_CHECK) && pptr->c_class->flags & CLASS_FLAGS_NOTILDE && !(pptr->c_flags & CLIENT_FLAGS_SPOOFED))
		{
			if (!(pptr->c_flags & CLIENT_FLAGS_GOTAUTH))
			{
				char *tptr = pptr->username;
				while((*tptr = *(tptr+1)))
					tptr++;
			}
		}
	}

	if (iphash_handle_client(pptr)) /* client_del is called in here */
		return -1;

	pptr->c_servptr = hash_find_client_incl_hm(pptr->servername);
	if (!pptr->c_servptr || !IsServer(pptr->c_servptr))
	{
		send_to_opers("Ignoring new client %s from %s -- unknown server: %s",
				pptr->name, lptr->name, pptr->servername);
		sprintf(exitbuf, "Unknown server: %s", pptr->servername);
		client_del(meptr, pptr, exitbuf);
		return -1;
	}

	if (IsLocal(pptr))
		if (vars_value_get_bynum(IRCD_VAR_CLIENTCONNECT))
			send_to_flag(CLIENT_NFLAGS_CLICONNECTS, NULL,
				"Client connecting: %s (%s@%s) [%s]",
				pptr->name, pptr->username, pptr->hostname, pptr->c_ipstring);
	if (IsSpoofed(pptr))
		send_to_client(pptr, "NOTICE * :*** Your address is among the chosen; your hostname has been masked! (%s@%s -> %s@%s)",
					pptr->rl_username, pptr->rl_hostname, pptr->username, pptr->hostname);

	pptr->c_type = CLIENT_TYPE_CLIENT;
	add_client_to_ulist(&(pptr->c_servptr->c_users), pptr);

	if (IsLocal(pptr))
	{
		char	cnamebuf[CLIENTNAMESIZE];

		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WELCOME, pptr->name),
				vars_value_get_bynum(IRCD_VAR_NETWORK_NAME), pptr->name);
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_YOURHOST, pptr->name),
				ircd_clientname(meptr, 0, cnamebuf), VERSION);
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_CREATED, pptr->name),
				CREATION);
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_MYINFO, pptr->name),
				meptr->name, VERSION);
		/* 005 */		/* maybe a "CHANMODES=b,e,q,hi,etc.."? */ /* OSET */ /* HMC */
		send_to_client(pptr, ":%s 005 %s %sOMODE OKICK GLINE%s FLAGS MAP PRIORITYLAG%s USERIP WHOISMSGS PREFIX=(ov)@+ NETWORK=%s NETHOST=%s :are available on this server", 
							meptr->name, pptr->name, 
#ifdef NO_HUB
							"NO_HUB ",
#else
							"",
#endif
#ifdef KLINES
							" KLINE",
#else
							"",
#endif
							(ImDebugging) ? " DEBUG_MSGS" : "", NetName, NetHost);
		send_to_client(pptr, ":%s 005 %s MAXCHANNELS=%i/%i AWAYLEN=%i KICKLEN=512 QUITLEN=%i TOPICLEN=%i NICKLEN=%i MODES=%i CHANTYPES=&# :are available on this server",
							meptr->name, pptr->name,
								vars_value_get_bynum(IRCD_VAR_DEFAULT_MAXCHANNELS), MAXCHANNELSPERUSER, 
									AWAYLEN, QUITLEN, TOPICLEN, NICKLEN, MAXMODEPARAMS);
		send_to_client(pptr, ":%s 005 %s MAXBANS=%i MAXEXCEPTS=%i MAXQUIETS=%i CHANLEN=%i CHARSET=rfc1459 :are available on this server",
							meptr->name, pptr->name,
								MAXBANS, MAXEXCEPTS, MAXQUIETS, CHANNELLEN);
		
		Globalinfo->m_totclients++;
		_lusers_send(pptr, NULL, 0);
		if (vars_value_get_bynum(IRCD_VAR_SHOW_CLIENTCOUNTER))
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :You are client #%d to connect since %s",
				pptr->name, Globalinfo->m_totclients,
				support_ctime(Globalinfo->m_totclientstime));
		_motd_send(pptr);
	
		m_flags(lptr, pptr, pptr->name, NULL);
	
		if (vars_value_get_bynum(IRCD_VAR_SHOW_FLAGS_NOTICE))
		{
			send_prefix_to_client(meptr, pptr, "NOTICE %s :This server uses a system of FLAGS to complement user modes.", pptr->name);
			send_prefix_to_client(meptr, pptr, "NOTICE %s :To use this system, type /FLAGS (or /quote FLAGS). You may alter these flags with /quote FLAGS (+/-)flagname (ie: /quote FLAGS +INVISIBLE)", pptr->name);
		}
		
	}

	if (IsLocal(pptr)) /* NICK sent to servers in m_nick, otherwise */
	{
		char ubuf[64];

		_umode_build(pptr, ubuf, SEND_UMODES);
		send_to_servers(lptr, "NICK %s %d %i %s %s %s %s :%s",
				pptr->name, pptr->c_hopcount+1, pptr->c_creation,
				ubuf, pptr->username, pptr->hostname, pptr->servername, 
				pptr->ircname);
		_umode_build(pptr, ubuf, -1);
		if (ubuf[1])
			send_to_client(pptr, ":%s MODE %s :%s",
					pptr->name, pptr->name, ubuf);
	}

	

	return 3;
}
#ifdef KLINES
static char *_reason_fix(char *reason)
{
	static char	reasonbuf[401];	/* reason can only be 200 long, so.. */
	char *dptr = reasonbuf;
	char *sptr = reason;

	for(;*sptr;)
	{
		if (*sptr == '\"')
			*dptr++ = '\\';
		*dptr++ = *sptr++;
	}
	*dptr = '\0';
	return reasonbuf;
}
#endif

int m_gnotice(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char 		*message = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	IRCD_CLIENT 	*cptr;

	if(!IsAnOper(pptr) && !IsServer(pptr) && MyClient(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if(!(pptr->c_oline_flags & OLINE_FLAGS_CANGNOTICE) && MyClient(pptr))
	{
		send_prefix_to_client(meptr, pptr, "NOTICE %s :You are not allowed to GNOTICE.", pptr->name);
		return 3;
	}

	for(cptr=Locals;cptr;cptr=cptr->c_lo_next)
	{
		if(!IsServer(cptr)) 
		{
			if (IsClient(cptr))
				send_prefix_to_client(pptr, cptr, "NOTICE %s :[%s@%s] %s", cptr->name, pptr->name, NetName, message);
			else
				send_to_client(cptr, ":%s NOTICE %s :[%s@%s]) %s", prefix, cptr->name, pptr->name, NetName, message);
		}
	}

	send_to_servers(IsServer(lptr) ? lptr : NULL, ":%s GNOTICE :%s", prefix, message);

	return 3;
}

static int _do_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *username, char *hostname, char *servername, char *ircname)
{
	if (IsLocal(pptr))
	{
#ifndef NO_DEFAULT_INVISIBLE
		if (!(pptr->c_umodes & CLIENT_UMODES_IMODE))
		{
			pptr->c_nflags |= CLIENT_NFLAGS_INVISIBLE;
			pptr->c_umodes |= CLIENT_UMODES_IMODE;
		}
#endif
		if(vars_value_get_bynum(IRCD_VAR_AUTH_CHECK))
		{
			if (!(pptr->c_flags & CLIENT_FLAGS_GOTAUTH))
			{
				/* other rl_*name fields are in conf.c */
#ifdef HMC
				strcpy(pptr->username, "U+~");
				support_strncpy(pptr->username+3, username, sizeof(pptr->username)-3);
#else
				*pptr->username = '~';
				support_strncpy(pptr->username+1, username, sizeof(pptr->username)-1);
#endif
				/*1 *pptr->rl_username = '~';
				support_strncpy(pptr->rl_username+1, username, sizeof(pptr->username)-1); */
			}
		}
		else
		{
#ifdef HMC
			strcpy(pptr->username, "U+");
			support_strncpy(pptr->username+2, username, sizeof(pptr->username)-2);
#else
			support_strncpy(pptr->username, username, sizeof(pptr->username));
#endif
			strcpy(pptr->rl_username, username);
		}

		support_strncpy(pptr->servername, meptr->name, sizeof(pptr->servername));
	}
	else
	{
		support_strncpy(pptr->username, username, sizeof(pptr->username));
		support_strncpy(pptr->hostname, hostname, sizeof(pptr->hostname));
		support_strncpy(pptr->servername, servername, sizeof(pptr->servername));
	}

	support_strncpy(pptr->ircname, ircname, sizeof(pptr->ircname));
	pptr->c_flags |= CLIENT_FLAGS_GOTUSER;

	if (pptr->name[0])
		return _register_user(lptr, pptr);
	return 3;
}

static int _lusers_send(IRCD_CLIENT *pptr, char *pattern, int sendcounter)
{
	int		g_opers = 0;
	int		g_clients = 0;
	int		g_invis = 0;
	int		g_servers = 0;
	int		m_servers = 0;
	int		m_clients = 0;
	int		m_unknown = 0;
	int		channels = 0;

	channels = Globalinfo->channels;
	if (!pattern || ((*pattern == '*') && (*(pattern+1) == '\0')))
	{
		g_opers = Globalinfo->g_opers;
		g_clients = Globalinfo->g_clients;
		g_invis = Globalinfo->g_invis;
		g_servers = Globalinfo->g_servers;
		m_servers = Globalinfo->m_servers;
		m_clients = Globalinfo->m_clients;
		m_unknown = Globalinfo->m_unknown;
	}
	else
	{
		int 		i = 0;
		IRCD_CLIENT	**list = Opers;
		IRCD_CLIENT	*acptr;

		for(;i<3;i++)
		{
			if (i==1)
				list = Clients;
			else if (i==2)
				list = Servers;
			for(acptr=*list;acptr;acptr=acptr->c_os_next)
			{
				if (((list != Servers) &&
							match(pattern, acptr->servername)) ||
					((list == Servers) && match(pattern, acptr->name)))
					continue;
				if ((list == Opers) || IsOper(acptr))
					g_opers++;
				if (list == Servers)
				{
					if (IsLocal(acptr))
						m_servers++;
					g_servers++;
				}
				else
				{
					if (IsLocal(acptr))
						m_clients++;
					g_clients++;
					if (IsInvisible(acptr))
						g_invis++;
				}
			}
		}
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSERCLIENT, pptr->name),
			g_clients - g_invis, g_invis, g_servers);
	if (g_opers)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSEROP, pptr->name),
			g_opers);
	if (m_unknown)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSERUNKNOWN, pptr->name),
			m_unknown);
	if (channels)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSERCHANNELS, pptr->name),
			channels);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_LUSERME, pptr->name),
		m_clients, m_servers);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_STATSCONN, pptr->name),
		Globalinfo->m_maxconn, Globalinfo->m_maxclients, support_ctime(Globalinfo->m_maxclientstime));
	if (Globalinfo->g_serversbursting)
		send_prefix_to_client(meptr, pptr, "NOTICE %s :%i server%s still bursting.",
			pptr->name, Globalinfo->g_serversbursting,
			Globalinfo->g_serversbursting == 1 ? "" : "s");
	if (sendcounter && vars_value_get_bynum(IRCD_VAR_SHOW_CLIENTCOUNTER))
		send_prefix_to_client(meptr, pptr, "NOTICE %s :%i client%s ha%s connected since %s",
			pptr->name, Globalinfo->m_totclients,
			Globalinfo->m_totclients == 1 ? "" : "s",
			Globalinfo->m_totclients == 1 ? "s" : "ve",
			support_ctime(Globalinfo->m_totclientstime));

	return 0;
}

static int _motd_send(IRCD_CLIENT *pptr)
{
	struct tm	tmbuf;
	char		**lines;

	if (Motd == NULL)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOMOTD, pptr->name));
		return 0;
	}
	tmbuf = motd_time(Motd);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_MOTDSTART, pptr->name),
				meptr->name);
	send_prefix_to_client(meptr, pptr, "%d %s :- %d/%d/%d %d:%02d",
		RPL_MOTD, pptr->name, tmbuf.tm_mday, tmbuf.tm_mon+1, 1900+tmbuf.tm_year,
		tmbuf.tm_hour, tmbuf.tm_min);
	for(lines=motd_lines(Motd);*lines;lines++)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_MOTD, pptr->name),
			*lines);
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFMOTD, pptr->name));
	return 0;
}

#if 0
int _hunt_server(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *format, int server, char *prefix, char *argptr)
{
	char		theargs[512];
	char		*args = theargs;
	char		*argss[10];
	int			numargs;
	IRCD_CLIENT	*acptr;

	if ((--server < 0) || (server > 7))
		return HUNTED_ISME;

	support_strncpy(theargs, argptr, sizeof(theargs));

	for(numargs=0;numargs < 8;numargs++)
		if (!(argss[numargs] = ircd_nextarg(&args, 0)))
		{
			while(numargs < 8)
				argss[numargs++] = NULL;
			break;
		}

	if (!argss[server] || !match(meptr->name, argss[server]) ||
			!match(argss[server], meptr->name))
		return HUNTED_ISME;
	if ((acptr = hash_find_client_incl_hm(argss[server])) != NULL)
		if (acptr->c_from == pptr->c_from && !IsLocal(acptr))
			acptr = NULL;
	if (!acptr)
	{
		for(acptr=*Servers;acptr;acptr=acptr->c_os_next)
		{
			if ((acptr->c_from == pptr->c_from) && IsLocal(acptr))
				continue;
			if (!match(acptr->name, argss[server]) ||
					!match(argss[server], acptr->name))
				break;
		}
	}
	if (acptr)
	{
		if (IsMe(acptr) || MyClient(acptr))
			return HUNTED_ISME;
		if (match(acptr->name, argss[server]))
			argss[server] = acptr->name;
		send_to_client(acptr, format, prefix, argss[0],
				argss[1], argss[2], argss[3], argss[4],
				argss[5], argss[6], argss[7]);
		return HUNTED_PASS;
	}
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHSERVER, pptr->name),
			argss[server]);
	return HUNTED_NOSUCH;
}
#endif

#if 0
static void _sendnick(IRCD_CLIENT *cptr, IRCD_CLIENT *acptr)
{
	static	char	ubuf[64];

	if (IsReg(acptr))
	{
		_umode_build(acptr, ubuf, SEND_UMODES);
		send_to_client(cptr, "NICK %s %d %ld %s %s %s %s :%s",
			acptr->name, acptr->c_hopcount + 1, acptr->c_creation, ubuf,
			acptr->username, acptr->hostname,
			acptr->servername, acptr->ircname);
    }
}
#endif
IRCD_MODE *_ircdmode_create(IRCD_MODE **ircdmodes, IRCD_MODE **lastircdmode)
{
	IRCD_MODE	*nircdmode;


	nircdmode = (IRCD_MODE *)MyMalloc(sizeof(IRCD_MODE));
	if (!*lastircdmode)
		*ircdmodes = nircdmode;
	else
		(*lastircdmode)->next = nircdmode;
	return (*lastircdmode = nircdmode);
}

void _ircdmode_free(IRCD_MODE **bucket)
{
	IRCD_MODE	*nextp;

	for(;*bucket;*bucket = nextp)
	{
		nextp = (*bucket)->next;
		MyFree(*bucket);
	}
}
#if 0
static char *_make_nick_user_host(char *nick, char *user, char *host, char *buf, char **hostptr)
{
	static char	star = '*';
	char	*ptr = buf;
	char	*sptr;
	int		n;

	if (!nick || !*nick)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=nick,n=NICKLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr++ = '!';
	if (!user || !*user)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=user,n=USERLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr++ = '@';
	if (hostptr)
		*hostptr = ptr;
	if (!host || !*host)
	{
		*ptr++ = star;
	}
	else
	{
		for(sptr=host,n=HOSTLEN;*sptr && n--;)
			*ptr++ = *sptr++;
	}
	*ptr = '\0';
	return buf;
}

static char *datestr(int thetime, char *buf)
{
	struct tm	gmbuf;
	struct tm	*tmptr;
	time_t		temptime;
	char		plusminus;
	int			minswest;

	if (!thetime)
		thetime = (int)time(NULL);
	temptime = (time_t)thetime;
	tmptr = gmtime(&temptime);
	if (tmptr)
		memcpy(&gmbuf, tmptr, sizeof(struct tm));
	else
	{
		strcpy(buf, "Error getting date");
		return buf;
	}
	tmptr = localtime(&temptime);

	if (tmptr->tm_yday == gmbuf.tm_yday)
		minswest = (gmbuf.tm_hour - tmptr->tm_hour) * 60 +
			(gmbuf.tm_min - tmptr->tm_min);
	else if (tmptr->tm_yday > gmbuf.tm_yday)
		minswest = (gmbuf.tm_hour - (tmptr->tm_hour + 24)) * 60;
	else
		minswest = ((gmbuf.tm_hour + 24) - tmptr->tm_hour) * 60;

	if (minswest > 0)
		plusminus = '-';
	else
	{
		plusminus = '+';
		minswest = -minswest;
	}

	sprintf(buf, "%s %s %d %04d -- %02d:%02d %c%02d:%02d",
		weekdays[tmptr->tm_wday], months[tmptr->tm_mon], tmptr->tm_mday,
		tmptr->tm_year + 1900, tmptr->tm_hour, tmptr->tm_min,
		plusminus, minswest/60, minswest%60);
	return buf;
}
#endif
static void _count_servers_and_users(IRCD_CLIENT *acptr, int *numservers, int *numusers)
{
	IRCD_CLIENT	*u;

	for(u=acptr->c_servers;u;u=u->c_u_next)
	{
		_count_servers_and_users(u, numservers, numusers);
	}
	for(u=acptr->c_users;u;u=u->c_u_next)
		(*numusers)++;
	(*numservers)++;
}

/*
** Can only come from local unknown connections now..
*/
int m_user(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*username = ircd_nextarg(&args, 0);
	char	*hostname = ircd_nextarg(&args, 0);
	char	*servername = ircd_nextarg(&args, 0);
	char	*ircname = ircd_nextarg(&args, 0);
	char	cnamebuf[CLIENTNAMESIZE];

	if (IsServer(lptr))
	{
		send_to_opers("Got USER from %s", ircd_clientname(lptr, 0, cnamebuf));
		return 3;
	}
	if (!ircname)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS, *pptr->name ? pptr->name : "*"), "USER");
		return 3;
	}
	if (pptr->c_flags & CLIENT_FLAGS_GOTUSER)
		return 3;
	hostname = pptr->hostname;
	servername = meptr->name;
	return _do_user(lptr, pptr, username, hostname, servername, ircname);
}

/*
** m_nick
**  parv[0] = sender prefix
**  parv[1] = nickname
**  parv[2] = optional hopcount when new user; TS when nick change
**  parv[3] = optional TS
**  parv[4] = optional umode
**  parv[5] = optional username
**  parv[6] = optional hostname
**  parv[7] = optional server
**  parv[9] = optional idle time
**  parv[8] = optional ircname
*/
int m_nick(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char		newnick[NICKLEN+1];
	char		*nick 		= ircd_nextarg(&args, 0);
	char		*hopcount 	= ircd_nextarg(&args, 0);
	char		*ts 		= ircd_nextarg(&args, 0);
	char		*umode 		= ircd_nextarg(&args, 0);
	char		*username 	= ircd_nextarg(&args, 0);
	char		*hostname 	= ircd_nextarg(&args, 0);
	char		*servername 	= ircd_nextarg(&args, 0);
	char		*ircname 	= ircd_nextarg(&args, 0);
	int		newts 	= 0;
	int		hc 	= 0;
	IRCD_CLIENT	*acptr;
	char		cnamebuf[CLIENTNAMESIZE];

	if (!nick)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NONICKNAMEGIVEN,
				*pptr->name ? pptr->name : "*"));
		return 3;
	}

	if (IsServer(lptr))
	{
		if (IsServer(pptr)) /* remote new nick */
		{
			if (!ircname)
			{
				send_to_opers("No ircname for new nick \"%s\" from remote \"%s\"",
					nick, pptr->name);
				return 3;
			}
			send_to_servers(lptr, ":%s NICK %s %d %s %s %s %s %s :%s",
				prefix, nick, atoi(hopcount)+1, ts,
				umode, username, hostname, servername, ircname);
			hc = atol(hopcount);
			newts = atol(ts);
		}
		else /* remote nick change */
		{
			send_to_servers(lptr, ":%s NICK %s %s",
				prefix, nick, hopcount?hopcount:"0");
			if (hopcount)
				newts = atol(hopcount);
			username = pptr->username;
			hostname = pptr->hostname;
		}
	}
	support_strncpy(newnick, nick, sizeof(newnick));
	{
		register char *ptr = newnick;

		while(*ptr && (*ptr != '~'))
			ptr++;
		*ptr = '\0';
	}
	if (!nickname_check(newnick) ||
			(IsServer(lptr) && strcmp(nick, newnick)))
	{
		if (IsServer(lptr))
		{
			/* send kill for new nickname everywhere */
			send_to_opers("Bad Nick: %s From: %s %s",
				nick, pptr->name, ircd_clientname(lptr, 0, cnamebuf));
			return 3;
		}
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_ERRONEUSNICKNAME,
					pptr->name), nick);
		return 3;
	}

	/* check to see if nick can be used */

	if (!IsServer(lptr) &&
		(!pptr->c_class ||
			!(pptr->c_class->flags & CLASS_FLAGS_BYPASSNICKNAMEDENY)) &&
		stringdeny_match(&_ircd_conf.nickname_denies, newnick, Now) && !nick_isdenied(newnick))
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_ERRONEUSNICKNAME,
				pptr->name), newnick);
		return 0;
	}

	/* check to see if nick exists */

	acptr = hash_find_client(newnick);
	if (acptr)
	{
		if (acptr == pptr) /* same nick */
		{
			if (!strcmp(acptr->name, newnick))
				return 3; /* trying to change to same nick */
			/* case change */
		}
		else if (IsLocal(acptr) && !IsReg(acptr))
		{
			/* get rid of client and fall through */
			client_del(meptr, acptr, "Overridden");
			acptr = NULL;
		}
		else if (!IsServer(lptr))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NICKNAMEINUSE,
					*pptr->name ? pptr->name : "*"), newnick);
			return 3;
		}
		else /* from remote */
		{
			int				sameuser;
			IRCD_CLIENT		*toexit;
	
			if (!newts || (newts == acptr->c_creation))
			{
				/* kill both */
				if (IsLocal(acptr))
				{
					send_prefix_to_client(meptr, acptr,
						IRCD_REPLY(ERR_NICKCOLLISION, acptr->name),
						acptr->name);
					acptr->c_flags |= CLIENT_FLAGS_NOSENDQUIT;
					acptr->c_flags |= CLIENT_FLAGS_KILLED;
				}
				client_del(meptr, acptr, "Nick collision");
				return (acptr == lptr) ? -1 : 3;
			}
			sameuser = (!mycmp(acptr->username, username) &&
							!mycmp(acptr->hostname, hostname));
			if ((sameuser && (newts < acptr->c_creation)) ||
				(!sameuser && (newts > acptr->c_creation)))
			{
				if (IsServer(pptr))
				{
					/* ignore the new NICK */
					return 3;
				}
				toexit = pptr;
			}
			else
			{
				toexit = acptr;
			}
			if (sameuser)
			{
				send_to_flag(CLIENT_NFLAGS_SKILLS, NULL,
					"Nick collision on %s(%s <- %s)(older killed)",
					acptr->name, acptr->c_from->name,
					ircd_clientname(lptr, 0, cnamebuf));
			}
			else
			{
				send_to_flag(CLIENT_NFLAGS_SKILLS, NULL,
					"Nick collision on %s(%s <- %s)(newer killed)",
					acptr->name, acptr->c_from->name,
					ircd_clientname(lptr, 0, cnamebuf));
			}

			/* kill off 'toexit' */
			if (IsLocal(toexit))
			{
				send_prefix_to_client(meptr, toexit,
					IRCD_REPLY(ERR_NICKCOLLISION, toexit->name), toexit->name);
				toexit->c_flags |= CLIENT_FLAGS_NOSENDQUIT;
				toexit->c_flags |= CLIENT_FLAGS_KILLED;
			}
			client_del(meptr, toexit, "Nick collision!");
			if (toexit == pptr)
				return 3; /* pptr is not local */
			acptr= NULL;
			/* fall through */
		}
	}

	if (IsServer(pptr)) /* new nick from remote server */
	{
		IRCD_CLIENT	*nptr;
		int		*s;
		int		flag;
		char	*m;

#if 1
		static int	nsize = offsetof(IRCD_CLIENT, c_auth_ptr);
#else
		static int	nsize = sizeof(IRCD_CLIENT);
#endif

		nptr = (IRCD_CLIENT *)MyMalloc(nsize);
		nptr->c_from = lptr;
		nptr->c_rem_ipaddr = -1;
		nptr->c_rem_port = 0;
		nptr->c_hopcount = hc;
		strcpy(nptr->name, newnick);
		support_strncpy(nptr->username, username, sizeof(nptr->username));
		support_strncpy(nptr->hostname, hostname, sizeof(nptr->hostname));
		nptr->c_creation = newts ? newts : (newts = Now);
		nptr->c_lasttime = nptr->c_firsttime = Now;
		m = umode + 1;
		for(;*m;m++)
		{
			for(s=_user_modes;(flag = *s);s+=2)
			{
				if (*m == *(s+1))
				{
					nptr->c_umodes |= flag;
					break;
				}
			}
		}
		hash_add_client(nptr);
		return _do_user(lptr, nptr, username, hostname, servername, ircname);
	}
	else if (pptr->name[0])
	{
		/* nick change */

		pptr->c_creation = newts ? newts : (newts = Now);
		if (IsReg(pptr)) /* send NICK +chan stuff to other servers */
		{
			whowas_add_history(pptr, 1);
			send_to_common_channels(pptr, "NICK :%s", newnick);
			if (IsLocal(pptr))
			{
				if (vars_value_get_bynum(IRCD_VAR_SHOW_NICKCHANGES))
					send_to_flag(CLIENT_NFLAGS_NICKCHANGES, NULL,
						"Nick change: From %s to %s [%s@%s]",
							prefix, newnick, pptr->username,
								pptr->hostname);
				send_to_servers(NULL, ":%s NICK %s :%i",
					pptr->name, newnick, pptr->c_creation);
			}
		}
		hash_del_client(pptr);
		strcpy(pptr->name, newnick);
		hash_add_client(pptr);
		return 3;
	}
	strcpy(pptr->name, newnick);
	pptr->c_creation = Now;
	hash_add_client(pptr);
	if (pptr->c_flags & CLIENT_FLAGS_GOTUSER)
		return _register_user(lptr, pptr);
	return 3;
}

#if 0
int m_version(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (!IsServer(lptr) && !IsReg(lptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_VERSION,
				pptr->name), VERSION, "", meptr->name, "http://darkircd.sourceforge.net");
		client_del(pptr, pptr, "Leaving");
		return -1;
	}

	if (_hunt_server(lptr, pptr, ":%s VERSION :%s", 1, prefix, args) == HUNTED_ISME)
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_VERSION,
				pptr->name), VERSION, "", meptr->name, "http://darkircd.sourceforge.net");
	return 3;
}
#endif

int m_quit(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*comment = ircd_nextarg(&args, 0);

	if (IsServer(pptr))
		return 3;
	if (!comment ||
		((lptr == pptr) && (!strncmp("Local Kill", comment, 10) ||
				!strncmp(comment, "Killed", 6))))
			comment = lptr->name;
	if (strlen(comment) > QUITLEN)
		comment[QUITLEN] = '\0';
	client_del(pptr, pptr, comment);
	if (pptr == lptr)
		return -1;
	return 3;
}

int m_kill(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	cnamebuf[CLIENTNAMESIZE];
	char	*inpath = ircd_clientname(lptr, 0, cnamebuf);
	char	*victim = ircd_nextarg(&args, 0);
	char	*comment = ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	char	*commentsave;
	char	buf[512];
	char	buf2[512];
	IRCD_CLIENT	*acptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!victim || !*victim)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "KILL");
		return 3;
	}

	if (!comment || !*comment)
		comment = pptr->name;

	commentsave = comment;

	if (!(IsServer(lptr) || IsAnOper(lptr)))
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (IsAnOper(lptr))
	{
		if (strlen(comment) > KILLLEN)
			comment[KILLLEN] = '\0';
	}

	if (!(acptr = hash_find_chasing(victim)))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOSUCHNICK, pptr->name), victim);
		return 3;
	}

	if (!IsLocal(acptr) && (lptr->c_umodes & CLIENT_UMODES_LOCOP))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (IsService(acptr) && !IsServer(lptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_ITSASERVICE, pptr->name),
					acptr->name, acptr->name);
		return 3;
	}

	if (IsServer(acptr) || (acptr == meptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_CANTKILLSERVER, pptr->name));
		return 3;
	}

	if (MyClient(pptr))
	{
		if (IsLocal(acptr) &&
					!(pptr->c_oline_flags & OLINE_FLAGS_CANLOCALKILL))
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :You are not allowed to LOCAL KILL", pptr->name);
			return 3;
		}
		if (!IsLocal(acptr) &&
					!(pptr->c_oline_flags & OLINE_FLAGS_CANGLOBALKILL))
		{
			send_prefix_to_client(meptr, pptr,
				"NOTICE %s :You are not allowed to GLOBAL KILL", pptr->name);
			return 3;
		}
	}
	if (!IsServer(lptr))
	{
		/* kill came from oper on this server */
		inpath = lptr->hostname;
		sprintf(buf, "%s%s (%s)",
			lptr->name, IsOper(pptr) ? "" : "(L)", comment);
		comment = buf;
	}

	if (IsServer(pptr))
	{
		send_to_flag(CLIENT_NFLAGS_SKILLS, NULL,
				"Received KILL message for %s. From %s Path: %s!%s",
				acptr->name, pptr->name, inpath, comment);
	}
	else
	{
		send_to_flag(CLIENT_NFLAGS_OKILLS, NULL,
				"Received KILL message for %s. From %s Path: %s!%s",
				acptr->name, pptr->name, inpath, comment);
	}

	if (!IsLocal(acptr) || !IsLocal(pptr) || !IsAnOper(pptr))
	{
		send_to_servers(lptr, ":%s KILL %s :%s!%s",
			prefix, acptr->name, inpath, comment);
		acptr->c_flags |= CLIENT_FLAGS_NOSENDQUIT;
	}

	if (IsLocal(acptr))
	{
		acptr->c_flags |= CLIENT_FLAGS_KILLED;

		if (IsClient(pptr))
			send_prefix_to_client(pptr, acptr, "KILL %s :%s!%s",
				acptr->name, inpath, comment);
		else
			send_to_client(acptr, ":%s KILL %s :%s!%s",
				prefix, acptr->name, inpath, comment);
	}

	if (IsLocal(acptr) && IsLocal(pptr) && IsAnOper(pptr))
		sprintf(buf2, "Local kill by %s (%s)", pptr->name, commentsave);
	else
	{
		char	*killer;

		if ((killer = strchr(comment, ' ')) != NULL)
		{
			while (*killer && (*killer != '!'))
				killer--;
			if (!*killer)
				killer = comment;
			else
				killer++;
		}
		else
			killer = comment;
		sprintf(buf2, "Killed (%s)", killer);
	}
/* above already
	acptr->c_flags |= CLIENT_FLAGS_KILLED;
*/
	client_del(pptr, acptr, buf2);
	return (acptr == lptr) ? -1 : 3;
}

#if 0
int m_stats(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_CLIENT		*sptr;
	IRCD_CLIENT		*acptr;
	IRCD_LISTENER	*liptr;
	char			cnamebuf[CLIENTNAMESIZE];
	char			*thing;
	char			*server;
	char			*rest;
	OLINE			*oline;
	SERVLINK		*servlink;
	int				now;
	CONFHOSTIPENTRY	*chentry;
	IRCD_CLASSENTRY	**ceptr;
	int				i;
	static char 	LFormat[] = "%d %s %s %u %u %u %u %u :%u";
	static int		ratenum = 0;
	static int		ratetime = 0;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (_hunt_server(lptr, pptr, ":%s STATS %s :%s", 2, prefix, args) != HUNTED_ISME)
	{
		return 6; /* remote queries are penalized more */
	}

	thing = ircd_nextarg(&args, 0);
	server = ircd_nextarg(&args, 0);
	rest = ircd_nextarg(&args, 0);

	if (!thing || !*thing)
	{
		send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "STATS");
		return 3;
	}

	send_to_flag(CLIENT_NFLAGS_STATSNOTICES, NULL,
			"STATS %s%c requested by %s (%s@%s) [%s]",
			*thing < ' ' ? "^" : "",
			(*thing) < ' ' ? (*thing)+0x40 : *thing,
			pptr->name, pptr->username, pptr->hostname,
			pptr->servername);

#define Meg 1024
#define Gig (Meg*1024)
#define Tera (Gig*1024)
#define IsMeg(x) (x>Meg)
#define IsGig(x) (x>Gig)
#define IsTera(x) (x>Tera)
#define ByteString(x) (IsTera(x) ? "Tb " : \
            IsGig(x) ? "Gig" : \
            IsMeg(x) ? "Meg" : \
            "KB ")
#define ByteConvert(x) (IsTera(x) ? (float)((float)x/(float)Tera) : \
            IsGig(x) ? (float)((float)x/(float)Gig) : \
            IsMeg(x) ? (float)((float)x/(float)Meg) : \
            (float)x)

	acptr = NULL;
	if (server && *server)
	{
		acptr = hash_find_chasing(server);
	}

	if ((*thing == 'L') || (*thing == 'l'))
	{
		if (!IsAnOper(pptr) && (!acptr || !IsClient(acptr)))
			*thing = '?';
	}

	if ((*thing != '?') && !IsLocal(pptr) && !IsAnOper(pptr) && ircd_ratelimit(pptr, &ratenum, &ratetime, (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_STATS_NUM), (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_TIME)))
		return 0;

	if (!strcasecmp(thing, "rate"))
	{
		int itotal = Globalinfo->bw_c_incoming_rate + Globalinfo->bw_s_incoming_rate;
		int ototal = Globalinfo->bw_c_outgoing_rate + Globalinfo->bw_s_outgoing_rate;

		send_prefix_to_client(meptr, pptr,
				"%d %s :Current Bandwidth Statistics (in bytes/second):",
					RPL_STATSDEBUG, pptr->name);
		send_prefix_to_client(meptr, pptr,
				"%d %s :Clients --  Incoming: %6i     Outgoing: %6i",
					RPL_STATSDEBUG, pptr->name, Globalinfo->bw_c_incoming_rate,
					Globalinfo->bw_c_outgoing_rate);
		send_prefix_to_client(meptr, pptr,
				"%d %s :Servers --  Incoming: %6i     Outgoing: %6i",
					RPL_STATSDEBUG, pptr->name,
					Globalinfo->bw_s_incoming_rate,
					Globalinfo->bw_s_outgoing_rate);
		send_prefix_to_client(meptr, pptr,
				"%d %s :  Total --  Incoming: %6i     Outgoing: %6i",
					RPL_STATSDEBUG, pptr->name, itotal, ototal);
	}
	else if (!strcasecmp(thing, "uline"))
	{
		IRCD_ULINE      **Ulines = &(_ircd_conf.ulines);
                IRCD_ULINE      *ul;

                for(ul=*Ulines;ul;ul=ul->next)
                {
                        send_prefix_to_client(meptr, pptr, "%d %s :%s",
                                                RPL_STATSULINE, pptr->name, ul->servername);
                }
	}
        else if (!strcasecmp(thing, "whois"))
        {
                IRCD_WL      **wlevels = &(_ircd_conf.wlevels);
                IRCD_WL      *wl;
                                   
                for(wl=*wlevels;wl;wl=wl->next)
                {
                        send_prefix_to_client(meptr, pptr, "%d %s :(%i) %s",
                                                RPL_STATSDEBUG, pptr->name, wl->number, wl->text);
                }
        }    
	else switch(*thing)
	{
		case '?':
			for(sptr=*Servers;sptr;sptr=sptr->c_os_next)
			{
				if (!IsLocal(sptr))
					continue;
				send_prefix_to_client(meptr, pptr, LFormat,
						RPL_STATSLINKINFO, pptr->name,
						IsAnOper(pptr) ? ircd_clientname(sptr, 1, cnamebuf) :
						ircd_clientname(sptr, 0, cnamebuf),
						sptr->c_queue_out.queue_size,
						sptr->c_send_msg,
						sptr->c_send_kbytes,
						sptr->c_recv_msg,
						sptr->c_recv_kbytes,
						Now - sptr->c_firsttime);
			}
			break;
		case 'C':
		case 'c':
			for(servlink=_ircd_conf.servers;servlink;servlink=servlink->next)
			{
				static char hidden[] = "127.0.0.1";
				char		flagsbuf[64];
				char		*ptr;
				char		*hname;

				ptr = flagsbuf;

				if (servlink->flags & SERVLINK_FLAGS_AUTOCONN)
				{
					if (ptr != flagsbuf)
						*ptr++ = ' ';
					ptr += sprintf(ptr, "autoconn");
				}
#if 0
				else if (servlink->flags & ... )
				{
					if (ptr != flagsbuf)
						*ptr++ = ' ';
					ptr += sprintf(ptr, "newflag");
				}
#endif

				*ptr = '\0';

				if ( /* !IsAnOper(pptr) && */
						(servlink->flags & SERVLINK_FLAGS_HIDEHOSTNAME))
					hname = hidden;
				else
					hname = servlink->hostname;

				send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSCLINE, pptr->name),
						'C', servlink->username ? servlink->username : "*",
						hname, servlink->servername,
						servlink->port, servlink->classname, flagsbuf);
				send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSNLINE, pptr->name),
						'N', servlink->username ? servlink->username : "*",
						hname, servlink->servername,
						servlink->hostmask, servlink->classname, flagsbuf);
			}
			break;
		case 'G':
		case 'g':
			show_glines(pptr, NULL);
			break;
		case 'H':
		case 'h':
			for(servlink=_ircd_conf.servers;servlink;servlink=servlink->next)
			{
				IRCD_STRINGLIST *stptr;

				for(stptr=servlink->hubs;stptr;stptr=stptr->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSHLINE, pptr->name),
						'H', stptr->string, servlink->servername, 0, -1);
				}
				for(stptr=servlink->leafs;stptr;stptr=stptr->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSLLINE, pptr->name),
						'L', stptr->string, servlink->servername, 0, -1);
				}
			}
			break;
		case 'I':
		case 'i':
			for(chentry=_ircd_conf.allow.conf_all;chentry;chentry=chentry->ch_next)
			{
				char *iname = NULL;
				char *hname = NULL;
				IRCD_CLASSENTRY	*classptr;

				if (chentry->classptr)
					classptr = (IRCD_CLASSENTRY *)chentry->classptr;
				else
				{
					int ret = class_find(&(_ircd_conf.classes),
							chentry->classname, &classptr);
					if (ret < 0)
						classptr = NULL;
					else
						chentry->classptr = (void *)classptr;
				}

				if (classptr && (classptr->flags & CLASS_FLAGS_HIDEHOSTNAME))
					continue;

				if (chentry->ipstring)
				{
					if (chentry->username)
					{
						iname = (char *)MyMalloc(strlen(chentry->username)+
										strlen(chentry->ipstring) + 2);
						sprintf(iname, "%s@%s", chentry->username,
										chentry->ipstring);
					}
					else
						iname = MyStrdup(chentry->ipstring);
				}
				if (chentry->hostname)
				{
					if (chentry->username)
					{
						hname = (char *)MyMalloc(strlen(chentry->username)+
										strlen(chentry->hostname) + 2);
						sprintf(hname, "%s@%s", chentry->username,
										chentry->hostname);
					}
					else
						hname = MyStrdup(chentry->hostname);
				}

				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_STATSILINE, pptr->name),
					'I', iname ? iname : "X", hname ? hname : "X",
					0, chentry->classname);

				if (iname)
					MyFree(iname);
				if (hname)
					MyFree(hname);
			}
			break;
#ifdef KLINES
		case 'K':
		case 'k':
		{
			char *user = NULL;
			char *host = NULL;
			static char _star[] = "*";
			int	numfound = 0;

			if (!IsOper(pptr) && (!rest || !*rest))
			{
				user = pptr->username;
				host = pptr->hostname;
			}
			else if (rest && *rest)
			{
				acptr = hash_find_client(rest);
				if (!acptr || IsServer(acptr))
				{
					user = host = rest;
					while(*host && (*host != '@'))
						host++;
					if (!*host)
					{
						host = rest;
						user = _star;
					}
					else
					{
						*host++ = '\0';
						if (!*host)
							host = _star;
					}
					if (!*user)
						user = _star;
				}
				else
				{
					user = acptr->username;
					host = acptr->hostname;
				}
			}
			for(chentry=_ircd_conf.deny.conf_all;chentry;chentry=chentry->ch_next)
			{
				char *iname = NULL;
				char *hname = NULL;

				if (chentry->expires && (chentry->expires < Now))
					continue;

				if (user)
				{
					if (chentry->username &&
							match(chentry->username, user) &&
							match(user, chentry->username))
						continue;
					if (!(((chentry->ipstring &&
						(!match(chentry->ipstring, host) ||
								!match(host, chentry->ipstring))) ||
						(chentry->hostname &&
						(!match(chentry->hostname, host) ||
								!match(host, chentry->hostname))))))
						continue;
				}
				if (chentry->ipstring)
				{
					if (chentry->username)
					{
						iname = (char *)MyMalloc(strlen(chentry->username)+
										strlen(chentry->ipstring) + 2);
						sprintf(iname, "%s@%s", chentry->username,
										chentry->ipstring);
					}
					else
						iname = MyStrdup(chentry->ipstring);
				}
				if (chentry->hostname)
				{
					if (chentry->username)
					{
						hname = (char *)MyMalloc(strlen(chentry->username)+
										strlen(chentry->hostname) + 2);
						sprintf(hname, "%s@%s", chentry->username,
										chentry->hostname);
					}
					else
						hname = MyStrdup(chentry->hostname);
				}

				numfound++;
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_STATSKLINE, pptr->name),
					'K', iname ? iname : "X",
					hname ? hname : "X", chentry->expires,
					chentry->flags,
					chentry->data ? chentry->data : "*");

				if (iname)
					MyFree(iname);
				if (hname)
					MyFree(hname);

				if (user && (numfound == 100))
				{
					send_prefix_to_client(meptr, pptr,
						"NOTICE %s :Maximum of 100 bans reached -- stopping request.",
						pptr->name);
					break;
				}
			}
			if (!numfound)
			{
				send_prefix_to_client(meptr, pptr,
					"NOTICE %s :No bans %s%s%s%s%s%sfound",
					pptr->name,
					user ? "matching" : "", user ? " " : "",
					user ? user : "", user ? "@": "",
					user ? host : "", user ? " " : "");
			}
		}
			break;
#endif
		case 'L':
		case 'l':
			if (acptr && IsClient(acptr))
			{
				if (IsLocal(acptr))
					send_prefix_to_client(meptr, pptr, LFormat,
						RPL_STATSLINKINFO, pptr->name,
						*thing == 'L' ? ircd_clientname(acptr, 1, cnamebuf):
						ircd_clientname(acptr, 0, cnamebuf),
						acptr->c_queue_out.queue_size,
						acptr->c_send_msg, acptr->c_send_kbytes,
						acptr->c_recv_msg, acptr->c_recv_kbytes,
						Now - acptr->c_firsttime);
				break;
			}
			/* IsAnOper(pptr) is TRUE because of check above switch() */

			for(liptr=_ircd_conf.listeners;liptr;liptr=liptr->l_next)
			{
				sprintf(cnamebuf, "%s[*@%s.%d]",
						meptr->name,
						liptr->l_ipstring, liptr->l_port);
				send_prefix_to_client(meptr, pptr, LFormat,
					RPL_STATSLINKINFO, pptr->name, cnamebuf, 0, 0, 0, 0, 0,
					Now - liptr->l_firsttime);
			}
			for(acptr=Locals;acptr;acptr=acptr->c_lo_next)
			{
				if (!IsClient(acptr) && !IsServer(acptr))
					continue;
				send_prefix_to_client(meptr, pptr, LFormat,
					RPL_STATSLINKINFO, pptr->name,
					*thing == 'L' ? ircd_clientname(acptr, 1, cnamebuf):
					ircd_clientname(acptr, 0, cnamebuf),
					acptr->c_queue_out.queue_size,
					acptr->c_send_msg, acptr->c_send_kbytes,
					acptr->c_recv_msg, acptr->c_recv_kbytes,
					Now - acptr->c_firsttime);
			}

			break;
		case 'O':
		case 'o':
			for(oline=_ircd_conf.olines;oline;oline=oline->next)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_STATSOLINE, pptr->name),
					oline->flags & OLINE_FLAGS_LOCAL ? 'o' : 'O', 
#if 0
					oline->username ? oline->username : "*",
					oline->hostname,
#else
					"*", "127.0.0.1",
#endif
					oline->name, 0, oline->classname);
			}
			break;
		case 'U':
		case 'u':
			now = Now - meptr->c_firsttime;
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_STATSUPTIME,
					pptr->name), now/86400, (now/3600)%24, (now/60)%60,
					now % 60);
			break;

		case 'Y':
		case 'y':
			for(ceptr=_ircd_conf.classes.entries,i=_ircd_conf.classes.num;i--;ceptr++)
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_STATSYLINE, pptr->name),
					'Y', (*ceptr)->classname, (*ceptr)->pingfreq,
					(*ceptr)->connfreq, (*ceptr)->maxclients,
					(*ceptr)->maxsendq, (*ceptr)->uhlimit,
					(*ceptr)->hlimit,
					(*ceptr)->floodnum, (*ceptr)->floodtime,
					(*ceptr)->floodexit, (*ceptr)->ipaddr_rate_num,
					(*ceptr)->ipaddr_rate_time);
			}
			break;
		case 'p':
		case 'P':
			i = 0;
			for(acptr=*Opers;acptr;acptr=acptr->c_os_next)
			{
				if (!IsLocal(acptr))
					continue;
				i++;
				send_prefix_to_client(meptr, pptr,
						"%d %s :[%c] %s (%s@%s) idle %is",
						RPL_STATSDEBUG, pptr->name,
						IsOper(acptr) ? 'O' : 'o',
						acptr->name, acptr->username,
						acptr->hostname, Now - acptr->c_idletime);
			}
			if (!i)
				send_prefix_to_client(meptr, pptr, "%d %s :No opers found",
						RPL_STATSDEBUG, pptr->name);
			break;
		case 'Q':
		case 'q':
			if (IsAnOper(pptr))
			{
				STRINGDENY	*sd;

				for(sd=_ircd_conf.nickname_denies;sd;sd=sd->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSQLINE, pptr->name),
						sd->name, "-", "-", sd->expires,
						sd->flags, sd->reason);
				}
				for(sd=_ircd_conf.channel_denies;sd;sd=sd->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSQLINE, pptr->name),
						sd->name, "-", "-", sd->expires,
						sd->flags, sd->reason);
				}
			}
			break;
		case 'X':
		case 'x':
			if (IsAnOper(pptr))
			{
				STRINGDENY	*sd;

				for(sd=_ircd_conf.ircname_denies;sd;sd=sd->next)
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_STATSXLINE, pptr->name),
						sd->name, sd->expires,
						sd->flags, sd->reason);
				}
			}
			break;
		case 'Z':
		case 'z':
		{
			unsigned int total_usage = 0;
			unsigned int iphash_num_nethash;
			unsigned int iphash_num_ips;
			unsigned int iphash_size_nethash;
			unsigned int whowas_num;
			unsigned int whowas_memory;
			unsigned int whowas_array_size;
			int i;
			IRCD_CLIENT	*list = Locals;
			IRCD_CLIENT *nextacptr;
			unsigned int lo_num = 0;
			unsigned int cl_num = 0;
			unsigned int extra_mem = 0;
			unsigned int ch_num = 0;
			unsigned int ch_mem = 0;
			unsigned int ban_num = 0;
			unsigned int ban_mem = 0;
			IRCD_CHANNEL	*chptr;

			iphash_count_memory(&iphash_num_nethash, &iphash_num_ips,
				&iphash_size_nethash);
			total_usage += iphash_num_nethash * sizeof(IRCD_NETHASH);
			total_usage += iphash_num_ips * sizeof(IRCD_IPENTRY);
			total_usage += iphash_size_nethash;

			send_prefix_to_client(meptr, pptr,
				"%d %s :IPHASH: Array: %i NetEntries: %i(%i) IPEntries: %i(%i)",
				RPL_STATSDEBUG, pptr->name,
				iphash_size_nethash, iphash_num_nethash,
				iphash_num_nethash * sizeof(IRCD_NETHASH),
				iphash_num_ips, iphash_num_ips * sizeof(IRCD_IPENTRY));
			whowas_count_memory(&whowas_num, &whowas_memory, &whowas_array_size);
			total_usage += whowas_memory + whowas_array_size;

			send_prefix_to_client(meptr, pptr,
				"%d %s :Whowas: Array: %i Entries: %i(%i)",
				RPL_STATSDEBUG, pptr->name,
				whowas_array_size, whowas_num, whowas_memory);

			for(i=0;i<4;i++)
			{
				if (i==1)
					list = *Clients;
				else if (i==2)
					list = *Opers;
				else if (i==3)
					list = *Servers;
				for(acptr=list;acptr;acptr=nextacptr)
				{
					if (list == Locals)
					{
						nextacptr = acptr->c_lo_next;
						lo_num++;
					}
					else
					{
						nextacptr = acptr->c_os_next;
						if (IsLocal(acptr))
							continue; /* already done */
						if (IsClient(acptr))
							cl_num++;
						else
							lo_num++;
					}
					if (acptr->c_away)
						extra_mem += strlen(acptr->c_away)+1;
					if (IsLocal(acptr) && acptr->c_connectby)
						extra_mem += strlen(acptr->c_connectby)+1;
					if (acptr->c_invites_list)
					{
						IRCD_INVITE	*iptr;
						for(iptr=acptr->c_invites_list;iptr;iptr=iptr->next)
							extra_mem += sizeof(IRCD_INVITE) + strlen(iptr->channel) + 1;
					}
					if (acptr->c_channels)
					{
						IRCD_CHANNELLIST	*chlptr;
						for(chlptr=acptr->c_channels;chlptr;chlptr=chlptr->next)
							extra_mem += sizeof(IRCD_CHANNELLIST) + sizeof(IRCD_CLIENTLIST);
					}
				}
			}
			total_usage += lo_num * sizeof(IRCD_CLIENT);
			total_usage += cl_num * offsetof(IRCD_CLIENT, c_auth_ptr);
			total_usage += extra_mem;
			total_usage += sizeof(Fds);
			send_prefix_to_client(meptr, pptr,
				"%d %s :Clients: Local: %i(%i) Remote: %i(%i) Extra: %i  Fds: %i",
				RPL_STATSDEBUG, pptr->name,
				lo_num, lo_num*sizeof(IRCD_CLIENT), cl_num,
				cl_num * offsetof(IRCD_CLIENT, c_auth_ptr),
				extra_mem, sizeof(Fds));
			for(chptr=Channels;chptr;chptr=chptr->next)
			{
				ch_num++;
				ch_mem += strlen(chptr->name);
				if (chptr->key)
					extra_mem += strlen(chptr->key) + 1;
				if (chptr->topic)
					extra_mem += strlen(chptr->topic) + 1;
				if (chptr->topic_who)
					extra_mem += strlen(chptr->topic_who) + 1;
				if (chptr->bans)
				{
					IRCD_BAN	*b;
					for(b=chptr->bans;b;b=b->next)
					{
						ban_num++;
						if (b->firstpart)
							ban_mem += strlen(b->firstpart) + 1;
						if (b->banstring)
							ban_mem += strlen(b->banstring) + 1;
						if (b->who)
							ban_mem += strlen(b->who) + 1;
					}
				}
			}
			total_usage += ch_mem + ban_mem + (ban_num * sizeof(IRCD_BAN));
			total_usage += extra_mem;
			send_prefix_to_client(meptr, pptr,
				"%d %s :Channels: %i(%i) Bans: %i(%i) Extra: %i",
				RPL_STATSDEBUG, pptr->name, ch_num,
				ch_mem, ban_num, ban_mem + (ban_num * sizeof(IRCD_BAN)),
				extra_mem);
			send_prefix_to_client(meptr, pptr,
				"%d %s :Total Usage: %i",
				RPL_STATSDEBUG, pptr->name, total_usage);

		}
			break;
	}

	send_prefix_to_client(meptr, pptr,
		IRCD_REPLY(RPL_ENDOFSTATS, pptr->name), *thing);

	return 3;
}
#endif /* '0' begin/m_stats */

int m_motd(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
        static int ratenum = 0;
        static int ratetime = 0;

        if (!IsServer(lptr) && !IsReg(pptr))
        {
                send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
                return 3;
        }
        if (_hunt_server(lptr, pptr, ":%s MOTD :%s", 1, prefix, args) != HUNTED_ISME)
        {
                return 6; /* remote info requests from local clients */
        }
        if (!IsLocal(pptr) && !IsAnOper(pptr) && ircd_ratelimit(pptr, &ratenum, &ratetime, (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_MOTD_NUM), (int)vars_value_get_bynum(IRCD_VAR_RATELIMIT_TIME)))
                return 0;
        _motd_send(pptr);
        return 3;
}

int m_nmotd(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_lusers(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{       
        char    *pattern;
                
        if (!IsServer(lptr) && !IsReg(pptr))
        {       
                send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
                return 3;
        }               
                
        if (_hunt_server(lptr, pptr, ":%s LUSERS %s :%s", 2, prefix, args) != HUNTED_ISME)
                return 3;
                        
        pattern = ircd_nextarg(&args, 0);
        if (pattern && !*pattern)
                pattern = NULL;
        
        _lusers_send(pptr, pattern, 1);

        return 3;
}

int m_help(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	return 3;
}

int m_away(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*away = ircd_nextarg(&args, 0);

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}
	else if (IsServer(pptr))
		return 0;

	if (away && !*away)
		away = NULL;
	else if (away && (strlen(away) > AWAYLEN))
		away[AWAYLEN] = '\0';
		
	if (pptr->c_away)
		MyFree(pptr->c_away);
	pptr->c_away = away ? MyStrdup(away) : away;
	if (IsLocal(pptr))
	{
		if (!away)
			send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_UNAWAY, pptr->name));
		else
			send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_NOWAWAY, pptr->name));
	}

	if (!away)
		send_to_servers(lptr, ":%s AWAY", pptr->name);
	else	/* we propogate these now -myr */
		send_to_servers(lptr, ":%s AWAY :%s", pptr->name, away);

	return 3;
}

int m_ping(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*origin = ircd_nextarg(&args, 0);
	char			*destination = ircd_nextarg(&args, 0);
	IRCD_CLIENT		*acptr;

	if (!origin || !*origin)
	{
		send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOORIGIN, pptr->name));
		return 3;
	}
	acptr = hash_find_client_incl_hm(origin);
	if (acptr && acptr != pptr)
		origin = lptr->name;
	if (destination && mycmp(destination, meptr->name))
	{
		acptr = hash_find_client_incl_hm(destination);
		if (!acptr || !IsServer(acptr))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOSUCHSERVER, pptr->name), destination);
		}
		else
		{
			if (IsServer(pptr))
				send_to_client(acptr, ":%s PING %s :%s",
					prefix, origin, destination);
			else
				send_prefix_to_client(pptr, acptr, "PING %s :%s",
					origin, destination);
		}
	}
	else
		send_prefix_to_client(meptr, pptr, "PONG %s :%s", 
			(destination) ? destination : meptr->name, origin);
	return 3;
}

int m_pong(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*origin = ircd_nextarg(&args, 0);
	char			*destination = ircd_nextarg(&args, 0);
	IRCD_CLIENT		*acptr;

	if (!origin || !*origin)
	{
		send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(ERR_NOORIGIN, pptr->name));
		return 3;
	}

	/* ping reset in ircd_parse() */

	if (!IsServer(lptr) && !IsReg(lptr))
		return 3;

	if (IsServer(pptr) && CliSentEOB(pptr) && !CliGotEOB(pptr))
	{
		if (pptr == lptr)
		{
			send_to_opers(
				"End-of-Burst acknowledged from %s",
				lptr->name);
		}
		Globalinfo->g_serversbursting--;
		CliSetGotEOB(pptr);
	}

	if (destination && mycmp(destination, meptr->name))
	{
		acptr = hash_find_client_incl_hm(destination);
		if (acptr)
		{
			if (IsClient(pptr))
				send_prefix_to_client(pptr, acptr, "PONG %s %s",
					origin, destination);
			else
				send_to_client(acptr, ":%s PONG %s %s",
					prefix, origin, destination);
		}
		else
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOSUCHSERVER, pptr->name),
				destination);
		}
	}
	return 3;
}

int m_oper(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*name = ircd_nextarg(&args, 0);
	char	*password = ircd_nextarg(&args, 0);
	OLINE	*oline;
	int		oldmodes;
	int 	ret;
	char	 *encr;
	char	buf[64];
	extern int UnderSVSNOOP;
	IRCD_CLASSENTRY	*classptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (IsServer(pptr) || IsServer(lptr)) /* eh ? */
		return 3;

	if (!password || !*password)
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "OPER");
		return 3;
	}

	if (IsOper(pptr))
	{
		if (IsLocal(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_YOUREOPER, pptr->name));
		return 3;
	}

	oldmodes = pptr->c_umodes;

	ret = oline_find(&(_ircd_conf.olines), name, pptr->username,
			pptr->hostname, pptr->c_rem_ipaddr, &oline);

	if (IsSpoofed(pptr) && ret)
		ret = oline_find(&(_ircd_conf.olines), name, pptr->rl_username,
				pptr->rl_hostname, pptr->c_rem_ipaddr, &oline);

	if (ret)
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOOPERHOST, pptr->name));
		return 3;
	}

#ifdef CRYPT_OPER_PASSWORD

	if (((encr = crypt(password, oline->passwd)) == NULL) ||
			strcmp(encr, oline->passwd))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_PASSWDMISMATCH, pptr->name));
		send_to_opers("Failed OPER attempt: %s (%s@%s) [%s]",
				pptr->name, pptr->username, pptr->hostname, name);

		if (vars_value_get_bynum(IRCD_VAR_BADOPER_LOG))
			log_printf(_ircd_conf.log_badoper, "%s %s!%s@%s tried [%s:%s]\n",
                    support_ctime(Now), pptr->name, pptr->username,
					pptr->hostname, oline->name, password);

		return 3;
	}
#else
	if (strcmp(password, oline->c_passwd))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_PASSWDMISMATCH, pptr->name));
		send_to_opers("Failed OPER attempt: %s (%s@%s) [%s]",
				pptr->name, pptr->username, pptr->hostname, name);
		if (vars_value_get_bynum(IRCD_VAR_BADOPER_LOG))
			log_printf(_ircd_conf.log_badoper, "%s %s!%s@%s tried [%s:%s]\n",
                    support_ctime(Now), pptr->name, pptr->username,
					pptr->hostname, oline->name, password);
		return 3;
	}

#endif
	if(UnderSVSNOOP)
	{
		send_prefix_to_client(meptr, pptr, "NOTICE %s :This server is under a SVSNOOP and may have no operators.", pptr->name);
		send_to_servers(NULL, ":%s WALLOPS :%s[@%s] was denied oper access. (SVSNOOP)", 
						meptr->name, pptr->name, meptr->name);
		return 3;
	}
	if (oline->flags & OLINE_FLAGS_LOCAL)
	{
		pptr->c_umodes |= CLIENT_UMODES_LOCOP;
	}
	else
	{
		pptr->c_umodes |= CLIENT_UMODES_OPER;
		Globalinfo->m_opers++;
		Globalinfo->g_opers++;
	}

	pptr->c_oline_flags = oline->flags;
	pptr->c_oline_name  = oline->name;


	assert(pptr->c_oslist);
	del_client_from_oslist(pptr->c_oslist, pptr);
	add_client_to_oslist(Opers, pptr);

	if(oline->wl)
		pptr->c_wl = oline->wl;

	if (oline->userflags)
		pptr->c_nflags = oline->userflags;
	else
	{
		pptr->c_nflags |= CLIENT_NFLAGS_SNOTICES;
		pptr->c_nflags |= CLIENT_NFLAGS_SWALLOPS;
		pptr->c_nflags |= CLIENT_NFLAGS_OWALLOPS;
		pptr->c_nflags |= CLIENT_NFLAGS_OKILLS;
		pptr->c_nflags |= CLIENT_NFLAGS_DENIES;
	}
	if (pptr->c_nflags & CLIENT_NFLAGS_SNOTICES)
		pptr->c_umodes |= CLIENT_UMODES_SMODE;
	if (pptr->c_nflags & CLIENT_NFLAGS_SWALLOPS)
		pptr->c_umodes |= CLIENT_UMODES_WMODE;
	if (pptr->c_nflags & CLIENT_NFLAGS_OKILLS)
		pptr->c_umodes |= CLIENT_UMODES_KMODE;

	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_YOUREOPER, pptr->name));
	
#if 0	/* moved a little bit lower */

	send_to_opers("%s (%s@%s) is now an operator (%c)", pptr->name, pptr->username, pptr->hostname,
			IsOper(pptr) ? 'O' : 'o');
	/*				^^^^^^^^^
	**				these should really be reversed
	**				but this is how old code is.
	*/
#endif

#ifdef HMC
	if(IsLocalAdmin(pptr))
		set_hmc_byletter(pptr, "L");
	if(IsNetAdmin(pptr))
		set_hmc_byletter(pptr, "N");
	if(!IsAnAdmin(pptr))
		set_hmc_byletter(pptr, "O"); /* just a regular oper */
#endif

        send_to_opers("%s (%s@%s) is now an operator (%c)", pptr->name, pptr->username, pptr->hostname,
				                        IsOper(pptr) ? 'O' : 'o');


	if(oline->setlevel)
	{
		pptr->c_oline_setlevel = oline->setlevel;
		if(SetLevelNotice)
			send_prefix_to_client(meptr, pptr, "NOTICE %s :Your SET level has been set to: %i", pptr->name, pptr->c_oline_setlevel);
	}
	else
		pptr->c_oline_setlevel = 0;

	if (vars_value_get_bynum(IRCD_VAR_GOODOPER_LOG))
		log_printf(_ircd_conf.log_goodoper, "%s %s!%s@%s [%s] opered\n",
			support_ctime(Now), pptr->name, pptr->username,
			pptr->hostname, oline->name);

	_umode_build_change(pptr, buf, oldmodes, SEND_UMODES);
#ifdef HMC
	send_to_servers(NULL, ":%s MODE %s %s !%s", prefix, pptr->name, buf, hmc_decide(pptr));
#else
	send_to_servers(NULL, ":%s MODE %s %s", prefix, pptr->name, buf);
#endif

	_umode_build_change(pptr, buf, oldmodes, -1);
#if 0
	send_prefix_to_client(pptr, pptr, "MODE %s %s", pptr->name, buf);
/* umodes only send client's nick, not full n!u@h */
#else
	send_to_client(pptr, ":%s MODE %s %s", prefix, pptr->name, buf);
#endif

	m_flags(lptr, pptr, prefix, NULL);

	ret = class_find(&(_ircd_conf.classes), oline->classname, &classptr);
	if (ret < 0)
		return 3;

	del_client_from_cllist(&(pptr->c_class->clients), pptr);
	if (!--pptr->c_class->num_on &&
		(pptr->c_class->flags & CLASS_FLAGS_ILLEGAL))
	{           
		class_del(&(_ircd_conf.classes), pptr->c_class);
	}

    add_client_to_cllist(&(classptr->clients), pptr);
    classptr->num_on++;
	pptr->c_class = classptr;
	pptr->c_maxsendq = classptr->maxsendq;
	pptr->c_pingfreq = classptr->pingfreq;
	pptr->c_rate_msg_limit = classptr->floodnum * 3;
	pptr->c_rate_time_limit = classptr->floodtime;

	return 3;
}

int m_pass(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*password = ircd_nextarg(&args, 0);

	if (!password || !*password)
	{
		if (IsClient(pptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NEEDMOREPARAMS,
				*pptr->name ? pptr->name : "*"), "PASS");
		return 3;
	}
	if (!IsLocal(pptr) || !IsUnknown(lptr))
	{
		if (IsClient(pptr))
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_ALREADYREGISTRED,
				pptr->name));
		return 3;
	}
	support_strncpy(pptr->c_passwd, password, sizeof(pptr->c_passwd));
	if (args && *args)
	{
		int	l = strlen(args);

		if (l < 2)
			return 3;
		if (!strcmp(args+l-2, "TS"))
			pptr->c_flags |= CLIENT_FLAGS_DOESTS;
	}
	return 3;
}

int m_wallops(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*message = ircd_nextarg(&args, 0);

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!message || !*message)
	{
		if (IsLocal(pptr))
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "WALLOPS");
		return 3;
	}

	if (IsClient(pptr))
	{
		if (MyClient(pptr) && !IsAnOper(pptr))
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
			return 3;
		}
		send_wallops_to_flag(CLIENT_NFLAGS_OWALLOPS, NULL, pptr, "%s",
				message);
	}
	else
		send_wallops_to_flag(CLIENT_NFLAGS_SWALLOPS, NULL, pptr, "%s",
				message);

	if (IsServer(pptr))
		send_to_servers(lptr, ":%s WALLOPS :%s", prefix, message);
	else
		send_to_servers(lptr, ":%s OPERWALL :%s", prefix, message);

	return 3;
}

int m_locops(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*message = ircd_nextarg(&args, 0);

	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!message || !*message)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "LOCOPS");
		return 3;
	}

	if (!IsAnOper(pptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}
	send_wallops_to_flag(CLIENT_NFLAGS_LWALLOPS, NULL, pptr, "LOCOPS - %s", message);
	return 3;
}

int m_trace(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*arg1;
	int				doall;
	IRCD_CLIENT		*acptr;
	char			cnamebuf[CLIENTNAMESIZE];
	int				numservers;
	int				numusers;
	int				i;
	IRCD_CLASSENTRY	**ceptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (args && *args && (_hunt_server(lptr, pptr,
					":%s TRACE %s :%s", 2, prefix, args) != HUNTED_ISME))
	{
			return 3;
	}

	switch (_hunt_server(lptr, pptr, ":%s TRACE :%s", 1, prefix, args))
	{
		case HUNTED_PASS:
		{
			char *name = 	ircd_nextarg(&args, 0);
			IRCD_CLIENT		*ac2ptr;

			if (!name || !*name)
				name = meptr->name;

			ac2ptr = hash_find_client_incl_hm(name);

			send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACELINK, pptr->name),
					VERSION, "", name, ac2ptr ? ac2ptr->c_from->name : "unknown");
		
			return 3;
		}
		case HUNTED_ISME:
			break;
		default:
			return 3;
	}

	arg1 = ircd_nextarg(&args, 0);

	if (!arg1 || !*arg1)
	{
		arg1 = meptr->name;
		doall = 1;
	}
	else
		doall = !match(arg1, meptr->name);

	if (!doall)
	{
		acptr = hash_find_client(arg1);
		if (acptr && IsLocal(acptr))
		{
			if (IsClient(acptr))
			{
				if (IsAnOper(acptr))
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TRACEOPERATOR, pptr->name),
						acptr->c_class->classname,
						ircd_clientname(acptr, 0, cnamebuf));
				else
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TRACEUSER, pptr->name),
						acptr->c_class->classname,
						ircd_clientname(acptr, 0, cnamebuf));
			}
			else if (IsServer(acptr))
			{
				numservers = numusers = 0;
				_count_servers_and_users(acptr, &numservers, &numusers);
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACESERVER, pptr->name),
					acptr->c_class->classname, numservers, numusers,
					ircd_clientname(acptr, 0, cnamebuf), acptr->c_connectby);
			}
		}
	}
	else
	{
		for(acptr=Locals;acptr;acptr=acptr->c_lo_next)
		{
			if (IsServer(acptr))
			{
				numservers = numusers = 0;
				_count_servers_and_users(acptr, &numservers, &numusers);
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACESERVER, pptr->name),
					acptr->c_class->classname, numservers, numusers,
					ircd_clientname(acptr, 0, cnamebuf), acptr->c_connectby);
			}
			else if (IsClient(acptr))
			{
				if (IsAnOper(acptr))
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TRACEOPERATOR, pptr->name),
						acptr->c_class->classname,
						ircd_clientname(acptr, 0, cnamebuf));
				}
				else if (IsAnOper(pptr))
				{
					send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_TRACEUSER, pptr->name),
						acptr->c_class->classname,
						ircd_clientname(acptr, 0, cnamebuf));
				}
			}
			else if (IsUnknown(acptr) && IsAnOper(pptr))
			{
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACEUNKNOWN, pptr->name),
					acptr->c_class ? acptr->c_class->classname : "unknown", ircd_clientname(acptr, 0, cnamebuf));
			}
		}
		if (IsAnOper(pptr))

			for(ceptr=_ircd_conf.classes.entries,i=_ircd_conf.classes.num;i--;ceptr++)
		{
			if ((*ceptr)->num_on > 0)
				send_prefix_to_client(meptr, pptr,
					IRCD_REPLY(RPL_TRACECLASS, pptr->name),
					(*ceptr)->classname, (*ceptr)->num_on);
		}

	}

	send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(RPL_ENDOFTRACE, pptr->name), arg1);

	return 3;
}


int m_userhost(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return m_useriphost(lptr, pptr, prefix, args, 0);
}

int m_userip(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return m_useriphost(lptr, pptr, prefix, args, 1);
}

static int m_useriphost(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args, int justip)
{
	char		*nicks = ircd_nextarg(&args, 0);
	int			maxuhost = 20;
	char		buf[475];
	IRCD_CLIENT *acptr;
	char		*end_ptr;
	char		*this_start_ptr;
	char		*sptr;
	char		*dptr = buf;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 1;
	}

	if (!nicks || !*nicks)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "USERHOST");
		return 1;
	}

	end_ptr = buf + sizeof(buf) - 1;
	*end_ptr = '\0';

	for(;maxuhost-- && nicks && *nicks;nicks=ircd_next_subarg_space(&args))
	{
		if ((acptr = hash_find_client(nicks)) != NULL)
		{
			if (IsClient(acptr))
			{
				this_start_ptr = dptr;

				if (dptr != buf)
					*dptr++ = ' ';

				/*
				** each section is: nick*=+~username@hostname 
				**  no * if not oper, - instead of + if away
				**
				** used to have snprintf/sprintf, but this is more efficient:
				*/

				for(sptr=acptr->name;dptr<end_ptr;sptr++,dptr++)
					if (!(*dptr=*sptr))
						break;

				if (dptr+4 >= end_ptr)
				{
					dptr = this_start_ptr;
					break;
				}

				if (IsAnOper(acptr))
					*dptr++ = '*';
				*dptr++ = '=';
				*dptr++ = acptr->c_away ? '-' : '+';

				for(sptr=acptr->username;dptr<end_ptr;sptr++,dptr++)
					if (!(*dptr=*sptr))
						break;

				if (dptr+2 >= end_ptr)
				{
					dptr = this_start_ptr;
					break;
				}

				*dptr++ = '@';
				
				if(justip)
				{
					for(sptr=acptr->c_ipstring;dptr<end_ptr;sptr++,dptr++)
						if (!(*dptr=*sptr))
							break;
				}
				else
				{
					for(sptr=acptr->hostname;dptr<end_ptr;sptr++,dptr++)
						if (!(*dptr=*sptr))
							break;
				}

				if (dptr >= end_ptr)
				{
					dptr = this_start_ptr;
					break;
				}
			}
		}
	}

	*dptr = '\0';

	send_prefix_to_client(meptr, pptr,
		IRCD_REPLY(RPL_USERHOST, pptr->name), buf);

	return 1;
}

int m_ison(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char			*nicks = ircd_nextarg(&args, 0);
	IRCD_CLIENT		*acptr;
	char			buf[475];
	char			*dptr = buf;
	char			*sptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 1;
	}

	if (!nicks || !*nicks)
	{
		send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(ERR_NEEDMOREPARAMS, pptr->name), "ISON");
		return 1;
	}

	for(;nicks && *nicks;nicks=ircd_next_subarg_space(&args))
	{
		if ((acptr = hash_find_client(nicks)) != NULL)
		{
			int len = strlen(acptr->name) + 1;

			if (dptr + len < buf + sizeof(buf))
			{
				if (dptr != buf)
					*dptr++ = ' ';
				for(sptr=acptr->name;*sptr;)
					*dptr++ = *sptr++;
			}
		}
	}
	*dptr = '\0';
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ISON, pptr->name), buf);
	return 1;
}

int m_operwall(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return m_wallops(lptr, pptr, prefix, args);
}

int m_note(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_service(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_squery(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_servlist(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_servset(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

int m_queue(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char		*command = ircd_nextarg(&args, 0);

	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!IsAnOper(lptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if (!command || !*command)
	{
		int	queue_size;
		int	queue_fl_size;
		int queue_malloced_size;

		queue_get_stats(&queue_size, &queue_fl_size, &queue_malloced_size);
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Queue space in use: %i out of %i allocated",
				pptr->name, queue_size, queue_malloced_size - queue_fl_size);
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Queue space in freelist: %i",
				pptr->name, queue_fl_size);
		return 3;
	}

	if (!strcasecmp(command, "flush"))
	{
		queue_cleanup();
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Queues flushed.",
				pptr->name);
	}
	else
		send_prefix_to_client(meptr, pptr, "NOTICE %s :Unknown queue command.",
				pptr->name);

	return 3;
}

int m_iphash(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return iphash_command(lptr, pptr, args);
}

int m_rejecthold(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_REJECTHOLD	*rptr;
	int				num = (int)vars_value_get_bynum(IRCD_VAR_REJECT_HOLD);

	if (IsServer(lptr))
		return 3;

	if (!IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!IsAnOper(lptr))
	{
		send_prefix_to_client(meptr, pptr,
			IRCD_REPLY(ERR_NOPRIVILEGES, pptr->name));
		return 3;
	}

	if ((rptr = Rejectholds) != NULL)
	{
		for(;rptr;rptr=rptr->rh_next)
		{
			ircd_uint32_t nipaddr = htonl(rptr->rh_rem_ipaddr);
			int e = (num <= 0) ? -1 : rptr->rh_firsttime - Now + num;

			if (e > 0)
				send_prefix_to_client(meptr, pptr, "NOTICE %s :%s put on hold %i second%s ago.  Expires in %i second%s.",
					pptr->name, inet_ntoa(*((struct in_addr *)&nipaddr)),
					Now - rptr->rh_firsttime, Now - rptr->rh_firsttime == 1 ? "" : "s",
					e, e == 1 ? "" : "s");
			else
				send_prefix_to_client(meptr, pptr, "NOTICE %s :%s put on hold %i second%s ago.  Entry expired.",
					pptr->name, inet_ntoa(*((struct in_addr *)&nipaddr)),
					Now - rptr->rh_firsttime, Now - rptr->rh_firsttime == 1 ? "" : "s");
		}
	}
	else
		send_prefix_to_client(meptr, pptr, "NOTICE %s :There are currently no IPs on REJECT_HOLD", pptr->name);
	return 3;
}
