/*
**   IRC - Internet Relay Chat, source/commands.c
**   Copyright (C) 1999 Chris Behrens
**
**   Portions Copyright (C) 1990 Jarkko Oikarinen and
**                      University of Oulu, Computing Center
**
**   Portions Copyright (C) 2001 Adam Loghry
** 
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"
//#include "cmds.h"
//#include "../lib/confparse/configparse.h"

RCSID("$Id: commands.c,v 1.20 2001/10/02 01:40:35 cbehrens Exp $");

/* this file houses whois and who (surprise surprise) --myriad */

static void do_who(IRCD_CLIENT *pptr, IRCD_CLIENT *acptr, IRCD_CHANNEL *chptr, IRCD_CLIENTLIST *clptr)
{
	char	status[5];
	int		i = 0;

	if (acptr->c_away)
		status[i++] = 'G';
	else
		status[i++] = 'H';
	if (IsAnOper(acptr))
		status[i++] = '*';
	if ((chptr != NULL) && (clptr == NULL)) /* should never happen */
	{
		IRCD_CHANNELLIST	*chlptr = IsMember(chptr, acptr);
		clptr = chlptr ? chlptr->clptr : NULL;
	}
	if (clptr)
	{
		if (clptr->modes & CHANNEL_MODES_CHANOP)
			status[i++] = '@';
		else if (clptr->modes & CHANNEL_MODES_VOICE)
			status[i++] = '+';
	}
	status[i] = '\0';
	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOREPLY, pptr->name),
			chptr ? chptr->name : "*",
			acptr->username,
			acptr->hostname,
			acptr->servername,
			acptr->name,
			status,
			acptr->c_hopcount,
			acptr->ircname);
}

int m_who(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char				*mask = ircd_nextarg(&args, 0);
	char				*mask2 = ircd_nextarg(&args, 0);
	IRCD_CHANNEL		*chptr;
	IRCD_CHANNELLIST	*chlptr;
	IRCD_CLIENT			*acptr;
	int					operonly = 0;
	int					maxmatches = 500;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if (!mask || !*mask)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFWHO, pptr->name),
			mask ? mask : "*");
		return 3;
	}

	if (mask2 && (*mask2 == 'o'))
		operonly = 1;

	collapse(mask);
	if (!operonly && (*mask == '*') && (!*(mask+1)))
	{
		if (!pptr->c_channels)
		{
			send_prefix_to_client(meptr, pptr,
				IRCD_REPLY(RPL_ENDOFWHO, pptr->name), mask);
			return 3;
		}
		chptr = pptr->c_channels->chptr;
		chlptr = pptr->c_channels;
	}
	else
	{
		chptr = hash_find_channel(mask);
		chlptr = NULL;
	}
	if (chptr)
	{
		IRCD_CLIENTLIST		*clptr;

		if (!chlptr)
			chlptr = IsMember(chptr, pptr);
		if (chlptr || CanWhoAll(pptr) || !(chptr->modes & CHANNEL_MODES_SECRET))
		{
			for(clptr=chptr->members;clptr;clptr=clptr->next)
			{
				if (operonly && !IsAnOper(clptr->cptr))
					continue;
				if (!chlptr && 
					((clptr->cptr->c_umodes & CLIENT_UMODES_IMODE) && !CanWhoAll(pptr)))
					continue;
				do_who(pptr, clptr->cptr, chptr, clptr);
			}
		}
	}
	else if (IsChannel(mask))
		; /* do nothing */
	else if ((acptr = hash_find_client(mask)) != NULL)
	{
		if (IsPerson(acptr) && (!operonly || IsAnOper(acptr)))
		{
			int					isinvis = acptr->c_umodes & CLIENT_UMODES_IMODE;
			IRCD_CHANNELLIST	*chlptr2;

			for(chlptr=acptr->c_channels;chlptr;chlptr=chlptr->next)
			{
				chlptr2 = _channel_ismember(chlptr->chptr, pptr);
				if (isinvis && !chlptr2 && !CanWhoAll(pptr))
					continue;
				if (chlptr2 || CanWhoAll(pptr) || (!isinvis && !(chlptr->chptr->modes & (CHANNEL_MODES_SECRET|CHANNEL_MODES_PRIVATE))))
					break;
			}
			do_who(pptr, acptr, chlptr ? chlptr->chptr : NULL,
						chlptr ? chlptr->clptr : NULL);
		}
	}
	else
	{
		int					i = 0;
		IRCD_CLIENT			**list;
		int					showperson;
		int					isinvis;
		IRCD_CHANNELLIST	*chlptr2;

		for(i=0;i<2;i++)
		{
			if (!i)
				list = Opers;
			else
			{
				if (operonly)
					break;
				list = Clients;
			}
			for(acptr=*list;acptr;acptr=acptr->c_os_next)
			{
				showperson = 0;
				isinvis = acptr->c_umodes & CLIENT_UMODES_IMODE;
				if (!(chlptr=acptr->c_channels) && !isinvis)
					showperson = 1;
				else if (chlptr)
				for(chptr=chlptr->chptr;chlptr;chlptr=chlptr->next)
				{
					chlptr2 = IsMember(chptr, pptr);
					if (!chlptr2 && isinvis)
						continue;
					if (chlptr2 || (!isinvis && !(chptr->modes &
								(CHANNEL_MODES_SECRET|CHANNEL_MODES_PRIVATE))))
					{
						showperson = 1;
						break;
					}
					if (!isinvis && (chptr->modes & CHANNEL_MODES_PRIVATE) &&
						!(chptr->modes & CHANNEL_MODES_SECRET))
						showperson = 1;
				}

				if ((showperson && !CanWhoAll(pptr)) &&
					(!match(mask, acptr->name) ||
					!match(mask, acptr->username) ||
					!match(mask, acptr->hostname) ||
					!match(mask, acptr->servername) ||
					!match(mask, acptr->ircname)))
				{
					do_who(pptr, acptr, chlptr ? chlptr->chptr : NULL,
								chlptr ? chlptr->clptr : NULL);
					if (!--maxmatches)
						break;
				}
			}
			if (!maxmatches)
				break;
		}
	}

	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFWHO, pptr->name),
			mask);

	return 3;
}

/* m_sw()
 * sets whois levels remotely. only used for ulined servers to set other clients,
 * and servers do NOT propogate changes like this when they internally alter someone's
 * whois level.
 */
/* FIX .. this is messy, re-write this before release */
int m_sw(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	IRCD_CLIENT	*cptr;
	char		*target	= ircd_nextarg(&args, 0);
	char		*blech	= ircd_nextarg(&args, 0); /* grr */
	int		number 	= 0;

	if(blech)
		number = atoi(blech);

	if(!IsServer(lptr))
		return 3;

	if(!target /* || !number */ )
		return 3; 

	cptr = hash_find_client(target);

	if(!cptr)
		return 3;

	if(!MyClient(cptr))
	{
		send_to_servers(lptr, ":%s SW %s %i", prefix, target, number);
		return 3; /* not mine, keep the show moving.. */
	}

	/* services only ? yeah, i guess.. */
	if(!is_ulined(lptr->name) /* && (cptr->servername != lptr->name) */ )
		return 3; /* a non-ulined server cannot change SW for a client that its own.. */

	if(!number)
	{
		cptr->c_wl = 0; /* removing the whois note */
	}
	else
	{	/* hooray, we have a valid number! */
		IRCD_WL **wlevels 	= &(_ircd_conf.wlevels);
		IRCD_WL *wl;
		int	wasgood 	= 0;

		for(wl=*wlevels;wl;wl=wl->next)
		{
			if(wl->number == number)
			{
				cptr->c_wl = number;
				cptr->c_old_wl = cptr->c_wl;
				wasgood = 1;
				break;
			}
		}

		if(!wasgood)
		{
			cptr->c_wl = 0;	
			cptr->c_old_wl = cptr->c_wl;
			return 3; /* we dont have that level on our list, remove
				   * the users previous whois level, if they even had one, and send us home */
		}

	}

	return 3;
}

static char *find_wlvl_text(int number)
{
        IRCD_WL      **wlevels = &(_ircd_conf.wlevels);
        IRCD_WL      *wl;

        for(wl=*wlevels;wl;wl=wl->next)
        {
		if(wl->number == number)
				return wl->text;
	}

	return NULL;
}

/* all this trouble for a few silly numerics.. *sigh* */
int m_rw(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	/* :irc.lackofdiscipline.com RW adam myr 123456 123456 :sucks the cock */
	/* :server.of.sending.client RW sender reciever sender_idle sender_ts :sender_wlvl */
	char 		*from		= ircd_nextarg(&args, 0);
	char		*to		= ircd_nextarg(&args, 0);
	char		*c_idle		= ircd_nextarg(&args, 0);
	char		*c_logon	= ircd_nextarg(&args, 0);
	char		*wlvlmsg	= ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	int		idle		= atoi(c_idle);
	int		logon		= atoi(c_logon);
	IRCD_CLIENT	*acptr;
	IRCD_CLIENT	*tptr;

	/* :betrayer.hashmark.net RW adam myr 12431 13413434 :i like cock */
	/* :irc.penis.com	  RW from to  from_TS from_IDLE :from_WHOIS */

	if(!from || !to)
		return 3;

	acptr 		= hash_find_client(from);		/* FROM IS THE CLIENT BEING WHOIS'D */
	tptr		= hash_find_client(to);			/* TO IS THE CLIENT WHO SENT THE WHOIS
								   AND IS NOW RECIEVING THE REPLY -myr */

	if(!acptr || !tptr)
		return 3;

	if(!MyClient(tptr) && !(strcasecmp(lptr->name, tptr->servername)))
	{
		acptr 	= hash_find_client(tptr->servername); /* reusing the memory */
		send_to_client(acptr, ":%s RW %s %s %s",
						prefix, from, to, args);
		return 3;
	}
	else
	{
		IRCD_CHANNELLIST	*chlptr;
		IRCD_CLIENTLIST		*clptr;
		char			buf[512];
		char			*stptr;
		char			*dptr = buf;
		char			*name;
		/* all the actual 'WHOIS' is done here. the original plan was to build it onto an extension of m_whois(),
		 * but i realized that would become very messy very quickly */

		/* fptr is the person being whois'd. */
		/* tptr is the person recieving the whois info [who requested the whois originally] */

		name    = *acptr->name ? acptr->name : "?";
		send_prefix_to_client(meptr, tptr, IRCD_REPLY(RPL_WHOISUSER, tptr->name),
				name, acptr->username, acptr->hostname, acptr->ircname);

		for(chlptr=acptr->c_channels;chlptr;chlptr=chlptr->next)
		{
			if (!((chlptr->chptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET)) && !CanSSC(tptr)) 
										|| (IsMember(chlptr->chptr, tptr) != NULL))
			{
				clptr = chlptr->clptr;
				if (chlptr->chptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET))
					*dptr++ = '%'; /* if its hidden, and they are an oper, tell them so. */
				if (clptr->modes & CHANNEL_MODES_CHANOP)
					*dptr++ = '@';
				else if (clptr->modes & CHANNEL_MODES_VOICE)
					*dptr++ = '+';
				for(stptr=chlptr->chptr->name;*stptr;)
					*dptr++ = *stptr++;
				if (dptr - buf > 200)
				{
					*dptr = '\0';
					send_prefix_to_client(meptr, tptr,
						IRCD_REPLY(RPL_WHOISCHANNELS, tptr->name),
						name, buf);
						dptr = buf;
				}
				else
					*dptr++ = ' ';
			}
		}
		if (dptr != buf)
		{
			*dptr = '\0';
			send_prefix_to_client(meptr, tptr,
					IRCD_REPLY(RPL_WHOISCHANNELS, tptr->name), name, buf);
		}

		if(IsRegistered(acptr))
			send_prefix_to_client(meptr, tptr, IRCD_REPLY(RPL_WHOISREGGED, tptr->name),
		                          name, (IsUmodeR(acptr)) ? "(only accepts msgs from registered users)" : "");

		if(vars_value_get_bynum(IRCD_VAR_WHOIS_DEAF_NOTE) && IsDeaf(acptr))
			send_prefix_to_client(meptr, tptr, IRCD_REPLY(RPL_WHOISDEAF, tptr->name), acptr->name);

		if(strlen(wlvlmsg) > 2) /* ... */
	                send_prefix_to_client(meptr, tptr, IRCD_REPLY(RPL_WHOISMESSAGE, tptr->name),
	                                name, wlvlmsg);

		if(IsAnOper(tptr) || !RestrictInfo)
			send_prefix_to_client(meptr, tptr, IRCD_REPLY(RPL_WHOISSERVER,
				tptr->name), name, (IsService(acptr)) ? NetHost : acptr->servername,
				(IsService(acptr)) ? NetName : acptr->c_servptr->ircname);

		if (acptr->c_away)
		{
			send_prefix_to_client(meptr, tptr, IRCD_REPLY(RPL_AWAY, tptr->name),
				name, acptr->c_away);
		}

		if (IsAnOper(acptr) && !IsService(acptr))
		{
			char *ptr = vars_value_get_bynum(IRCD_VAR_ISOPERSTRING);

			if (!ptr)
				ptr = "is an IRC Operator";

			send_prefix_to_client(meptr, tptr, "%d %s %s :%s",
					RPL_WHOISOPERATOR, tptr->name, acptr->name, ptr);
		}
		send_prefix_to_client(meptr, tptr, IRCD_REPLY(RPL_WHOISIDLE, tptr->name),
				name, Now - idle, logon);

		send_prefix_to_client(meptr, tptr, IRCD_REPLY(RPL_ENDOFWHOIS, tptr->name), acptr->name);
		
		return 3;
	}

	return 3;
}

int m_whois(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char		argssave[512];
	char		*whostr;
	char		*nick;
	char		*server;
	char		*sendtos[21];
	char		**st;
	char		**std;
	char		*name;
	char		*wlvl_msg = NULL;
	int		num = 0;
	IRCD_CLIENT	*acptr;

	if (!IsServer(lptr) && !IsReg(pptr))
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOTREGISTERED, NULL));
		return 3;
	}

	if(IsServer(lptr))
		return 3; /* .... retards */
	
	if (args)
		support_strncpy(argssave, args, sizeof(argssave));

	whostr = ircd_nextarg(&args, 0);
	server = ircd_nextarg(&args, 0);

	if (server && *server)
	{
		if (_hunt_server(lptr, pptr, ":%s WHOIS %s :%s", 1, prefix, argssave) != HUNTED_ISME)
			return 3;
		whostr = server;
	}

	if (!whostr || !*whostr)
	{
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NONICKNAMEGIVEN, pptr->name));
		return 3;
	}

	st = sendtos;
	for(;(nick = ircd_next_subarg_comma(&whostr)) != NULL;)
	{
		if (!*nick)
			continue;
		*st = nick;
		st++;
		if (++num == 20)
			break;
	}
	*st = NULL;
	st = sendtos;

	for(;;)
	{
		nick = *st;
		st++;
		if (!nick)
			break;
		/* already done it ? */
		for(std=sendtos;*std != nick;std++)
			if (!strcmp(*std, nick))
				break;
		if (*std != nick)
			continue;

		acptr 	= hash_find_client(nick);

		if (!acptr)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(ERR_NOSUCHNICK,
				pptr->name), nick);
			continue;
		}
		/*	if they are my client, and they arent on a ulined server, send the whois command somewhere else.
		 *	we want to handle ulined server requests locally.. */
		else if (!MyClient(acptr) && !is_ulined(acptr->servername) && !IsService(acptr))
		{
			IRCD_CLIENT *sptr2 = hash_find_client(acptr->servername);
			send_to_client(sptr2, ":%s WHOIS %s :%s", pptr->name, acptr->name, acptr->name);
			continue;
		}
		else if (!MyClient(pptr) /* && RestrictInfo */ )
		{
			char times[40];
			IRCD_CLIENT *sptr2 = hash_find_client(pptr->servername);
			sprintf(times, "%i %i", acptr->c_idletime, acptr->c_firsttime);
			send_to_client(sptr2, ":%s RW %s %s %s :%s", 
						meptr->name, acptr->name, pptr->name, times,
							(acptr->c_wl) ? find_wlvl_text(acptr->c_wl) : "");
			/* betrayer.hashmark.net RW UserA UserB UserA's_idle UserA's_signon :UserA's_whois_lvl */
			continue;
		}
		else if (!IsClient(acptr))
			continue;

		name 	= *acptr->name ? acptr->name : "?";
		
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISUSER, pptr->name),
				name, acptr->username, acptr->hostname, acptr->ircname);
		/* channels */
		if(!IsService(acptr))
		{
			IRCD_CHANNELLIST	*chlptr;
			IRCD_CLIENTLIST		*clptr;
			char				buf[512];
			char				*stptr;
			char				*dptr = buf;

			for(chlptr=acptr->c_channels;chlptr;chlptr=chlptr->next)
			{
				if (!((chlptr->chptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET)) && !CanSSC(pptr)) 
											|| (IsMember(chlptr->chptr, pptr) != NULL))
				{
					clptr = chlptr->clptr;
					if (chlptr->chptr->modes & (CHANNEL_MODES_PRIVATE|CHANNEL_MODES_SECRET))
						*dptr++ = '%'; /* if its hidden, and they are an oper, tell them so. */
					if (clptr->modes & CHANNEL_MODES_CHANOP)
						*dptr++ = '@';
					else if (clptr->modes & CHANNEL_MODES_VOICE)
						*dptr++ = '+';
					for(stptr=chlptr->chptr->name;*stptr;)
						*dptr++ = *stptr++;
					if (dptr - buf > 200)
					{
						*dptr = '\0';
						
						send_prefix_to_client(meptr, pptr,
							IRCD_REPLY(RPL_WHOISCHANNELS, pptr->name),
							name, buf);
						dptr = buf;
					}
					else
						*dptr++ = ' ';
				}
			}
			if (dptr != buf)
			{
				*dptr = '\0';
				send_prefix_to_client(meptr, pptr,
						IRCD_REPLY(RPL_WHOISCHANNELS, pptr->name), name, buf);
			}
		}

		if(IsRegistered(acptr))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISREGGED, pptr->name),
		                          name, (IsUmodeR(acptr)) ? "(only accepts msgs from registered users)" : "");
		}
		if(IsSpoofed(acptr) && MyClient(acptr) && (pptr->c_oline_flags & OLINE_FLAGS_CANSEESPOOFS))
		{
		 	send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISSPOOFED, pptr->name),
					pptr->name, acptr->name, acptr->rl_username, acptr->rl_hostname, acptr->rl_ipstring);
		}
		
		if(vars_value_get_bynum(IRCD_VAR_WHOIS_DEAF_NOTE) && IsDeaf(acptr))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISDEAF, pptr->name), acptr->name);
		}

		wlvl_msg = find_wlvl_text(acptr->c_wl);

		if(acptr->c_wl && wlvl_msg != NULL)
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISMESSAGE, pptr->name), name, wlvl_msg);
	
#if 0
		if((!vars_value_get_bynum(IRCD_VAR_RESTRICT_INFO) && !IsService(acptr)) || (vars_value_get_bynum(IRCD_VAR_RESTRICT_INFO) && IsAnOper(pptr)))
#endif
#if 0
		if(!RestrictInfo || (RestrictInfo && IsAnOper(pptr))) 
#endif
		if(IsAnOper(pptr) || !RestrictInfo)
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISSERVER,
				pptr->name), name, (IsService(acptr)) ? NetHost : acptr->servername,
				(IsService(acptr)) ? NetName : acptr->c_servptr->ircname);

		if (acptr->c_away)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_AWAY, pptr->name),
				name, acptr->c_away);
		}
#if 0
		wlvl_msg = find_wlvl_text(acptr->c_wl);	

		if (acptr->c_wl && wlvl_msg != NULL)
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISMESSAGE, pptr->name),
						name, wlvl_msg);
		}
#endif
		if (IsAnOper(acptr) && !IsService(acptr))
		{
			char *ptr = vars_value_get_bynum(IRCD_VAR_ISOPERSTRING);

			if (!ptr)
				ptr = "is an IRC Operator";

			send_prefix_to_client(meptr, pptr, "%d %s %s :%s",
					RPL_WHOISOPERATOR, pptr->name, acptr->name, ptr);
#if 0
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISOPERATOR, pptr->name),
				name);
#endif
		}
		if (IsLocal(acptr) && !IsService(acptr))
		{
			send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_WHOISIDLE, pptr->name),
				name, Now - acptr->c_idletime, acptr->c_firsttime);
		}
	}

	name = strchr(argssave, ' ');
	if (name)
		*name = '\0';

	if(MyClient(pptr))
		send_prefix_to_client(meptr, pptr, IRCD_REPLY(RPL_ENDOFWHOIS, pptr->name),
			argssave);

	return 3;
}

int m_whowas(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return whowas_command(lptr, pptr, prefix, args);
}
