#ifndef __VARS_H__
#define __VARS_H__

typedef struct _ircd_var	IRCD_VAR;

#include "defines.h"
#include "structs.h"

int vars_logfilename_check(IRCD_CLIENT *pptr, IRCD_VAR *v, char *value);
int vars_motdfilename_check(IRCD_CLIENT *pptr, IRCD_VAR *v, char *value);
int vars_netname_check(IRCD_CLIENT *pptr, IRCD_VAR *v, char *value);
int vars_nethost_check(IRCD_CLIENT *pptr, IRCD_VAR *v, char *value);
int vars_whowas_change(IRCD_CLIENT *pptr, IRCD_VAR *v);
int vars_logfile_change(IRCD_CLIENT *pptr, IRCD_VAR *v);

struct _ircd_var
{
	char	*v_name;

#define IRCD_VAR_TYPE_INT		1
#define IRCD_VAR_TYPE_STRING		2
#define IRCD_VAR_TYPE_BOOL		3

	char	v_type;

	int		v_val_min;	/* for ints, minimum value..for string, min length */
	int		v_val_max;	/* for ints, maximum value..for string, max length */

	void	*v_value;

	int		(*v_functionbefore)(IRCD_CLIENT *pptr, IRCD_VAR *vptr, char *value);
	int		(*v_functionafter)(IRCD_CLIENT *pptr, IRCD_VAR *vptr);

	int	v_level;
	int	v_net;
};

#define IRCD_VAR_CHANOPSONSPLIT			0
#define IRCD_VAR_MAXCLIENTS			1
#define IRCD_VAR_IDLE_CHECK			2
#define IRCD_VAR_IDLE_TIME			3
#define IRCD_VAR_GLINE_LOG			4
#define IRCD_VAR_GLINE_LOGFILE			5
#define IRCD_VAR_GOODOPER_LOG			6
#define IRCD_VAR_GOODOPER_LOGFILE		7
#define IRCD_VAR_BADOPER_LOG			8
#define IRCD_VAR_BADOPER_LOGFILE		9
#define IRCD_VAR_AUTH_CHECK			10
#define IRCD_VAR_SHOW_NICKCHANGES		11
#define IRCD_VAR_ABUSE_EMAIL			12
#define IRCD_VAR_MOTD_FILE			13
#define IRCD_VAR_WHOIS_DEAF_NOTE		14
#define IRCD_VAR_CLONES_LOG			15
#define IRCD_VAR_CLONES_LOGFILE			16
#define IRCD_VAR_USERS_LOG			17
#define IRCD_VAR_USERS_LOGFILE			18
#define IRCD_VAR_USERNAMECHARS			19
#define IRCD_VAR_ISOPERSTRING			20
#define IRCD_VAR_CLIENTCONNECT			21
#define IRCD_VAR_IPADDR_ACCEPT_RATE_NUM		22
#define IRCD_VAR_IPADDR_ACCEPT_RATE_TIME	23
#define IRCD_VAR_IPADDR_CLIENT_RATE_NUM		24
#define IRCD_VAR_IPADDR_CLIENT_RATE_TIME	25
#define IRCD_VAR_WHOWAS_SIZE			26
#define IRCD_VAR_RATELIMIT_TIME			27
#define IRCD_VAR_RATELIMIT_MOTD_NUM		28
#define IRCD_VAR_RATELIMIT_INFO_NUM		29
#define IRCD_VAR_RATELIMIT_STATS_NUM		30
#define IRCD_VAR_REJECT_HOLD			31
#define IRCD_VAR_PRIORITY_IDLE			32
#define IRCD_VAR_PRIORITY_LAG			33
#define IRCD_VAR_PRIORITY_FDLISTS		34
#define IRCD_VAR_DEFAULT_MAXCHANNELS		35
#define IRCD_VAR_SETLEVELNOTICE			36
#define IRCD_VAR_OPERSCANJOINONSPLIT		37
#define IRCD_VAR_SHOW_CLIENTCOUNTER		38
#define IRCD_VAR_IGNORE_REDUNDANT_MODES		39
#define IRCD_VAR_RESTRICT_INFO			40
#define IRCD_VAR_NETWORK_HOST			41
#define IRCD_VAR_NETWORK_NAME			42
#define IRCD_VAR_MAX_GLINESEARCH		43
#define IRCD_VAR_SHOW_GLINE_MSGS		44
#define IRCD_VAR_SETDEFAULTCMODES		45
#define IRCD_VAR_SERVICES_SERVER		46
#define IRCD_VAR_SHOW_GLINE_EXPIRES		47
#define IRCD_VAR_SHOW_FLAGS_NOTICE		48
#define IRCD_VAR_HUB				49
#define IRCD_VAR_HIDECHANQUIETS			50
#define IRCD_VAR_HIDECHANEXCEPTS		51
#define IRCD_VAR_HIDECHANBANS			52
#define IRCD_VAR_WARN_CHANNEL_DENY		53

#ifdef DEFINE_VARS_DEFINES

IRCD_VAR	Vars_Table[] =
{
/*      	name                type       min   max   default value  * * level(0,1,2)) "target"(1 = network, 0 = local, note that other servers can change this) */
	{ "CHANOPSONSPLIT",	IRCD_VAR_TYPE_INT, 0,	999, (void *)0, NULL, NULL, 10, 1 },
	{ "MAXCLIENTS",		IRCD_VAR_TYPE_INT,	0,	0, (void *)1024, NULL, NULL, 20, 0 },
	{ "IDLE_CHECK",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)0, NULL, NULL, 20, 0 },
	{ "IDLE_TIME",		IRCD_VAR_TYPE_INT,	300,0,	(void *)300, NULL, NULL, 20, 0 },
	{ "GLINE_LOG",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)0, NULL, vars_logfile_change, 30, 0 },
	{ "GLINE_LOGFILE",	IRCD_VAR_TYPE_STRING,	0,	255, "logs/gline.log", vars_logfilename_check, vars_logfile_change, 30, 0 },
	{ "GOODOPER_LOG",	IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)0, NULL, vars_logfile_change, 30, 0 },
	{ "GOODOPER_LOGFILE",	IRCD_VAR_TYPE_STRING,	0,	255, "logs/good-opers.log", vars_logfilename_check, vars_logfile_change, 30, 0 },
	{ "BADOPER_LOG",	IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)0, NULL, vars_logfile_change, 30, 0 },
	{ "BADOPER_LOGFILE",	IRCD_VAR_TYPE_STRING,	0,	255, "logs/bad-opers.log", vars_logfilename_check, vars_logfile_change, 30, 0 },
	{ "AUTH_CHECK",		IRCD_VAR_TYPE_BOOL, 0,	0,	(void *)0, NULL, NULL, 20, 0 },
	{ "SHOW_NICKCHANGES",	IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)0, NULL, NULL, 20, 0 },
	{ "ABUSE_EMAIL",                IRCD_VAR_TYPE_STRING,   1,      80,     (void *)DEFAULT_ABUSE_EMAIL, NULL, NULL, 30, 1 },
	{ "MOTD_FILE",		IRCD_VAR_TYPE_STRING,	0,	255,	"ircd.motd", vars_motdfilename_check, NULL, 30, 0 },
	{ "WHOIS_DEAF_NOTE",    IRCD_VAR_TYPE_BOOL,     0,      0,      	(void *)0, NULL, NULL, 10, 1 },
	{ "CLONES_LOG",		IRCD_VAR_TYPE_BOOL,	0,	0,		(void *)0, NULL, vars_logfile_change, 30, 0 },
	{ "CLONES_LOGFILE",	IRCD_VAR_TYPE_STRING,	0,	255, "logs/clones.log", vars_logfilename_check, vars_logfile_change, 30, 0 },
	{ "USERS_LOG",		IRCD_VAR_TYPE_BOOL,	0,	0,		(void *)0, NULL, vars_logfile_change, 30, 0 },
	{ "USERS_LOGFILE",	IRCD_VAR_TYPE_STRING,	0,	255, "logs/users.log", vars_logfilename_check, vars_logfile_change, 30, 0 },
	{ "USERNAMECHARS",	IRCD_VAR_TYPE_STRING,	0,	0,		(void *)"-_", NULL, NULL, 20, 1 },
	{ "ISOPERSTRING",	IRCD_VAR_TYPE_STRING,	1, 80, "is an IRC Operator", NULL, NULL, 20, 0 },
	{ "CLIENTCONNECT",	IRCD_VAR_TYPE_BOOL,	0,	0,		(void *)0, NULL, NULL, 10, 0 },
	{ "IPADDR_ACCEPT_RATE_NUM",	IRCD_VAR_TYPE_INT,	0,	0, 	(void *)20, NULL, NULL, 20, 0 },
	{ "IPADDR_ACCEPT_RATE_TIME",	IRCD_VAR_TYPE_INT,	0, 0, 		(void *)10, NULL, NULL, 20, 0 },
	{ "IPADDR_CLIENT_RATE_NUM",	IRCD_VAR_TYPE_INT,	0,	0, 	(void *)3, NULL, NULL, 20, 0 },
	{ "IPADDR_CLIENT_RATE_TIME",	IRCD_VAR_TYPE_INT,	0, 0, 		(void *)10, NULL, NULL, 20, 0 },
	{ "WHOWAS_SIZE",	IRCD_VAR_TYPE_INT,	1,	100000,		(void *)(16*1024), NULL, vars_whowas_change, 20, 0 },
	{ "RATELIMIT_TIME",			IRCD_VAR_TYPE_INT,	0, 0, 	(void *)6, NULL, NULL, 20, 0 },
	{ "RATELIMIT_MOTD_NUM",		IRCD_VAR_TYPE_INT,	0, 0, 		(void *)3, NULL, NULL, 20, 0 },
	{ "RATELIMIT_INFO_NUM",		IRCD_VAR_TYPE_INT,	0, 0, 		(void *)3, NULL, NULL, 20, 0 },
	{ "RATELIMIT_STATS_NUM",	IRCD_VAR_TYPE_INT,	0, 0, 		(void *)3, NULL, NULL, 20, 0 },
	{ "REJECT_HOLD",			IRCD_VAR_TYPE_INT, 0, 3600, 	(void *)0, NULL, NULL, 20, 0 },
	{ "PRIORITY_IDLE",			IRCD_VAR_TYPE_INT,	120, 0, (void *)(20*60), NULL, NULL, 30, 0 },
	{ "PRIORITY_LAG",			IRCD_VAR_TYPE_INT,	1, 100, (void *)3, NULL, NULL, 30, 0 },
	{ "PRIORITY_FDLISTS",		IRCD_VAR_TYPE_INT,	0, 99, 		(void *)4, NULL, NULL, 30, 0 },
	{ "DEFAULT_MAXCHANNELS",	IRCD_VAR_TYPE_INT,	0, MAXCHANNELSPERUSER, (void *)10, NULL, NULL, 20, 0 },
	{ "SET_LEVEL_NOTICE",           IRCD_VAR_TYPE_BOOL,     0,      0,      (void *)DEFAULT_SET_LEVEL_NOTICE,       NULL, NULL, 20, 0 },
	{ "OPERSCANJOINONSPLIT",	IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)0, NULL, NULL, 30, 0 },
	{ "SHOW_CLIENTCOUNTER",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_SHOW_CLIENTCOUNTER, NULL, NULL, 10, 1 },
	{ "IGNORE_REDUNDANT_MODES",	IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_IGNORE_REDUNDANT_MODES, NULL, NULL, 10, 1 },
	{ "RESTRICT_INFO",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_RESTRICT_INFO, NULL, NULL, 30, 1 },
	{ "NETWORK_HOST",		IRCD_VAR_TYPE_STRING,	1,	100, 	(void *)DEFAULT_NETWORK_HOST, vars_nethost_check, NULL, 30, 1 },
	{ "NETWORK_NAME",		IRCD_VAR_TYPE_STRING,	1,	100, 	(void *)DEFAULT_NETWORK_NAME, vars_netname_check, NULL, 30, 1 },
	{ "MAX_GLINESEARCH",		IRCD_VAR_TYPE_INT,	0,	0,	(void *)DEFAULT_MAX_GLINESEARCH, NULL, NULL, 20, 0 },
	{ "SHOW_GLINE_MSGS",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_SHOW_GLINE_MSGS, NULL, NULL, 30, 1 },
	{ "SET_DEFAULTCMODES",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_SET_DEFAULTCMODES, NULL, NULL, 30, 1  },
	{ "SERVICES_SERVER",		IRCD_VAR_TYPE_STRING,	1,	100, 	(void *)SERVICES_SERVER, NULL, NULL, 30, 1 },
	{ "SHOW_GLINE_EXPIRES",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_SHOW_GLINE_EXPIRES, NULL, NULL, 30, 0 },
	{ "SHOW_FLAGS_NOTICE",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_SHOW_FLAGS_NOTICE, NULL, NULL, 10, 0 },
#ifdef NO_HUB
	{ "HUB",			IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)0, NULL, NULL, 101, 0 },
#else
	{ "HUB",			IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_HUB_STATUS, NULL, NULL, 40, 0 },
#endif
	{ "HIDE_CHANNELQUIETS",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_HIDE_CHANQUIETS,	NULL, NULL, 20, 1 },
	{ "HIDE_CHANNELEXCEPTS",		IRCD_VAR_TYPE_BOOL,	0,	0, 	(void *)DEFAULT_HIDE_CHANEXCEPTS,	NULL, NULL, 20, 1 },
	{ "HIDE_CHANNELBANS",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_HIDE_CHANBANS,		NULL, NULL, 20, 1 },
	{ "WARN_CHANNEL_DENY",		IRCD_VAR_TYPE_BOOL,	0,	0,	(void *)DEFAULT_WARN_CHANNEL_DENY,	NULL, NULL, 20, 1 },
	{ NULL, 0, 0, 0, NULL }
};

#else

extern IRCD_VAR Vars_Table[];

#endif

#define vars_value_get_bynum(__x)	Vars_Table[(__x)].v_value

#ifndef NO_HUB
#define ImAHub				vars_value_get_bynum(IRCD_VAR_HUB)
#else
#define	ImAHub				0
#endif

/* saves typing, not used in every spot in the code where such
 * variables are referenced */
#define NetName				vars_value_get_bynum(IRCD_VAR_NETWORK_NAME)
#define NetHost				vars_value_get_bynum(IRCD_VAR_NETWORK_HOST)
#define RestrictInfo			vars_value_get_bynum(IRCD_VAR_RESTRICT_INFO)
#define HideCQuiets			vars_value_get_bynum(IRCD_VAR_HIDECHANQUIETS)
#define HideCExcepts			vars_value_get_bynum(IRCD_VAR_HIDECHANEXCEPTS)
#define HideCBans			vars_value_get_bynum(IRCD_VAR_HIDECHANBANS)
#define SetLevelNotice			vars_value_get_bynum(IRCD_VAR_SETLEVELNOTICE)

/* doesnt really fit in here, but its better than the other areas i suppose */
#ifdef DEBUG_MSGS
#define ImDebugging	1
#else
#define ImDebugging	0
#endif

void vars_init(void);
IRCD_VAR *vars_find_byname(char *name);
void *vars_value_get_byname(char *name);
int vars_value_set_byname(IRCD_CLIENT *pptr, char *name, void *value);
int vars_value_set_bynum(IRCD_CLIENT *pptr, int num, void *value);
IRCD_VAR *vars_value_getall(void);
IRCD_VAR *vars_value_getnext(IRCD_VAR *v);
int vars_init_reload(void *varsptr);
int vars_deinit_reload(void **varsptr);

#endif
