/*
**   IRC - Internet Relay Chat, source/services.c
**   Copyright (C) 2001 Adam Loghry
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "ircd_includes.h"
#include "commands.h"

RCSID("$Id: services.c,v 1.1.1.1 2001/12/18 07:39:59 myr Exp $");

int is_ulined(char *servername)
{
                IRCD_ULINE      **Ulines = &(_ircd_conf.ulines);
                IRCD_ULINE      *ul;

                for(ul=*Ulines;ul;ul=ul->next)
                {
			if(!myncmp(servername, ul->servername, sizeof(ul->servername)))
				return 1;
                }

	return 0; /* not ulined */
}

static int nickname_check(char *nick)
{
	char	*ch;

	if ((*nick == '-') || isdigit(*nick))
		return 0;
	for (ch = nick; *ch && (ch - nick) < NICKLEN; ch++)
		if (!isvalid(*ch) || isspace(*ch))
			return 0;

	return 1; /* its good */
}


/* possible commands;
 *
 * SVSNICK %s %s :%i
 * SVSMODE %s %s :%i
 * SVSJOIN %s %i :%s
 * SVSPART %s %i :%s
 * SVSNOOP %s [+/-]
 */

/* svsmode
 * (an interpretation of bahamut's code)
 * example: SVSMODE myriad +r :123456
 */ 
int m_svsmode(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char	*nick			= ircd_nextarg(&args, 0);
	char	*modechange		= ircd_nextarg(&args, 0);
	char	*cmd_ts			= ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	IRCD_CLIENT 	*acptr;
	int	ts			= atoi(cmd_ts);
	int	oldmodes;
	int	*i;
	char	*ptr;
	char	pm = '+';
	extern int _user_modes[];

	if(!is_ulined(lptr->name) || !modechange)
		return 3; /* dont care about them, not ulined or bad syntax */

	if(IsChannel(nick))
		return 3; /* we arent doing this yet, dunno if we want to or not */

	if(!(acptr = hash_find_client(nick)))
		return 3; /* no user on that nick */

	if(!MyClient(acptr))
	{
		/* its not my client, so just propogate the command and drop.. */
		send_to_servers(lptr, ":%s SVSMODE %s %s", prefix, acptr->name, modechange);
		return 3;
	}


	if((acptr->c_creation != ts) && (ts != 0))
		return 3; /* wrong TS, not our client! */
	
	/* m_mode() */
	oldmodes 	= acptr->c_umodes;
	for(ptr=modechange;*ptr;ptr++)
	{
		switch(*ptr)
		{
			case '+':
			case '-':
				pm = *ptr;
				break;
			case ' ':
			case '\n':
			case '\r':
			case '\t':
				break;
			default:
				for(i=_user_modes;*i;i+=2)
					if(*ptr == *(i+1))
						break;

			if(pm == '+')
				acptr->c_umodes |= *i;
			else
				acptr->c_umodes &= ~(*i);
		}
	}

	if (oldmodes == pptr->c_umodes)
		return 3;

	/* tell the shmuck */
	send_to_client(acptr, ":%s MODE %s %s", (vars_value_get_bynum(IRCD_VAR_RESTRICT_INFO)) ? vars_value_get_bynum(IRCD_VAR_NETWORK_HOST) : pptr->name, acptr->name, modechange);

	/* propogate the change -- all servers must accept this, or we get "issues" */
	// send_to_servers(lptr, ":%s SVSMODE %s %s", prefix, acptr->name, modechange);
	send_to_servers(NULL, ":%s MODE %s %s", acptr->name, acptr->name, modechange);

	return 3;
}

/* m_svsnick()
 * (another interpretation)
 * some of the things i deemed slightly odd in the bahamut version
 * of this command have been removed/altered .. no wonder those commands arent too
 * reliable .. */
/* example: SVSNICK myriad dairym :123456 */
int m_svsnick(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	char 		*oldnick 		= ircd_nextarg(&args, 0);
	char 		*newnick		= ircd_nextarg(&args, 0);
	char		*cmd_ts			= ircd_nextarg(&args, NEXTARG_FLAGS_LASTARG);
	int		ts			= atoi(cmd_ts);
	IRCD_CLIENT 	*olduptr;
	IRCD_CLIENT	*newuptr;

	if((!is_ulined(pptr->name) && !is_ulined(prefix)) || !newnick)
		return 3;

	if(!nickname_check(newnick) || !nickname_check(oldnick))
		return 3; /* grr */

	olduptr		= hash_find_client(oldnick);
	newuptr		= hash_find_client(newnick);

	if(!olduptr)
	{
		/* uh, svsnick sent us a command for a nick that doesnt exist. */
#if 0
		send_to_servers(lptr, ":%s SVSNICK %s %s", prefix, oldnick, newnick);
#endif
		return 3;
	}

	if(!MyClient(olduptr))
	{
		/* old guy isnt my client, propogate the nick change and let the guy's own server 
		 * handle his business */

		 send_to_servers(lptr, ":%s SVSNICK %s %s", prefix, oldnick, newnick);
		 return 3;
	}

	if((olduptr->c_creation != ts) && (ts != 0))
		return 3; /* not the same guy the SVS command was meant for. */
	
	/* okay, by now, we have a confirmed client(that we own) on the old nick, and there is a client
	 * sitting on the old nick. take care of him, then change our guy's nick */
	/* ::sigh:: the checking here is done by whatever server owned the remote client, it is ASSUMED to have
	 * deleted its own user in favor of the new one. if it doesnt, we get some rather fucked up situations */
	if(newuptr)
	{
		char tmp[KILLLEN];

		if(!MyClient(newuptr))
		{
			/* theres a client on that nick! boo! kill it! */
			send_to_client(newuptr, ":%s NOTICE %s :*** SVSNICK Collision! (%s->%s) [%s]", 
						prefix, newuptr->name, oldnick, newnick, "REMOTE");
			send_to_servers(lptr, ":%s KILL %s :%s->%s (SVSNICK Collision!)",
						prefix, newnick, oldnick, newnick);
			newuptr->c_flags |= CLIENT_FLAGS_KILLED;
			sprintf(tmp, "Killed (%s (%s->%s (SVSNICK Collision!)))", prefix, oldnick, newnick);
			client_del(meptr, newuptr, tmp);
			/* client_del(meptr, newuptr, ("Killed ( " + prefix + " (" + oldnick + "->" + newnick + "(SVSNICK Collision!)))")); */
		}
		else 
		{
			/* its mine, i have to kill the connection myself */
			send_to_client(newuptr, ":%s NOTICE %s :*** SVSNICK Collision! (%s->%s) [%s]", 
						prefix, newuptr->name, oldnick, newnick, "LOCAL");
			newuptr->c_flags |= CLIENT_FLAGS_KILLED;
			sprintf(tmp, "Killed (%s (%s->%s (SVSNICK Collision!)))", prefix, oldnick, newnick);
			client_del(meptr, newuptr, tmp);
			/* client_del(meptr, newuptr, ("Killed ( " + prefix + " (" + oldnick + "->" + newnick + "(SVSNICK Collision!)))")); */
		}

	}

	/* okay, took care of the other guy(if there was another guy), now take care of our own
	 * dude's nick.. quite a bit of this is out of m_nick() */

	olduptr->c_creation = Now; 

	if(IsRegistered(olduptr))
		olduptr->c_umodes &= ~CLIENT_UMODES_REGISTERED;

	whowas_add_history(olduptr, 1);
	send_to_common_channels(olduptr, "NICK :%s", newnick);

	send_to_servers(NULL, ":%s NICK %s %i",
			olduptr->name, newnick, olduptr->c_creation);

	hash_del_client(olduptr);
	strcpy(olduptr->name, newnick);
	hash_add_client(olduptr);

	/* we dont send out a SVSNICK if we already changed the guys nick.. so dont sweat it */
#if 0
	send_to_servers(lptr, ":%s SVSNICK %s %s", prefix, oldnick, newnick);
#endif

	if(vars_value_get_bynum(IRCD_VAR_SHOW_NICKCHANGES))
			send_to_flag(CLIENT_NFLAGS_NICKCHANGES, NULL,
				"Forced Nick Change: From %s to %s through SVSNICK [%s@%s]",
					oldnick, newnick, olduptr->username, olduptr->hostname);
	return 3;
}

/* m_svsnoop()
 * designed to be identical to the bahamut version
 * syntax is:
 * :server SVSNOOP other.server.net +/-
 * 	+ -> removes all opers on the server(adding a SVSNOOP)
 * 	- -> returns all the olines to the server(removing that SVSNOOP)
 */
int m_svsnoop(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	extern int 	UnderSVSNOOP;
	char		*target		= ircd_nextarg(&args, 0);
	int		add = 0;
	IRCD_CLIENT 	*srvptr;

        if((!is_ulined(pptr->name) && !is_ulined(prefix)) || !target)
		                return 3;
	if(*target == '+') {
		add = 1;
		target++;
	} else if(*target == '-') {
		add = 0;
		target++;
	} else {
		return 3;	/* didnt get a sane msg from the other server. */
	}

	srvptr = hash_find_client(target);

	if(IsMe(srvptr))
		UnderSVSNOOP = (add) ? 1 : 0;
	else
		send_to_servers(lptr, ":%s SVSNOOP %s%s", prefix, 
					(add) ? "+" : "-", target);
	return 3;
}

/* :%s SVSJOIN %s %i :%s */
int m_svsjoin(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}

#if 0 /* make services KICK the client, if they are supposed to get
	 out forcibly.. */
int m_svspart(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}
#endif

/* :%s SVSDENY ?????? */
int m_svsdeny(IRCD_CLIENT *lptr, IRCD_CLIENT *pptr, char *prefix, char *args)
{
	return 3;
}
